/************************************************************************/
/* @file   aes.s                                                        */
/* @author yangx                                                        */
/* @date   2022-4-11                                                    */
/* @brief  two functions                                                */
/*         -- AES_Encrypt_GCB [public]                                  */
/*            Implementation of AES-128 masked by code-based masking,   */ 
/*            please refer to xxxxxxx to see the detailed masking       */
/*            scheme. The generator matrix is fixed as [1, 0; 1, 1].    */
/*            This function has no output and the input is the          */ 
/*            address of cyphertext.                                    */ 
/*         -- Sbox [private]                                            */
/*            Implementations of SubBytes masked by code-based masking. */
/*            Note that S-box can be implemented by an inverse and an   */
/*            affine transformation.                                    */
/************************************************************************/

.section .data
        temp: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	 code_pt: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	pt_state: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	  temp_2: .byte 0x00, 0x00
	  temp_4: .byte 0x00, 0x00
	 temp_16: .byte 0x00, 0x00
	  temp_n: .byte 0x00, 0x00
			
.section .text
.syntax unified
.thumb
.global AES_Encrypt_GCB
	
		/*====================================*/
		/*-----function: AES_Encrypt_GCB------*/
		/*====================================*/
AES_Encrypt_GCB:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the adress of result */
		
		/* fix some Regs */
		MOV R12, R0
		// randomSeed stores all the random numbers utlized in AES
		LDR R11, =randomSeed 
		
		/* encode the plaintext */
		MOV R10, #0xFF
		LDR R5, =pt_aes // plaintext
		LDR R4, =temp // container for the encoded plaintext and key
		// encode
		LDR R0, [R5]
		LDR R1, [R11]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R1, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R1, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #2]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R1, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #4]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R1, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #6]
		LDR R0, [R5, #4]
		LDR R1, [R11, #4]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R1, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #8]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R1, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #10]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R1, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #12]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R1, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #14]
		LDR R0, [R5, #8]
		LDR R1, [R11, #8]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R1, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #16]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R1, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #18]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R1, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #20]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R1, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #22]
		LDR R0, [R5, #12]
		LDR R1, [R11, #12]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R1, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #24]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R1, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #26]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R1, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #28]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R1, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #30]
		ADD R11, #16 // move the pointer of randomSeed
		
		/* add encoded key to temp */
		LDR R8, =maskedKeys // note that the key has been extended and encoded
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		ADD R8, #32 // move the pointer of maskedKeys
		
		LDR R10, =random1 // container for R1 
		LDR R9, =random2 // container for R2 * 16
		LDR R7, =code_pt 
		LDR R6, =pt_state
		
		/*--add Round Key--*/
		// R2 construction
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// add round key
		MOV R0, R4 // address of temp
		MOV R1, R6 // address of pt_state
		BL Gadget_CodeLs_AddRoundKey
		
		/*----AES Rounds----*/
		MOV R5, #0 // counter for rounds
loopRounds:
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2 // move the address of pt_state
		ADD R7, #2 // move the address of code_pt
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		// R2 construction
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// shiftRows
		MOV R0, R7
		MOV R1, R6
		BL Gadget_CodeLs_ShiftRows
		
		/*--MixColumns--*/
		// R2 construction
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// mix columns
		MOV R0, R6
		MOV R1, R4
		BL Gadget_CodeLs_MixColumns
		
		/* add the encoded key of next round to temp */
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		ADD R8, #32 // move the pointer of maskedKeys
		
		/*--AddRoundKey--*/
		// R2 construction
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// add round key
		MOV R0, R4
		MOV R1, R6
		BL Gadget_CodeLs_AddRoundKey
		
		ADD R5, #1
		CMP R5, #9
		IT NE 
		BNE loopRounds

		/*----the last round----*/
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		LDR R6, =pt_state // reset pt_state
		LDR R7, =code_pt // reset code_pt
		
		/*--ShiftRows--*/
		// R2 construction 
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// shiftRows
		MOV R0, R7
		MOV R1, R4
		BL Gadget_CodeLs_ShiftRows
		
		/* add key to temp */
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		
		/* add the encoded key of the last round to temp */
		// R2 construction
		// the 0th R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// the 1th R2
		LDRB R0, [R11, #2]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #4]
		LDRB R0, [R11, #3]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #6]
		// the 2th R2
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #8]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #10]
		// the 3th R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #12]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #14]
		// the 4th R2
		LDRB R0, [R11, #8]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #16]
		LDRB R0, [R11, #9]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #18]
		// the 5th R2
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #20]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #22]
		// the 6th R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #24]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #26]
		// the 7th R2
		LDRB R0, [R11, #14]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #28]
		LDRB R0, [R11, #15]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #30]
		// the 8th R2
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #32]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #34]
		// the 9th R2
		LDRB R0, [R11, #18]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #36]
		LDRB R0, [R11, #19]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #38]
		// the 10th R2
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #40]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #42]
		// the 11th R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #44]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #46]
		// the 12th R2
		LDRB R0, [R11, #24]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #48]
		LDRB R0, [R11, #25]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #50]
		// the 13th R2
		LDRB R0, [R11, #26]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #52]
		LDRB R0, [R11, #27]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #54]
		// the 14th R2
		LDRB R0, [R11, #28]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #56]
		LDRB R0, [R11, #29]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #58]
		// the 15th R2
		LDRB R0, [R11, #30]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #60]
		LDRB R0, [R11, #31]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #62]
		ADD R11, #32 // move the pointer of randomSeed
		// add round key
		MOV R0, R4
		MOV R1, R7
		BL Gadget_CodeLs_AddRoundKey
		
		
		/* decode */
		LDR R0, [R7] // r7 (code_pt) stores the address of final encoded cyphertext
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12] // r12 stores the address of ct_aes
		LSR R0, R0, #16
		STRB R0, [R12, #1]
		LDR R0, [R7, #4]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #2]
		LSR R0, R0, #16
		STRB R0, [R12, #3]
		LDR R0, [R7, #8]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #4]
		LSR R0, R0, #16
		STRB R0, [R12, #5]
		LDR R0, [R7, #12]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #6]
		LSR R0, R0, #16
		STRB R0, [R12, #7]
		LDR R0, [R7, #16]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #8]
		LSR R0, R0, #16
		STRB R0, [R12, #9]
		LDR R0, [R7, #20]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #10]
		LSR R0, R0, #16
		STRB R0, [R12, #11]
		LDR R0, [R7, #24]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #12]
		LSR R0, R0, #16
		STRB R0, [R12, #13]
		LDR R0, [R7, #28]
		EOR R0, R0, R0, LSR #8
		STRB R0, [R12, #14]
		LSR R0, R0, #16
		STRB R0, [R12, #15]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-----------function: Sbox-----------*/
		/*====================================*/
Sbox:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R8,LR}
		/* r6 stores the address of src */
		/* r7 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r10 stores the address of random1 */
		/* r11 stores the address of randomSeed */
		
		/* fix some Regs */
		LDR R8, =temp_2
		LDR R5, =temp_4
		LDR R4, =temp_16
		LDR R3, =temp_n
		
		// construction R2
		LDRB R0, [R11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #1]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// x^2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, #0 // parameter indicating .^2
		BL Gadget_CodeL
		
		// construction R1 & R2
		LDRH R0, [R11, #2]
		ORR R0, R0, R0, LSL #16
		STR R0, [R10]
		LDRB R0, [R11, #4]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #5]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// tempn = x * x ^ 2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// construction R2
		LDRB R0, [R11, #6]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #7]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// tempn^4
		MOV R0, R3 // temp_n
		MOV R1, R5 // temp_4
		MOV R2, #1 // parameter indicating .^4
		BL Gadget_CodeL
		
		// construction R1 & R2
		LDRH R0, [R11, #8]
		ORR R0, R0, R0, LSL #16
		STR R0, [R10]
		LDRB R0, [R11, #10]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #11]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// dst = (temp4) * (tempn)
		MOV R0, R5 // temp_4
		MOV R1, R3 // temp_n
		MOV R2, R7 // dst
		BL Gadget_CodeMuL
		
		// construction R2
		LDRB R0, [R11, #12]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #13]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// dst^16
		MOV R0, R7 // dst
		MOV R1, R4 // temp_16
		MOV R2, #2 // parameter indicating .^16
		BL Gadget_CodeL
		
		// construction R1 & R2
		LDRH R0, [R11, #14]
		ORR R0, R0, R0, LSL #16
		STR R0, [R10]
		LDRB R0, [R11, #16]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #17]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// dst = (temp4) * (temp16)
		MOV R0, R4 // temp_16
		MOV R1, R5 // temp_4
		MOV R2, R6 // src
		BL Gadget_CodeMuL
		
		// construction R1 & R2
		LDRH R0, [R11, #18]
		ORR R0, R0, R0, LSL #16
		STR R0, [R10]
		LDRB R0, [R11, #20]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #21]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// tempn = src * (temp2)
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// construction R2
		LDRB R0, [R11, #22]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9]
		LDRB R0, [R11, #23]
		ORR R0, R0, R0, LSL #8
		STRH R0, [R9, #2]
		// dst = aff(tempn)
		MOV R0, R3 // temp_n
		MOV R1, R7 // dst
		MOV R2, #3 // parameter indicating affine transformation
		BL Gadget_CodeL
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R8,LR}
		/* return */
		BX LR
		/*====================================*/
		

.end
	