/************************************************************************/
/* @file   comm.c                                                       */
/* @author yangx                                                        */
/* @date   2022-4-12                                                    */
/* @brief  functions for Usart in STM32F407                             */
/************************************************************************/

#include <stm32f4xx_rcc.h>
#include "comm.h"

/************************************************************************/
/*                          Private Function                            */
/************************************************************************/

void NVIC_Conf()
{
	NVIC_InitTypeDef NVIC_InitStructure;

	/* Enable the USARTx Interrupt */
	NVIC_InitStructure.NVIC_IRQChannel = USART2_IRQn;         // configure the USART2 interrupts
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0; // set the priority group of the USART1 interrupts 
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;        // set the subpriority inside the group
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

}

void USART_Conf(uint32_t baudrate)
{
	USART_InitTypeDef USART_InitStructure;

	/* USARTx configuration ------------------------------------------------------*/
	 /* USARTx configured as follows:
		- BaudRate = 9600 baud
		- Word Length = 8 Bits
		- One Stop Bit
		- No parity
		- Hardware flow control disabled (RTS and CTS signals)
		- Receive and transmit enabled
	*/
	USART_InitStructure.USART_BaudRate = baudrate;
	USART_InitStructure.USART_WordLength = USART_WordLength_8b;
	USART_InitStructure.USART_StopBits = USART_StopBits_1;
	USART_InitStructure.USART_Parity = USART_Parity_No;
	USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;

	USART_Init(USART2, &USART_InitStructure);
}

void GPIO_Conf()
{
	GPIO_InitTypeDef GPIO_InitStructure;

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2 | GPIO_Pin_3;  // Pins 2 (TX) and 3 (RX) are used
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF; 			// the pins are configured as alternate function so the USART peripheral has access to them
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;		// this defines the IO speed and has nothing to do with the baudrate!
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;			// this defines the output type as push pull mode (as opposed to open drain)
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;

	GPIO_Init(GPIOA, &GPIO_InitStructure);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource2, GPIO_AF_USART2);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource3, GPIO_AF_USART2);
}

/************************************************************************/
/*                           Public Function                            */
/************************************************************************/

void Usart_Init(uint32_t baudrate)
{
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA, ENABLE);

	// NVIC Configuration
	NVIC_Conf();

	// USART Configuration
	USART_Conf(baudrate);

	// Disable the USART2 Receive interrupt
	USART_ITConfig(USART2, USART_IT_RXNE, DISABLE);

	// GPIO Configuration
	GPIO_Conf();

	// Enable the complete USART2 peripheral
	USART_Cmd(USART2, ENABLE);
}

void Usart_Send(volatile char* data, int len)
{

	while (len) {
		// wait until data register is empty
		//while( !(USART2->SR & 0x00000040) );
		while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
		USART_SendData(USART2, *data);
		data++;
		len--;
	}
	// wait until transmit completes
	while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET) {};

}

unsigned char Usart_GetByte()
{
	while (USART_GetFlagStatus(USART2, USART_FLAG_RXNE) == RESET) {}
	return (unsigned char)USART_ReceiveData(USART2);
}

// this is the interrupt request handler (IRQ) for ALL USART1 interrupts
void USART1_IRQHandler(void){

	// check if the USART1 receive interrupt flag was set
	if( USART_GetITStatus(USART2, USART_IT_RXNE) ){

		USART_ClearITPendingBit(USART2, USART_IT_RXNE); // clear interrupt flag
		
	}
}




