/***************************************************************************************/
/* @file   gadget_CodeLs_MixColumns.s                                                  */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1, 0;1, 1] and the f function is fixed as         */
/*		   MixColumns of AES-128.                                                      */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_MixColumns
	
Gadget_CodeLs_MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */

		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		MOV R11, R0
		MOV R10, R1
		MOV R9, #0x00FF00FF
		MOV R8, #0xFF
		
		/*-------the 1/4.1 part--------*/
		// Part A
		// call part A 16 / 4 times 
		// for [0, 1, 2, 3]
		LDR R3, [R11]
		LDR R4, [R11, #4]
		// mix columns
		AND R5, R3, #0x00FF00FF
		AND R6, R4, #0x00FF00FF
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		MOV R7, R1
		AND R5, R9, R3, LSR #8
		AND R6, R9, R4, LSR #8
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		// multiply by A
		AND R0, R7, #0xFF
		AND R2, R1, #0xFF
		ORR R3, R0, R2, LSL #16
		AND R0, R8, R7, LSR #8
		AND R2, R8, R1, LSR #8
		ORR R4, R0, R2, LSL #16
		AND R0, R8, R7, LSR #16
		AND R2, R8, R1, LSR #16
		ORR R5, R0, R2, LSL #16
		AND R0, R8, R7, LSR #24
		AND R2, R8, R1, LSR #24
		ORR R6, R0, R2, LSL #16
		// refresh with R2
		LDR R1, [R12]
		EOR R3, R1
		LDR R1, [R12, #4]
		EOR R4, R1
		LDR R1, [R12, #8]
		EOR R5, R1
		LDR R1, [R12, #12]
		EOR R6, R1
		// eor and store
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R10]
		STRH R4, [R10, #2]
		STRH R5, [R10, #4]
		STRH R6, [R10, #6]
		
		/*-------the 1/4.2 part--------*/
		// Part A
		// call part A 16 / 4 times 
		// for [4, 5, 6, 7]
		LDR R3, [R11, #8]
		LDR R4, [R11, #12]
		// mix columns
		AND R5, R3, #0x00FF00FF
		AND R6, R4, #0x00FF00FF
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		MOV R7, R1
		AND R5, R9, R3, LSR #8
		AND R6, R9, R4, LSR #8
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		// multiply by A
		AND R0, R7, #0xFF
		AND R2, R1, #0xFF
		ORR R3, R0, R2, LSL #16
		AND R0, R8, R7, LSR #8
		AND R2, R8, R1, LSR #8
		ORR R4, R0, R2, LSL #16
		AND R0, R8, R7, LSR #16
		AND R2, R8, R1, LSR #16
		ORR R5, R0, R2, LSL #16
		AND R0, R8, R7, LSR #24
		AND R2, R8, R1, LSR #24
		ORR R6, R0, R2, LSL #16
		// refresh with R2
		LDR R1, [R12, #16]
		EOR R3, R1
		LDR R1, [R12, #20]
		EOR R4, R1
		LDR R1, [R12, #24]
		EOR R5, R1
		LDR R1, [R12, #28]
		EOR R6, R1
		// eor and store
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R10, #8]
		STRH R4, [R10, #10]
		STRH R5, [R10, #12]
		STRH R6, [R10, #14]
		
		/*-------the 1/4.3 part--------*/
		// Part A
		// call part A 16 / 4 times 
		// for [8, 9, 10, 11]
		LDR R3, [R11, #16]
		LDR R4, [R11, #20]
		// mix columns
		AND R5, R3, #0x00FF00FF
		AND R6, R4, #0x00FF00FF
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		MOV R7, R1
		AND R5, R9, R3, LSR #8
		AND R6, R9, R4, LSR #8
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		// multiply by A
		AND R0, R7, #0xFF
		AND R2, R1, #0xFF
		ORR R3, R0, R2, LSL #16
		AND R0, R8, R7, LSR #8
		AND R2, R8, R1, LSR #8
		ORR R4, R0, R2, LSL #16
		AND R0, R8, R7, LSR #16
		AND R2, R8, R1, LSR #16
		ORR R5, R0, R2, LSL #16
		AND R0, R8, R7, LSR #24
		AND R2, R8, R1, LSR #24
		ORR R6, R0, R2, LSL #16
		// refresh with R2
		LDR R1, [R12, #32]
		EOR R3, R1
		LDR R1, [R12, #36]
		EOR R4, R1
		LDR R1, [R12, #40]
		EOR R5, R1
		LDR R1, [R12, #44]
		EOR R6, R1
		// eor and store
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R10, #16]
		STRH R4, [R10, #18]
		STRH R5, [R10, #20]
		STRH R6, [R10, #22]
		
		/*-------the 1/4.4 part--------*/
		// Part A
		// call part A 16 / 4 times 
		// for [12, 13, 14, 15]
		LDR R3, [R11, #24]
		LDR R4, [R11, #28]
		// mix columns
		AND R5, R3, #0x00FF00FF
		AND R6, R4, #0x00FF00FF
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		MOV R7, R1
		AND R5, R9, R3, LSR #8
		AND R6, R9, R4, LSR #8
		ORR R5, R5, R5, LSR #8
		ORR R6, R6, R6, LSR #8
		LSL R5, R5, #16
		LSR R5, R5, #16
		ORR R0, R5, R6, LSL #16
		BL mul_mat_vec_4
		// multiply by A
		AND R0, R7, #0xFF
		AND R2, R1, #0xFF
		ORR R3, R0, R2, LSL #16
		AND R0, R8, R7, LSR #8
		AND R2, R8, R1, LSR #8
		ORR R4, R0, R2, LSL #16
		AND R0, R8, R7, LSR #16
		AND R2, R8, R1, LSR #16
		ORR R5, R0, R2, LSL #16
		AND R0, R8, R7, LSR #24
		AND R2, R8, R1, LSR #24
		ORR R6, R0, R2, LSL #16
		// refresh with R2
		LDR R1, [R12, #48]
		EOR R3, R1
		LDR R1, [R12, #52]
		EOR R4, R1
		LDR R1, [R12, #56]
		EOR R5, R1
		LDR R1, [R12, #60]
		EOR R6, R1
		// eor and store
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R10, #24]
		STRH R4, [R10, #26]
		STRH R5, [R10, #28]
		STRH R6, [R10, #30]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	