/***************************************************************************************/
/* @file   gadget_CodeLs_ShiftRows.s                                                   */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1, 0;1, 1] and the f function is fixed as         */
/*		   shiftRows of AES-128.                                                       */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_ShiftRows
	
Gadget_CodeLs_ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the adress of src */
		/* r1 stores the adress of dst */

		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		
		/*-------the 1/4.1 part--------*/
		// Part A
		// call part A 16 / 4 times
		// for [0, 1, 2, 3]
		LDRH R3, [R0]
		LDRH R4, [R0, #2]
		LDRH R5, [R0, #4]
		LDRH R6, [R0, #6]
		// shift rows
		// multiply by A
		AND R7, R3, #0xFF
		AND R8, R4, #0xFF
		AND R9, R5, #0xFF
		AND R10, R6, #0xFF
		LSR R3, R3, #8
		LSR R4, R4, #8
		LSR R5, R5, #8
		LSR R6, R6, #8
		// refresh with R2 
		// changed to [0, 13, 10, 7]
		ORR R3, R7, R3, LSL #16
		ORR R4, R8, R4, LSL #16
		ORR R5, R9, R5, LSL #16
		ORR R6, R10, R6, LSL #16
		LDR R11, [R12]
		EOR R3, R11
		LDR R11, [R12, #52]
		EOR R4, R11
		LDR R11, [R12, #40]
		EOR R5, R11
		LDR R11, [R12, #28]
		EOR R6, R11
		// eor and store 
		// store in [0, 13, 10, 7]
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R1]
		STRH R4, [R1, #26]
		STRH R5, [R1, #20]
		STRH R6, [R1, #14]
		
		/*-------the 1/4.2 part--------*/
		// Part A
		// call part A 16 / 4 times
		// for [4, 5, 6, 7]
		LDRH R3, [R0, #8]
		LDRH R4, [R0, #10]
		LDRH R5, [R0, #12]
		LDRH R6, [R0, #14]
		// shift rows
		// multiply by A
		AND R7, R3, #0xFF
		AND R8, R4, #0xFF
		AND R9, R5, #0xFF
		AND R10, R6, #0xFF
		LSR R3, R3, #8
		LSR R4, R4, #8
		LSR R5, R5, #8
		LSR R6, R6, #8
		// refresh with R2 
		// changed to [4, 1, 14, 11]
		ORR R3, R7, R3, LSL #16
		ORR R4, R8, R4, LSL #16
		ORR R5, R9, R5, LSL #16
		ORR R6, R10, R6, LSL #16
		LDR R11, [R12, #16]
		EOR R3, R11
		LDR R11, [R12, #4]
		EOR R4, R11
		LDR R11, [R12, #56]
		EOR R5, R11
		LDR R11, [R12, #44]
		EOR R6, R11
		// eor and store 
		// store in [4, 1, 14, 11]
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R1, #8]
		STRH R4, [R1, #2]
		STRH R5, [R1, #28]
		STRH R6, [R1, #22]
		
		/*-------the 1/4.1 part--------*/
		// Part A
		// call part A 16 / 4 times
		// for [8, 9, 10, 11]
		LDRH R3, [R0, #16]
		LDRH R4, [R0, #18]
		LDRH R5, [R0, #20]
		LDRH R6, [R0, #22]
		// shift rows
		// multiply by A
		AND R7, R3, #0xFF
		AND R8, R4, #0xFF
		AND R9, R5, #0xFF
		AND R10, R6, #0xFF
		LSR R3, R3, #8
		LSR R4, R4, #8
		LSR R5, R5, #8
		LSR R6, R6, #8
		// refresh with R2 
		// change to [8, 5, 2, 15]
		ORR R3, R7, R3, LSL #16
		ORR R4, R8, R4, LSL #16
		ORR R5, R9, R5, LSL #16
		ORR R6, R10, R6, LSL #16
		LDR R11, [R12, #32]
		EOR R3, R11
		LDR R11, [R12, #20]
		EOR R4, R11
		LDR R11, [R12, #8]
		EOR R5, R11
		LDR R11, [R12, #60]
		EOR R6, R11
		// eor and store 
		// store in [8, 5, 2, 15]
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R1, #16]
		STRH R4, [R1, #10]
		STRH R5, [R1, #4]
		STRH R6, [R1, #30]
		
		/*-------the 1/4.1 part--------*/
		// Part A
		// call part A 16 / 4 times
		// for [12, 13, 14, 15]
		LDRH R3, [R0, #24]
		LDRH R4, [R0, #26]
		LDRH R5, [R0, #28]
		LDRH R6, [R0, #30]
		// shift rows
		// multiply by A
		AND R7, R3, #0xFF
		AND R8, R4, #0xFF
		AND R9, R5, #0xFF
		AND R10, R6, #0xFF
		LSR R3, R3, #8
		LSR R4, R4, #8
		LSR R5, R5, #8
		LSR R6, R6, #8
		// refresh with R2 
		// change to [12, 9, 6, 3]
		ORR R3, R7, R3, LSL #16
		ORR R4, R8, R4, LSL #16
		ORR R5, R9, R5, LSL #16
		ORR R6, R10, R6, LSL #16
		LDR R11, [R12, #48]
		EOR R3, R11
		LDR R11, [R12, #36]
		EOR R4, R11
		LDR R11, [R12, #24]
		EOR R5, R11
		LDR R11, [R12, #12]
		EOR R6, R11
		// eor and store 
		// store in [12, 9, 6, 3]
		EOR R3, R3, R3, LSR #16
		EOR R4, R4, R4, LSR #16
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		STRH R3, [R1, #24]
		STRH R4, [R1, #18]
		STRH R5, [R1, #12]
		STRH R6, [R1, #6]
		
		/* copy registers value from the stack */
		POP {R0,R1,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	