/****************************************************************************/
/* @file   gadget_CodeMuL.s                                                 */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of gadget CodeMuL proposed in XXXXXXXXXXXX.       */
/*         This function calculates the byte_wise multiplication of two     */
/*		   vectors over F_{2^8}^{2} in the masked domain. And the generator */
/*         matrix is fixed as [1,0;1,1].                                    */
/****************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeMuL
	
Gadget_CodeMuL:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R9,R10,R11,R12,LR}
		/* r0 stores the address of src1 */
		/* r1 stores the address of src2 */
		/* r2 stores the address of result */
		
		/* fix some Regs */
		LDR R12, =random1 // store R1
		LDR R11, =random2 // store R2 
		LDR R3, =bhtable // for field multiplication
		LDR R4, =bltable
		LDRH R0, [R0] // r0 stores the vec1
		LDRH R1, [R1] // r1 stores the vec2
		MOV R9, #0xFF
		
		/*-------The 0th row of matrix--------*/
		/*--outer product--*/
		// prepare the byte
		AND R5, R9, R0
		LSL R5, R5, #4
		// the first byte of vec (1, 2)
		AND R6, R9, R1
		ORR R7, R5, R6, LSR #4
		LDRB R7, [R3, R7]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R10, R7, R6
		// the second byte of vec (1, 2)
		AND R6, R9, R1, LSR #8
		ORR R7, R5, R6, LSR #4
		LDRB R7, [R3, R7]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R7, R6
		ORR R10, R10, R7, LSL #8
		// r10 stores the vec : (1, 2)
		
		/*--refresh with R1--*/
		LDRH R5, [R12]
		EOR R10, R5
		// r10 stores the vec : (1, 2)
		
		/*--multiply by M0--*/
		// equals byte xor byte
		EOR R10, R10, R10, LSR #8
		// r10 stores the vec : (1, 1)
		
		/*--multiply by A--*/
		AND R10, #0xFF
		// r10 stores the vec : (1, 2)
		
		/*--refresh with R2--*/
		LDRH R5, [R11]
		EOR R10, R5
		// r10 stores the vec : (1, 2)
		
		
		/*-------The 1th row of matrix--------*/
		/*--outer product--*/
		// prepare the byte
		AND R5, R9, R0, LSR #8
		LSL R5, R5, #4
		// the first byte of vec (1, 2)
		AND R6, R9, R1
		ORR R7, R5, R6, LSR #4
		LDRB R7, [R3, R7]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R7, R6
		// the second byte of vec (1, 2)
		AND R6, R9, R1, LSR #8
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R9, R6
		ORR R9, R7, R9, LSL #8
		// r9 stores the vec : (1, 2)
		
		/*--refresh with R1--*/
		LDRH R5, [R12, #2]
		EOR R9, R5
		// r8 stores the vec : (1, 2)
		
		/*--multiply by M1--*/
		// equals byte xor byte
		EOR R9, R9, R9, LSR #8
		// r8 stores the vec : (1, 1)
		
		/*--multiply by A--*/
		AND R9, #0xFF
		// r8 stores the vec : (1, 2)
		
		/*--refresh with R2--*/
		LDRH R5, [R11, #2]
		EOR R9, R5
		// r9 stores the vec : (1, 2)
		
		/*-------Eor and store--------*/
		EOR R10, R9
		STRH R10, [R2]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	