/****************************************************************************/
/* @file   mul_mat_vec_4.s                                                  */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of the product for a matrix defined as            */
/*         [2,3,1,1;1,2,3,1;1,1,2,3;3,1,1,2] and a vector over F_{2^8}^{4}. */
/*		   Actually this is the MixColumns operated on a column. The input  */
/*		   is the vector stored in the 32-bit registers R0.                 */
/****************************************************************************/

.section .data  
.section .text
.syntax unified
.thumb
.global mul_mat_vec_4

/* input */
/* r0 stores the vec */
/* r1 stores the result */

/* output */
mul_mat_vec_4:
		/* copy registers value to the stack */
		PUSH {R0,R2,R3,R4,R5,R6,R8,R9,R10,R11,R12}
		/* r0 stores the vec */
        /* r1 stores the result */
		
		/* fix some Regs */
		LDR R2, =tab_2
		LDR R3, =tab_3
		MOV R8, #0xFF
		
		// r9 stores the inner product of the first row of matrix and the vector 
		// r10 stores the inner product of the second row of matrix and the vector 
		// r11 stores the inner product of the third row of matrix and the vector 
		// r12 stores the inner product of the forth row of matrix and the vector 
		
		/* deal with the 0th byte of vec */
		AND R10, R0, R8
		LDRB R9, [R2, R10]
		LDRB R12, [R3, R10]
		MOV R11, R10
		
		/* deal with the 1th byte of vec */
		AND R5, R8, R0, LSR #8
		EOR R11, R5
		EOR R12, R5
		LDRB R6, [R3, R5]
		EOR R9, R6
		LDRB R5, [R2, R5]
		EOR R10, R5
		
		/* deal with the 2th byte of vec */
		AND R5, R8, R0, LSR #16
		EOR R9, R5
		EOR R12, R5
		LDRB R6, [R3, R5]
		EOR R10, R6
		LDRB R5, [R2, R5]
		EOR R11, R5
		
		/* deal with the 3th byte of vec */
		AND R5, R8, R0, LSR #24
		EOR R9, R5
		EOR R10, R5
		LDRB R6, [R3, R5]
		EOR R11, R6
		LDRB R5, [R2, R5]
		EOR R12, R5
		
		/* store the result to R1 */
		MOV R1, R9
		ORR R1, R1, R10, LSL #8
		ORR R1, R1, R11, LSL #16
		ORR R1, R1, R12, LSL #24

		/* copy registers value from the stack */
		POP {R0,R2,R3,R4,R5,R6,R8,R9,R10,R11,R12}
		/* return */
		BX LR
.end
	