/**************************************************************************/
/* @file   aes.s                                                          */
/* @author yangx                                                          */
/* @date   2022-4-11                                                      */
/* @brief  eight functions                                                */
/*         -- AES_Encrypt_GCB [public]                                    */
/*            Implementation of AES-128 masked by code-based masking,     */ 
/*            please refer to xxxxxxx to see the detailed masking         */
/*            scheme. The generator matrix is fixed as                    */
/*            [1,0;91,1]. This function has no output and the input is    */
/*            the address of cyphertext.                                  */ 
/*         -- AddRoundKey [private]                                       */
/*            Implementation of AddRoundKey masked by code-based masking. */
/*         -- Sbox [private]                                              */
/*            Implementation of SubBytes masked by code-based masking.    */
/*            Note that S-box can be implemented by an inverse and an     */
/*            affine transformation.                                      */
/*         -- ShiftRows [private]                                         */
/*            Implementation of ShiftRows masked by code-based masking.   */
/*         -- MixColumns [private]                                        */
/*            Implementation of MixColumns masked by code-based masking.  */
/*         -- three functions for constructing R2                         */
/*            --- GetR21 for addRoundKey                                  */
/*            --- GetR22 for shiftRows                                    */
/*            --- GetR23 for mixColumns                                   */
/**************************************************************************/

.section .data
        temp: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	 code_pt: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	pt_state: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	  temp_2: .byte 0x00, 0x00
	  temp_4: .byte 0x00, 0x00
	 temp_16: .byte 0x00, 0x00
	  temp_n: .byte 0x00, 0x00
			
.section .text
.syntax unified
.thumb
.global AES_Encrypt_GCB, Sbox
	
		/*====================================*/
		/*-----function: AES_Encrypt_GCB------*/
		/*====================================*/
AES_Encrypt_GCB:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of result */
		
		/* fix some Regs */
		MOV R12, R0
		// randomSeed stores all the random numbers utlized in AES
		// note that the first 16 byte numbers are the product of random numbers
		// stored in rawRand and 91.
		LDR R11, =randomSeed
		
		/* encode the plaintext */
		MOV R10, #0xFF
		LDR R5, =pt_aes // plaintext
		LDR R4, =temp // container for the encoded plaintext and key
		LDR R6, =rawRand // random numbers for encoding
		// encode
		LDR R0, [R5]
		LDR R1, [R11]
		LDR R7, [R6]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R7, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R7, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #2]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R7, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #4]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R7, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #6]
		LDR R0, [R5, #4]
		LDR R1, [R11, #4]
		LDR R7, [R6, #4]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R7, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #8]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R7, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #10]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R7, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #12]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R7, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #14]
		LDR R0, [R5, #8]
		LDR R1, [R11, #8]
		LDR R7, [R6, #8]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R7, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #16]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R7, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #18]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R7, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #20]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R7, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #22]
		LDR R0, [R5, #12]
		LDR R1, [R11, #12]
		LDR R7, [R6, #12]
		EOR R0, R1
		AND R2, R0, R10
		AND R3, R7, R10
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #24]
		AND R2, R10, R0, LSR #8
		AND R3, R10, R7, LSR #8
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #26]
		AND R2, R10, R0, LSR #16
		AND R3, R10, R7, LSR #16
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #28]
		AND R2, R10, R0, LSR #24
		AND R3, R10, R7, LSR #24
		ORR R2, R2, R3, LSL #8
		STRH R2, [R4, #30]
		ADD R11, #16 // move the pointer of randomSeed
		
		/* add encoded key to temp */
		LDR R8, =maskedKeys // note that the key has been extended and encoded
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		ADD R8, #32 // move the pointer of maskedKeys
		
		LDR R10, =random1 // container for R1 
		LDR R9, =random2 // container for R2 * 16
		LDR R7, =code_pt
		LDR R6, =pt_state
		
		/*--add Round Key--*/
		BL AddRoundKey
		ADD R11, #32 // move the pointer of randomSeed
		
		/*----AES Round----*/
		MOV R5, #0
loopRounds:
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2 // move the address of pt_state
		ADD R7, #2 // move the address of code_pt
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		BL ShiftRows
		ADD R11, #32 // move the pointer of randomSeed
		
		/*--MixColumns--*/
		BL MixColumns
		ADD R11, #32 // move the pointer of randomSeed
		
		/* add the encoded key of next round to temp */
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		ADD R8, #32 // move the pointer of maskedKeys
		
		/*--AddRoundKey--*/
		BL AddRoundKey
		ADD R11, #32 // move the pointer of randomSeed
		
		ADD R5, #1
		CMP R5, #9
		IT NE 
		BNE loopRounds

		/*----the last round----*/
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		ADD R6, #2
		ADD R7, #2
		BL Sbox
		ADD R11, #24 // move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		MOV R6, R4
		BL ShiftRows
		ADD R11, #32 // move the pointer of randomSeed
		
		/* add the encoded key of the last round to temp */
		LDR R0, [R8]
		STR R0, [R4, #32]
		LDR R0, [R8, #4]
		STR R0, [R4, #36]
		LDR R0, [R8, #8]
		STR R0, [R4, #40]
		LDR R0, [R8, #12]
		STR R0, [R4, #44]
		LDR R0, [R8, #16]
		STR R0, [R4, #48]
		LDR R0, [R8, #20]
		STR R0, [R4, #52]
		LDR R0, [R8, #24]
		STR R0, [R4, #56]
		LDR R0, [R8, #28]
		STR R0, [R4, #60]
		
		/*--AddRoundKey--*/
		LDR R6, =pt_state
		BL AddRoundKey
		
		/* decode */
		LDR R3, =tabGByte // for field multiplication with 91
		
		LDRH R0, [R6] // r6 (code_pt) stores the address of final encoded cyphertext
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12] // r12 stores the address of ct_aes
		LDRH R0, [R6, #2]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #1]
		LDRH R0, [R6, #4]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #2]
		LDRH R0, [R6, #6]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #3]
		LDRH R0, [R6, #8]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #4]
		LDRH R0, [R6, #10]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #5]
		LDRH R0, [R6, #12]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #6]
		LDRH R0, [R6, #14]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #7]
		LDRH R0, [R6, #16]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #8]
		LDRH R0, [R6, #18]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #9]
		LDRH R0, [R6, #20]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #10]
		LDRH R0, [R6, #22]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #11]
		LDRH R0, [R6, #24]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #12]
		LDRH R0, [R6, #26]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #13]
		LDRH R0, [R6, #28]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #14]
		LDRH R0, [R6, #30]
		LSR R1, R0, #8
		LDRB R1, [R3, R1]
		EOR R0, R1
		STRB R0, [R12, #15]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-------function: AddRoundKey--------*/
		/*====================================*/
AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R5,R7,R12,LR}
		/* r4 stores the address of src */
		/* r5 stores the pointer of R9 */
		/* r6 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* fix some Regs */
		LDR R7, =tabGByte // for field multiplication with 91
		
		MOV R5, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		// R2 construction 
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		
		// add round key
		MOV R0, R4
		MOV R1, R6
		BL Gadget_CodeLs_AddRoundKey
	
		
		/* copy registers value from the stack */
		POP {R0,R1,R5,R7,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-----------function: Sbox-----------*/
		/*====================================*/
Sbox:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* r6 stores the address of src */
		/* r7 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r10 stores the address of random1 */
		/* r11 stores the address of randomSeed */
		
		/* fix some Regs */
		LDR R8, =temp_2
		LDR R3, =temp_n
		LDR R4, =temp_4
		LDR R5, =temp_16
		LDR R12, =tabGByte // for field multiplication with 91
		
		// R2 construction
		LDRB R0, [R11]
		LDRB R1, [R11, #1]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// x^2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, #0 // parameter indicating .^2
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #2]
		AND R1, R0, #0xFF
		LDRB R1, [R12, R1]
		LSR R2, R0, #8
		LDRB R2, [R12, R2]
		ORR R1, R1, R2, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R10]
		// R2 construction
		LDRB R0, [R11, #4]
		LDRB R1, [R11, #5]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// tempn = x * x ^ 2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRB R0, [R11, #6]
		LDRB R1, [R11, #7]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// tempn^4
		MOV R0, R3 // temp_n
		MOV R1, R4 // temp_4
		MOV R2, #1 // parameter indicating .^4
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #8]
		AND R1, R0, #0xFF
		LDRB R1, [R12, R1]
		LSR R2, R0, #8
		LDRB R2, [R12, R2]
		ORR R1, R1, R2, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R10]
		// R2 construction
		LDRB R0, [R11, #10]
		LDRB R1, [R11, #11]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// dst = (temp4) * (tempn)
		MOV R0, R4 // temp_4
		MOV R1, R3 // temp_n
		MOV R2, R7 // dst
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRB R0, [R11, #12]
		LDRB R1, [R11, #13]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// dst^16
		MOV R0, R7 // dst
		MOV R1, R5 // temp_16
		MOV R2, #2 // parameter indicating .^16
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #14]
		AND R1, R0, #0xFF
		LDRB R1, [R12, R1]
		LSR R2, R0, #8
		LDRB R2, [R12, R2]
		ORR R1, R1, R2, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R10]
		// R2 construction
		LDRB R0, [R11, #16]
		LDRB R1, [R11, #17]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// dst = (temp4) * (temp16)
		MOV R0, R5 // temp_16
		MOV R1, R4 // temp_4
		MOV R2, R6 // src
		BL Gadget_CodeMuL
		
		// R1 construction
		LDRH R0, [R11, #18]
		AND R1, R0, #0xFF
		LDRB R1, [R12, R1]
		LSR R2, R0, #8
		LDRB R2, [R12, R2]
		ORR R1, R1, R2, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R10]
		// R2 construction
		LDRB R0, [R11, #20]
		LDRB R1, [R11, #21]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// tempn = src * (temp2)
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRB R0, [R11, #22]
		LDRB R1, [R11, #23]
		LDRB R2, [R12, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R12, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R0, R1, LSL #16
		STR R0, [R9]
		// dst = aff(n)
		MOV R0, R3 // temp_n
		MOV R1, R7 // dst
		MOV R2, #3 // parameter indicating affine transformation
		BL Gadget_CodeL
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: ShiftRows---------*/
		/*====================================*/
ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R4,R5,R12,LR}
		/* r6 stores the address of dst */
		/* r7 stores the address of src */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		
		/* fix some Regs */
		LDR R4, =tabGByte // for field multiplication with 91
		
		MOV R5, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		
		// shift rows
		MOV R0, R7
		MOV R1, R6
		BL Gadget_CodeLs_ShiftRows
	
		/* copy registers value from the stack */
		POP {R0,R1,R4,R5,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: MixColumns--------*/
		/*====================================*/
MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R5,R7,R12,LR}
		/* r4 stores the address of dst */
		/* r5 stores the pointer of R9 */
		/* r6 stores the address of src */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* fix some Regs */
		LDR R7, =tabGByte // for field multiplication with 91
		
		MOV R5, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		
		// mix columns
		MOV R0, R6
		MOV R1, R4
		BL Gadget_CodeLs_MixColumns
	
		/* copy registers value from the stack */
		POP {R0,R1,R5,R7,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: R2 for add--------*/
		/*====================================*/
GetR21:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,LR}
		/* r5 stores the pointer of R9 */
		/* r7 stores the address of tabGByte */
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		LDRH R0, [R11, R12]
		AND R1, R0, #0xFF
		LSR R0, R0, #8
		LDRB R2, [R7, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R7, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R9, R5]
		ADD R12, #2
		ADD R5, #4

		/* copy registers value from the stack */
		POP {R0,R1,R2,LR}
		/* return */
		BX LR
		/*====================================*/
		
		/*====================================*/
		/*-----function: R2 for shiftRows-----*/
		/*====================================*/
GetR22:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,LR}
		/* r4 stores the address of tabGByte */
		/* r5 stores the pointer of R9
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		LDRH R0, [R11, R12]
		AND R1, R0, #0xFF
		LSR R0, R0, #8
		LDRB R2, [R4, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R4, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R9, R5]
		ADD R12, #2
		ADD R5, #4

		/* copy registers value from the stack */
		POP {R0,R1,R2,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-----function: R2 for mixColumn-----*/
		/*====================================*/
GetR23:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,LR}
		/* r5 stores the pointer of R9 */
		/* r7 stores the address of tabGByte */
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		LDRH R0, [R11, R12]
		AND R1, R0, #0xFF
		LSR R0, R0, #8
		LDRB R2, [R7, R0]
		ORR R0, R2, R0, LSL #8
		LDRB R2, [R7, R1]
		ORR R1, R2, R1, LSL #8
		ORR R0, R1, R0, LSL #16
		STR R0, [R9, R5]
		ADD R12, #2
		ADD R5, #4

		/* copy registers value from the stack */
		POP {R0,R1,R2,LR}
		/* return */
		BX LR
		/*====================================*/
		
.end
	