/***************************************************************************************/
/* @file   gadget_CodeLs_AddRoundKey.s                                                 */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1, 0;91, 1] and the f function is fixed as        */
/*		   addRoundKey, or saying an Eor operation between two vectors defined in      */
/*		   F_{2^8}^{16}.                                                               */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_AddRoundKey
	
Gadget_CodeLs_AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R3,R4,R5,R6,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */

		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		LDR R2, =tabGByte // for field multiplication with 91
		
		/*-------1/16.1 part--------*/
		// Part A
		LDRB R3, [R0]
		LDRB R4, [R0, #1]
		LDRB R5, [R0, #32]
		LDRB R6, [R0, #33]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1]
		
		/*-------1/16.2 part--------*/
		// Part A
		LDRB R3, [R0, #2]
		LDRB R4, [R0, #3]
		LDRB R5, [R0, #34]
		LDRB R6, [R0, #35]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #4]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #2]
		
		/*-------1/16.3 part--------*/
		// Part A
		LDRB R3, [R0, #4]
		LDRB R4, [R0, #5]
		LDRB R5, [R0, #36]
		LDRB R6, [R0, #37]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #8]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #4]
		
		/*-------1/16.4 part--------*/
		// Part A
		LDRB R3, [R0, #6]
		LDRB R4, [R0, #7]
		LDRB R5, [R0, #38]
		LDRB R6, [R0, #39]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #12]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #6]
		
		/*-------1/16.5 part--------*/
		// Part A
		LDRB R3, [R0, #8]
		LDRB R4, [R0, #9]
		LDRB R5, [R0, #40]
		LDRB R6, [R0, #41]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #16]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #8]
		
		/*-------1/16.6 part--------*/
		// Part A
		LDRB R3, [R0, #10]
		LDRB R4, [R0, #11]
		LDRB R5, [R0, #42]
		LDRB R6, [R0, #43]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #20]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #10]
		
		/*-------1/16.7 part--------*/
		// Part A
		LDRB R3, [R0, #12]
		LDRB R4, [R0, #13]
		LDRB R5, [R0, #44]
		LDRB R6, [R0, #45]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #24]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #12]
		
		/*-------1/16.8 part--------*/
		// Part A
		LDRB R3, [R0, #14]
		LDRB R4, [R0, #15]
		LDRB R5, [R0, #46]
		LDRB R6, [R0, #47]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #28]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #14]
		
		/*-------1/16.9 part--------*/
		// Part A
		LDRB R3, [R0, #16]
		LDRB R4, [R0, #17]
		LDRB R5, [R0, #48]
		LDRB R6, [R0, #49]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #32]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #16]
		
		/*-------1/16.10 part--------*/
		// Part A
		LDRB R3, [R0, #18]
		LDRB R4, [R0, #19]
		LDRB R5, [R0, #50]
		LDRB R6, [R0, #51]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #36]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #18]
		
		/*-------1/16.11 part--------*/
		// Part A
		LDRB R3, [R0, #20]
		LDRB R4, [R0, #21]
		LDRB R5, [R0, #52]
		LDRB R6, [R0, #53]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #40]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #20]
		
		/*-------1/16.12 part--------*/
		// Part A
		LDRB R3, [R0, #22]
		LDRB R4, [R0, #23]
		LDRB R5, [R0, #54]
		LDRB R6, [R0, #55]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #44]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #22]
		
		/*-------1/16.13 part--------*/
		// Part A
		LDRB R3, [R0, #24]
		LDRB R4, [R0, #25]
		LDRB R5, [R0, #56]
		LDRB R6, [R0, #57]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #48]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #24]
		
		/*-------1/16.14 part--------*/
		// Part A
		LDRB R3, [R0, #26]
		LDRB R4, [R0, #27]
		LDRB R5, [R0, #58]
		LDRB R6, [R0, #59]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #52]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #26]
		
		/*-------1/16.15 part--------*/
		// Part A
		LDRB R3, [R0, #28]
		LDRB R4, [R0, #29]
		LDRB R5, [R0, #60]
		LDRB R6, [R0, #61]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #56]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #28]
		
		/*-------1/16.16 part--------*/
		// Part A
		LDRB R3, [R0, #30]
		LDRB R4, [R0, #31]
		LDRB R5, [R0, #62]
		LDRB R6, [R0, #63]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// F Function
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		EOR R3, R5
		// Refresh with R2
		LDR R4, [R12, #60]
		EOR R3, R4
		// Eor and Store
		EOR R3, R3, R3, LSR #16
		STRH R3, [R1, #30]
		
		/* copy registers value from the stack */
		POP {R0,R1,R3,R4,R5,R6,R12,LR}
		/* return */
		BX LR
.end
	