/***************************************************************************************/
/* @file   gadget_CodeLs_MixColumns.s                                                  */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1, 0;91, 1] and the f function is fixed as        */
/*		   MixColumns of AES-128.                                                      */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_MixColumns
	
Gadget_CodeLs_MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */
		
		/* fix some Regs */
		LDR R12, =random2 // store r2 * 16
		MOV R2, R0 
		MOV R3, R1
		LDR R4, =tabGByte // for field multiplication with 91
		MOV R9, #0xFF
		
		/*-------1/4 Part.1--------*/
		// for [0, 1, 2, 3]
		// Part A
		LDRB R0, [R2]
		LDRB R1, [R2, #2]
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #4]
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #6]
		EOR R0, R0, R1, LSL #24
		BL mul_mat_vec_4 
		// r1 stores the res
		MOV R8, R1
		LDRB R0, [R2, #1]
		LDRB R0, [R4, R0] // multiply by M1
		LDRB R1, [R2, #3]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #5]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #7]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #24 
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A
		AND R5, R9, R8
		AND R0, R9, R1
		ORR R5, R5, R0, LSL #16
		AND R6, R9, R8, LSR #8
		AND R0, R9, R1, LSR #8
		ORR R6, R6, R0, LSL #16
		AND R7, R9, R8, LSR #16
		AND R0, R9, R1, LSR #16
		ORR R7, R7, R0, LSL #16
		AND R8, R9, R8, LSR #24
		AND R0, R9, R1, LSR #24
		ORR R8, R8, R0, LSL #16
		// Refresh with R2
		LDR R0, [R12]
		EOR R5, R0
		LDR R0, [R12, #4]
		EOR R6, R0
		LDR R0, [R12, #8]
		EOR R7, R0
		LDR R0, [R12, #12]
		EOR R8, R0
		// Eor and Store
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		EOR R7, R7, R7, LSR #16
		EOR R8, R8, R8, LSR #16
		STRH R5, [R3]
		STRH R6, [R3, #2]
		STRH R7, [R3, #4]
		STRH R8, [R3, #6]
		
		/*-------1/4 Part.2--------*/
		// for [4, 5, 6, 7]
		// Part A
		LDRB R0, [R2, #8]
		LDRB R1, [R2, #10]
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #12]
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #14]
		EOR R0, R0, R1, LSL #24
		BL mul_mat_vec_4 
		// r1 stores the res
		MOV R8, R1
		LDRB R0, [R2, #9]
		LDRB R0, [R4, R0] // multiply by M1
		LDRB R1, [R2, #11]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #13]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #15]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #24 
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A
		AND R5, R9, R8
		AND R0, R9, R1
		ORR R5, R5, R0, LSL #16
		AND R6, R9, R8, LSR #8
		AND R0, R9, R1, LSR #8
		ORR R6, R6, R0, LSL #16
		AND R7, R9, R8, LSR #16
		AND R0, R9, R1, LSR #16
		ORR R7, R7, R0, LSL #16
		AND R8, R9, R8, LSR #24
		AND R0, R9, R1, LSR #24
		ORR R8, R8, R0, LSL #16
		// Refresh with R2
		LDR R0, [R12, #16]
		EOR R5, R0
		LDR R0, [R12, #20]
		EOR R6, R0
		LDR R0, [R12, #24]
		EOR R7, R0
		LDR R0, [R12, #28]
		EOR R8, R0
		// Eor and Store
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		EOR R7, R7, R7, LSR #16
		EOR R8, R8, R8, LSR #16
		STRH R5, [R3, #8]
		STRH R6, [R3, #10]
		STRH R7, [R3, #12]
		STRH R8, [R3, #14]
		
		/*-------1/4 Part.3--------*/
		// for [8, 9, 10, 11]
		// Part A
		LDRB R0, [R2, #16]
		LDRB R1, [R2, #18]
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #20]
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #22]
		EOR R0, R0, R1, LSL #24
		BL mul_mat_vec_4 
		// r1 stores the res
		MOV R8, R1
		LDRB R0, [R2, #17]
		LDRB R0, [R4, R0] // multiply by M1
		LDRB R1, [R2, #19]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #21]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #23]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #24 
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A
		AND R5, R9, R8
		AND R0, R9, R1
		ORR R5, R5, R0, LSL #16
		AND R6, R9, R8, LSR #8
		AND R0, R9, R1, LSR #8
		ORR R6, R6, R0, LSL #16
		AND R7, R9, R8, LSR #16
		AND R0, R9, R1, LSR #16
		ORR R7, R7, R0, LSL #16
		AND R8, R9, R8, LSR #24
		AND R0, R9, R1, LSR #24
		ORR R8, R8, R0, LSL #16
		// Refresh with R2
		LDR R0, [R12, #32]
		EOR R5, R0
		LDR R0, [R12, #36]
		EOR R6, R0
		LDR R0, [R12, #40]
		EOR R7, R0
		LDR R0, [R12, #44]
		EOR R8, R0
		// Eor and Store
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		EOR R7, R7, R7, LSR #16
		EOR R8, R8, R8, LSR #16
		STRH R5, [R3, #16]
		STRH R6, [R3, #18]
		STRH R7, [R3, #20]
		STRH R8, [R3, #22]
		
		/*-------1/4 Part.4--------*/
		// for [12, 13, 14, 15]
		// Part A
		LDRB R0, [R2, #24]
		LDRB R1, [R2, #26]
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #28]
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #30]
		EOR R0, R0, R1, LSL #24
		BL mul_mat_vec_4 
		// r1 stores the res
		MOV R8, R1
		LDRB R0, [R2, #25]
		LDRB R0, [R4, R0] // multiply by M1
		LDRB R1, [R2, #27]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #8
		LDRB R1, [R2, #29]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #16
		LDRB R1, [R2, #31]
		LDRB R1, [R4, R1] // multiply by M1
		EOR R0, R0, R1, LSL #24 
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A
		AND R5, R9, R8
		AND R0, R9, R1
		ORR R5, R5, R0, LSL #16
		AND R6, R9, R8, LSR #8
		AND R0, R9, R1, LSR #8
		ORR R6, R6, R0, LSL #16
		AND R7, R9, R8, LSR #16
		AND R0, R9, R1, LSR #16
		ORR R7, R7, R0, LSL #16
		AND R8, R9, R8, LSR #24
		AND R0, R9, R1, LSR #24
		ORR R8, R8, R0, LSL #16
		// Refresh with R2
		LDR R0, [R12, #48]
		EOR R5, R0
		LDR R0, [R12, #52]
		EOR R6, R0
		LDR R0, [R12, #56]
		EOR R7, R0
		LDR R0, [R12, #60]
		EOR R8, R0
		// Eor and Store
		EOR R5, R5, R5, LSR #16
		EOR R6, R6, R6, LSR #16
		EOR R7, R7, R7, LSR #16
		EOR R8, R8, R8, LSR #16
		STRH R5, [R3, #24]
		STRH R6, [R3, #26]
		STRH R7, [R3, #28]
		STRH R8, [R3, #30]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R12,LR}
		/* return */
		BX LR
.end
	