/***************************************************************************************/
/* @file   gadget_CodeLs_ShiftRows.s                                                   */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1, 0;91, 1] and the f function is fixed as        */
/*		   shiftRows of AES-128.                                                       */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_ShiftRows
	
Gadget_CodeLs_ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R3,R4,R5,R6,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */

		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		LDR R2, =tabGByte // for field multiplication with 91
		
		/*-------1/8 part.1--------*/
		// for [0, 1]
		// Part A
		LDRB R3, [R0]
		LDRB R4, [R0, #1]
		LDRB R5, [R0, #2]
		LDRB R6, [R0, #3]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [0, 13]
		LDR R4, [R12]
		EOR R3, R4
		LDR R4, [R12, #52]
		EOR R5, R4
		// Eor and Store
		// store in [0, 13]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1]
		STRH R5, [R1, #26]
		
		/*-------1/8 part.2--------*/
		// for [2, 3]
		// Part A
		LDRB R3, [R0, #4]
		LDRB R4, [R0, #5]
		LDRB R5, [R0, #6]
		LDRB R6, [R0, #7]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [10, 7]
		LDR R4, [R12, #40]
		EOR R3, R4
		LDR R4, [R12, #28]
		EOR R5, R4
		// Eor and Store
		// store in [10, 7]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #20]
		STRH R5, [R1, #14]
		
		/*-------1/8 part.3--------*/
		// for [4, 5]
		// Part A
		LDRB R3, [R0, #8]
		LDRB R4, [R0, #9]
		LDRB R5, [R0, #10]
		LDRB R6, [R0, #11]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [4, 1]
		LDR R4, [R12, #16]
		EOR R3, R4
		LDR R4, [R12, #4]
		EOR R5, R4
		// Eor and Store
		// store in [4, 1]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #8]
		STRH R5, [R1, #2]
		
		/*-------1/8 part.4--------*/
		// for [6, 7]
		// Part A
		LDRB R3, [R0, #12]
		LDRB R4, [R0, #13]
		LDRB R5, [R0, #14]
		LDRB R6, [R0, #15]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [14, 11]
		LDR R4, [R12, #56]
		EOR R3, R4
		LDR R4, [R12, #44]
		EOR R5, R4
		// Eor and Store
		// store in [14, 11]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #28]
		STRH R5, [R1, #22]
		
		/*-------1/8 part.5--------*/
		// for [8, 9]
		// Part A
		LDRB R3, [R0, #16]
		LDRB R4, [R0, #17]
		LDRB R5, [R0, #18]
		LDRB R6, [R0, #19]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [8, 5]
		LDR R4, [R12, #32]
		EOR R3, R4
		LDR R4, [R12, #20]
		EOR R5, R4
		// Eor and Store
		// store in [8, 5]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #16]
		STRH R5, [R1, #10]
		
		/*-------1/8 part.6--------*/
		// for [10, 11]
		// Part A
		LDRB R3, [R0, #20]
		LDRB R4, [R0, #21]
		LDRB R5, [R0, #22]
		LDRB R6, [R0, #23]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [2, 15]
		LDR R4, [R12, #8]
		EOR R3, R4
		LDR R4, [R12, #60]
		EOR R5, R4
		// Eor and Store
		// store in [2, 15]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #4]
		STRH R5, [R1, #30]
		
		/*-------1/8 part.7--------*/
		// for [12, 13]
		// Part A
		LDRB R3, [R0, #24]
		LDRB R4, [R0, #25]
		LDRB R5, [R0, #26]
		LDRB R6, [R0, #27]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [12, 9]
		LDR R4, [R12, #48]
		EOR R3, R4
		LDR R4, [R12, #36]
		EOR R5, R4
		// Eor and Store
		// store in [12, 9]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #24]
		STRH R5, [R1, #18]
		
		/*-------1/8 part.8--------*/
		// for [14, 15]
		// Part A
		LDRB R3, [R0, #28]
		LDRB R4, [R0, #29]
		LDRB R5, [R0, #30]
		LDRB R6, [R0, #31]
		LDRB R4, [R2, R4] // multiply by M1
		LDRB R6, [R2, R6] // multiply by M1
		// Multiply by A
		ORR R3, R3, R4, LSL #16
		ORR R5, R5, R6, LSL #16
		// Refresh with R2 
		// changed to [6, 3]
		LDR R4, [R12, #24]
		EOR R3, R4
		LDR R4, [R12, #12]
		EOR R5, R4
		// Eor and Store
		// store in [6, 3]
		EOR R3, R3, R3, LSR #16
		EOR R5, R5, R5, LSR #16
		STRH R3, [R1, #12]
		STRH R5, [R1, #6]
		
		/* copy registers value from the stack */
		POP {R0,R1,R3,R4,R5,R6,R12,LR}
		/* return */
		BX LR
.end
	