/****************************************************************************/
/* @file   gadget_CodeMuL.s                                                 */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of gadget CodeMuL proposed in XXXXXXXXXXXX.       */
/*         This function calculates the byte_wise multiplication of two     */
/*		   vectors over F_{2^8}^{2} in the masked domain. And the generator */
/*         matrix is fixed as [1,0;91,1].                                   */
/****************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeMuL
	
Gadget_CodeMuL:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src1 */
		/* r1 stores the address of src2 */
		/* r2 stores the address of result */
		
		/* fix some Regs */
		LDR R11, =random1 // store R1
		LDR R10, =random2 // store R2 
		LDR R3, =bhtable // for field multiplication
		LDR R4, =bltable
		LDRH R0, [R0] // r0 stores the vec1
		LDRH R1, [R1] // r1 stored the vec2
		MOV R12, #0xFF
		
		/*-------The 0th row of matrix--------*/
		// Outer Product & Refresh with R1 & Multiply M0
		// prepare the byte
		AND R7, R12, R0
		LSL R7, R7, #4
		// the first byte of vec (1, 2)
		AND R8, R12, R1
		ORR R9, R7, R8, LSR #4
		LDRB R9, [R3, R9]
		AND R8, #0xF
		ORR R8, R7, R8
		LDRB R8, [R4, R8]
		EOR R5, R8, R9
		LDRB R8, [R11] // refresh with R1
		EOR R5, R8
		// the second byte of vec (1, 2)
		AND R8, R12, R1, LSR #8
		ORR R9, R7, R8, LSR #4
		LDRB R9, [R3, R9]
		AND R8, #0xF
		ORR R8, R7, R8
		LDRB R8, [R4, R8]
		EOR R9, R8
		LDRB R8, [R11, #1] // refresh with R1
		EOR R9, R8
		LDR R6, =tabGByte // multiply by M0
		LDRB R9, [R6, R9] 
		EOR R9, R5
		// r9 stores the vec : (1, 1)
		
		// Multiply by A
		// r9 stores the vec : (1, 2)
		
		// Refresh with R2
		LDRH R5, [R10]
		EOR R9, R5
		// r9 stores the vec : (1, 2)
		
		/*-------The 1th row of matrix--------*/
		// Outer Product & Refresh with R1 & Multiply M1
		// prepare the byte
		AND R7, R12, R0, LSR #8
		LSL R7, R7, #4
		// the first byte of vec (1, 2)
		AND R8, R12, R1
		ORR R6, R7, R8, LSR #4
		LDRB R6, [R3, R6]
		AND R8, #0xF
		ORR R8, R7, R8
		LDRB R8, [R4, R8]
		EOR R5, R8, R6 
		LDRB R8, [R11, #2] // refresh with R1
		EOR R5, R8
		LDR R6, =tabGByte // multiply by M1
		LDRB R5, [R6, R5]
		// the second byte of vec (1, 2)
		AND R8, R12, R1, LSR #8
		ORR R6, R7, R8, LSR #4
		LDRB R6, [R3, R6]
		AND R8, #0xF
		ORR R8, R7, R8
		LDRB R8, [R4, R8]
		EOR R6, R8 
		LDRB R8, [R11, #3] // refresh with R1
		EOR R6, R8
		LDR R7, =tabMByte // multiply by M1
		LDRB R6, [R7, R6]
		EOR R5, R6
		// r5 vec : (1, 1)
		
		// Multiply by A
		// r5 stores the vec : (1, 2)
		
		// Refresh with R2
		LDRH R8, [R10, #2]
		EOR R5, R8
		// r5 stores the vec : (1, 2)
		
		// Eor and Store
		EOR R9, R5
		STRH R9, [R2]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	