/**************************************************************************/
/* @file   aes.s                                                          */
/* @author yangx                                                          */
/* @date   2022-4-11                                                      */
/* @brief  eight functions                                                */
/*         -- AES_Encrypt_GCB [public]                                    */
/*            Implementation of AES-128 masked by code-based masking,     */ 
/*            please refer to xxxxxxx to see the detailed masking         */
/*            scheme. The generator matrix is fixed as                    */
/*            [1,0,0;1,1,0;1,0,1]. This function has no output and the    */
/*            input is the address of cyphertext.                         */ 
/*         -- AddRoundKey [private]                                       */
/*            Implementation of AddRoundKey masked by code-based masking. */
/*         -- Sbox [private]                                              */
/*            Implementation of SubBytes masked by code-based masking.    */
/*            Note that S-box can be implemented by an inverse and an     */
/*            affine transformation.                                      */
/*         -- ShiftRows [private]                                         */
/*            Implementation of ShiftRows masked by code-based masking.   */
/*         -- MixColumns [private]                                        */
/*            Implementation of MixColumns masked by code-based masking.  */
/*         -- three functions for constructing R2                         */
/*            --- GetR21 for addRoundKey                                  */
/*            --- GetR22 for shiftRows                                    */
/*            --- GetR23 for mixColumns                                   */
/**************************************************************************/

.section .data
        temp: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	 code_pt: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	code_key: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	pt_state: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	  temp_2: .byte 0x00, 0x00, 0x00
	  temp_4: .byte 0x00, 0x00, 0x00
	 temp_16: .byte 0x00, 0x00, 0x00
	  temp_n: .byte 0x00, 0x00, 0x00
			
.section .text
.syntax unified
.thumb
.global AES_Encrypt_GCB
	
		/*====================================*/
		/*-----function: AES_Encrypt_GCB------*/
		/*====================================*/
AES_Encrypt_GCB:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the adress of result */
		
		/* fix some Regs */
		MOV R12, R0
		// randomSeed stores all the random numbers utlized in AES
		LDR R11, =randomSeed
		
		/* encode the plaintext */
		LDR R5, =temp // container for the encoded plaintext and key
		LDR R6, =pt_aes // plaintext
		MOV R10, #0xFF
		// encode
		LDR R0, [R6]
		LDR R1, [R11]
		LDR R2, [R11, #4]
		EOR R0, R1
		EOR R0, R2
		AND R3, R0, R10
		AND R4, R1, R10
		AND R8, R2, R10
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5]
		AND R3, R10, R0, LSR #8
		AND R4, R10, R1, LSR #8
		AND R8, R10, R2, LSR #8
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #3]
		AND R3, R10, R0, LSR #16
		AND R4, R10, R1, LSR #16
		AND R8, R10, R2, LSR #16
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #6]
		AND R3, R10, R0, LSR #24
		AND R4, R10, R1, LSR #24
		AND R8, R10, R2, LSR #24
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #9]
		LDR R0, [R6, #4]
		LDR R1, [R11, #8]
		LDR R2, [R11, #12]
		EOR R0, R1
		EOR R0, R2
		AND R3, R0, R10
		AND R4, R1, R10
		AND R8, R2, R10
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #12]
		AND R3, R10, R0, LSR #8
		AND R4, R10, R1, LSR #8
		AND R8, R10, R2, LSR #8
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #15]
		AND R3, R10, R0, LSR #16
		AND R4, R10, R1, LSR #16
		AND R8, R10, R2, LSR #16
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #18]
		AND R3, R10, R0, LSR #24
		AND R4, R10, R1, LSR #24
		AND R8, R10, R2, LSR #24
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #21]
		LDR R0, [R6, #8]
		LDR R1, [R11, #16]
		LDR R2, [R11, #20]
		EOR R0, R1
		EOR R0, R2
		AND R3, R0, R10
		AND R4, R1, R10
		AND R8, R2, R10
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #24]
		AND R3, R10, R0, LSR #8
		AND R4, R10, R1, LSR #8
		AND R8, R10, R2, LSR #8
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #27]
		AND R3, R10, R0, LSR #16
		AND R4, R10, R1, LSR #16
		AND R8, R10, R2, LSR #16
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #30]
		AND R3, R10, R0, LSR #24
		AND R4, R10, R1, LSR #24
		AND R8, R10, R2, LSR #24
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #33]
		LDR R0, [R6, #12]
		LDR R1, [R11, #24]
		LDR R2, [R11, #28]
		EOR R0, R1
		EOR R0, R2
		AND R3, R0, R10
		AND R4, R1, R10
		AND R8, R2, R10
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #36]
		AND R3, R10, R0, LSR #8
		AND R4, R10, R1, LSR #8
		AND R8, R10, R2, LSR #8
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #39]
		AND R3, R10, R0, LSR #16
		AND R4, R10, R1, LSR #16
		AND R8, R10, R2, LSR #16
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STR R3, [R5, #42]
		AND R3, R10, R0, LSR #24
		AND R4, R10, R1, LSR #24
		AND R8, R10, R2, LSR #24
		ORR R3, R3, R4, LSL #8
		ORR R3, R3, R8, LSL #16
		STRH R3, [R5, #45]
		LSR R3, R3, #16
		STRB R3, [R5, #47]
		ADD R11, #32 // move the pointer of randomSeed
		
		/* add encoded key to temp */
		LDR R8, =maskedKeys // note that the key has been extended and encoded
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		ADD R8, #48 // move the pointer of maskedKeys
		
		LDR R10, =random1 // container for R1
		LDR R9, =random2 // container for R2 * 16
		LDR R7, =code_pt
		LDR R6, =pt_state
		
		/*--add Round Key--*/
		// add round key
		BL AddRoundKey
		ADD R11, #96 // move the pointer of randomSeed
		
		/*----AES Round----*/
		MOV R4, #0 // counter for AES rounds
loopRounds:
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3 // move the address of pt_state
		ADD R7, #3 // move the address of code_pt
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72
		// move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		BL ShiftRows
		ADD R11, #96 // move the pointer of randomSeed
		
		/*--MixColumns--*/
		BL MixColumns
		ADD R11, #96 // move the pointer of randomSeed
		
		/* add encoded key of next round to temp */
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		ADD R8, #48 // move the pointer of maskedKeys
		
		/*--AddRoundKey--*/
		BL AddRoundKey
		ADD R11, #96 // move the pointer of randomSeed
		
		ADD R4, #1
		CMP R4, #9
		IT NE 
		BNE loopRounds

		/*----the last round----*/
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		MOV R6, R5
		BL ShiftRows
		ADD R11, #96 // move the pointer of randomSeed
		
		/* add the encoded key of the last round to temp */
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		
		/*--AddRoundKey--*/
		LDR R6, =pt_state
		BL AddRoundKey
		
		/* decode */
		LDR R0, [R6] // r6 (code_pt) stores the address of final encoded cyphertext
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12] // r12 stores the address of ct_aes
		LDR R0, [R6, #3]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #1]
		LDR R0, [R6, #6]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #2]
		LDR R0, [R6, #9]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #3]
		LDR R0, [R6, #12]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #4]
		LDR R0, [R6, #15]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #5]
		LDR R0, [R6, #18]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #6]
		LDR R0, [R6, #21]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #7]
		LDR R0, [R6, #24]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #8]
		LDR R0, [R6, #27]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #9]
		LDR R0, [R6, #30]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #10]
		LDR R0, [R6, #33]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #11]
		LDR R0, [R6, #36]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #12]
		LDR R0, [R6, #39]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #13]
		LDR R0, [R6, #42]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #14]
		LDR R0, [R6, #45]
		EOR R1, R0, R0, LSR #8
		EOR R1, R1, R0, LSR #16
		STRB R1, [R12, #15]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-------function: AddRoundKey--------*/
		/*====================================*/
AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R10,R12,LR}
		/* r5 stores the address of src */
		/* r4 stores the pointer of R9 */
		/* r6 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		// initialize the pointers
		MOV R4, #0
		MOV R12, #0
		// R2 construction
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		
		// add round key
		MOV R0, R5
		MOV R1, R6
		BL Gadget_CodeLs_AddRoundKey
	
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R10,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-----------function: Sbox-----------*/
		/*====================================*/
Sbox:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* r6 stores the address of src */
		/* r7 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r10 stores the address of random1 */
		/* r11 stores the address of randomSeed */
		
		/* fix some Regs */
		LDR R8, =temp_2
		LDR R5, =temp_4
		LDR R4, =temp_16
		LDR R3, =temp_n
		
		// R2 construction
		LDRH R0, [R11]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #2]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #4]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// x^2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, #0 // parameter indicating .^2
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #6]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10]
		STRB R0, [R10, #4]
		LSR R0, R0, #8
		STRB R0, [R10, #8]
		LDRH R0, [R11, #8]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10, #1]
		STRB R0, [R10, #5]
		LSR R0, R0, #8
		STRB R0, [R10, #9]
		LDRH R0, [R11, #10]
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		STRH R1, [R10, #2]
		AND R1, R0, #0xFF
		STRH R1, [R10, #6]
		LSR R0, R0, #8
		STRH R0, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #12]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #14]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #16]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// tempn = x * x ^ 2
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #18]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #20]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #22]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// tempn^4
		MOV R0, R3 // temp_n
		MOV R1, R5 // temp_4
		MOV R2, #1 // parameter indicating .^4
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #24]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10]
		STRB R0, [R10, #4]
		LSR R0, R0, #8
		STRB R0, [R10, #8]
		LDRH R0, [R11, #26]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10, #1]
		STRB R0, [R10, #5]
		LSR R0, R0, #8
		STRB R0, [R10, #9]
		LDRH R0, [R11, #28]
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		STRH R1, [R10, #2]
		AND R1, R0, #0xFF
		STRH R1, [R10, #6]
		LSR R0, R0, #8
		STRH R0, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #30]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #32]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #34]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// dst = (temp4) * (tempn)
		MOV R0, R5 // temp_4
		MOV R1, R3 // temp_n
		MOV R2, R7 // dst
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #36]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #38]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #40]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// dst^16
		MOV R0, R7 // dst
		LDR R1, =temp_16
		MOV R2, #2 // parameter indicating .^16
		BL Gadget_CodeL
		
		// R1 construction
		LDRH R0, [R11, #42]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10]
		STRB R0, [R10, #4]
		LSR R0, R0, #8
		STRB R0, [R10, #8]
		LDRH R0, [R11, #44]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10, #1]
		STRB R0, [R10, #5]
		LSR R0, R0, #8
		STRB R0, [R10, #9]
		LDRH R0, [R11, #46]
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		STRH R1, [R10, #2]
		AND R1, R0, #0xFF
		STRH R1, [R10, #6]
		LSR R0, R0, #8
		STRH R0, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #48]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #50]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #52]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// src = (temp4) * (temp16)
		MOV R0, R4 // temp_16
		MOV R1, R5 // temp_4
		MOV R2, R6 // src
		BL Gadget_CodeMuL
		
		// R1 construction
		LDRH R0, [R11, #54]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10]
		STRB R0, [R10, #4]
		LSR R0, R0, #8
		STRB R0, [R10, #8]
		LDRH R0, [R11, #56]
		EOR R1, R0, R0, LSR #8
		STRB R1, [R10, #1]
		STRB R0, [R10, #5]
		LSR R0, R0, #8
		STRB R0, [R10, #9]
		LDRH R0, [R11, #58]
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		STRH R1, [R10, #2]
		AND R1, R0, #0xFF
		STRH R1, [R10, #6]
		LSR R0, R0, #8
		STRH R0, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #60]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #62]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #64]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// tempn = src * (temp2)
		MOV R0, R6 // src
		MOV R1, R8 // temp_2
		MOV R2, R3 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #66]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRH R0, [R11, #68]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #8
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #8
		LDRH R0, [R11, #70]
		AND R12, R0, #0xFF
		ORR R1, R1, R12, LSL #16
		LSR R12, R0, #8
		ORR R2, R2, R12, LSL #16
		EOR R12, R1, R2
		STR R1, [R9]
		STR R2, [R9, #4]
		STR R12, [R9, #8]
		// dst = aff(tempn)
		MOV R0, R3 // temp_n
		MOV R1, R7 // dst
		MOV R2, #3 // parameter indicating affine transformation
		BL Gadget_CodeL
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: ShiftRows---------*/
		/*====================================*/
ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R5,R12,LR}
		/* r6 stores the address of dst */
		/* r7 stores the address of src */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		
		/* initialize the pointers */
		MOV R5, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		
		// shift rows
		MOV R0, R7
		MOV R1, R6
		BL Gadget_CodeLs_ShiftRows
	
		
		/* copy registers value from the stack */
		POP {R0,R1,R5,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: MixColumns--------*/
		/*====================================*/
MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R4,R12,LR}
		/* r6 stores the address of src */
		/* r4 stores the pointer of R9 */
		/* r5 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* initialize the pointers */
		MOV R4, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		
		// mix columns
		MOV R0, R6
		MOV R1, R5
		BL Gadget_CodeLs_MixColumns
	
		
		/* copy registers value from the stack */
		POP {R0,R1,R4,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: R2 for add--------*/
		/*====================================*/
GetR21:
		/* copy registers value to the stack */
		PUSH {R0,R5,R6,R7,LR}
		/* r4 stores the pointer of R9 */
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of randomSeed */
		
		LDRH R0, [R11, R12]
		ADD R12, #2
		AND R6, R0, #0xFF
		LSR R7, R0, #8
		LDRH R0, [R11, R12]
		ADD R12, #2
		AND R5, R0, #0xFF
		ORR R6, R6, R5, LSL #8
		LSR R5, R0, #8
		ORR R7, R7, R5, LSL #8
		LDRH R0, [R11, R12]
		ADD R12, #2
		AND R5, R0, #0xFF
		ORR R6, R6, R5, LSL #16
		LSR R5, R0, #8
		ORR R7, R7, R5, LSL #16
		EOR R5, R6, R7
		STR R5, [R9, R4]
		ADD R4, #4
		STR R6, [R9, R4]
		ADD R4, #4
		STR R7, [R9, R4]
		ADD R4, #4
		
		/* copy registers value from the stack */
		POP {R0,R5,R6,R7,LR}
		/* return */
		BX LR
		/*====================================*/
		
		/*====================================*/
		/*-----function: R2 for shiftRows-----*/
		/*====================================*/
GetR22:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,LR}
		/* r5 stores the pointer of R9 */
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of randomSeed */
		
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R5]
		ADD R5, #4
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R5]
		ADD R5, #4
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R5]
		ADD R5, #4
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,LR}
		/* return */
		BX LR
		/*====================================*/
		
		/*====================================*/
		/*-----function: R2 for mixColumn-----*/
		/*====================================*/
GetR23:
		/* copy registers value to the stack */
		PUSH {R0,R1,LR}
		/* r4 stores the pointer of R9 */
		/* r9 stores the address of random2 */ 
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R4]
		ADD R4, #4
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R4]
		ADD R4, #4
		LDRH R0, [R11, R12]
		ADD R12, #2
		EOR R1, R0, R0, LSR #8
		AND R1, 0xFF
		ORR R1, R1, R0, LSL #8
		STR R1, [R9, R4]
		ADD R4, #4

		/* copy registers value from the stack */
		POP {R0,R1,LR}
		/* return */
		BX LR
		/*====================================*/
		
.end
	