/****************************************************************************************/
/* @file   gadget_CodeLs_AddRoundKey.s                                                  */
/* @author yangx                                                                        */
/* @date   2022.4.11                                                                    */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored  */
/*         for the generator matrix [1,0,0;1,1,0;1,0,1] and the f function is fixed as  */
/*		   addRoundKey, or saying an Eor operation between two vectors defined in       */
/*		   F_{2^8}^{16}.                                                                */
/****************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_AddRoundKey
	
Gadget_CodeLs_AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */
		
		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		MOV R11, #0xFF
		
		/*-------the 1/8.1 part--------*/
		/*----part A----*/
		LDR R2, [R0]
		LDR R3, [R0, #3]
		LDR R4, [R0, #48]
		LDR R5, [R0, #51]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12]
		EOR R4, R2
		LDR R2, [R12, #4]
		EOR R5, R2
		LDR R2, [R12, #8]
		EOR R6, R2
		LDR R2, [R12, #12]
		EOR R7, R2
		LDR R2, [R12, #16]
		EOR R8, R2
		LDR R2, [R12, #20]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #3]
		
		/*-------the 1/8.2 part--------*/
		/*----part A----*/
		LDR R2, [R0, #6]
		LDR R3, [R0, #9]
		LDR R4, [R0, #54]
		LDR R5, [R0, #57]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #24]
		EOR R4, R2
		LDR R2, [R12, #28]
		EOR R5, R2
		LDR R2, [R12, #32]
		EOR R6, R2
		LDR R2, [R12, #36]
		EOR R7, R2
		LDR R2, [R12, #40]
		EOR R8, R2
		LDR R2, [R12, #44]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #6]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #9]
		
		/*-------the 1/8.3 part--------*/
		/*----part A----*/
		LDR R2, [R0, #12]
		LDR R3, [R0, #15]
		LDR R4, [R0, #60]
		LDR R5, [R0, #63]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #48]
		EOR R4, R2
		LDR R2, [R12, #52]
		EOR R5, R2
		LDR R2, [R12, #56]
		EOR R6, R2
		LDR R2, [R12, #60]
		EOR R7, R2
		LDR R2, [R12, #64]
		EOR R8, R2
		LDR R2, [R12, #68]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #12]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #15]
		
		/*-------the 1/8.4 part--------*/
		/*----part A----*/
		LDR R2, [R0, #18]
		LDR R3, [R0, #21]
		LDR R4, [R0, #66]
		LDR R5, [R0, #69]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #72]
		EOR R4, R2
		LDR R2, [R12, #76]
		EOR R5, R2
		LDR R2, [R12, #80]
		EOR R6, R2
		LDR R2, [R12, #84]
		EOR R7, R2
		LDR R2, [R12, #88]
		EOR R8, R2
		LDR R2, [R12, #92]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #18]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #21]
		
		/*-------the 1/8.5 part--------*/
		/*----part A----*/
		LDR R2, [R0, #24]
		LDR R3, [R0, #27]
		LDR R4, [R0, #72]
		LDR R5, [R0, #75]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #96]
		EOR R4, R2
		LDR R2, [R12, #100]
		EOR R5, R2
		LDR R2, [R12, #104]
		EOR R6, R2
		LDR R2, [R12, #108]
		EOR R7, R2
		LDR R2, [R12, #112]
		EOR R8, R2
		LDR R2, [R12, #116]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #24]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #27]
		
		/*-------the 1/8.6 part--------*/
		/*----part A----*/
		LDR R2, [R0, #30]
		LDR R3, [R0, #33]
		LDR R4, [R0, #78]
		LDR R5, [R0, #81]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #120]
		EOR R4, R2
		LDR R2, [R12, #124]
		EOR R5, R2
		LDR R2, [R12, #128]
		EOR R6, R2
		LDR R2, [R12, #132]
		EOR R7, R2
		LDR R2, [R12, #136]
		EOR R8, R2
		LDR R2, [R12, #140]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #30]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #33]
		
		/*-------the 1/8.7 part--------*/
		/*----part A----*/
		LDR R2, [R0, #36]
		LDR R3, [R0, #39]
		LDR R4, [R0, #84]
		LDR R5, [R0, #87]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #144]
		EOR R4, R2
		LDR R2, [R12, #148]
		EOR R5, R2
		LDR R2, [R12, #152]
		EOR R6, R2
		LDR R2, [R12, #156]
		EOR R7, R2
		LDR R2, [R12, #160]
		EOR R8, R2
		LDR R2, [R12, #164]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #36]
		EOR R7, R8
		EOR R7, R9
		STR R7, [R1, #39]
		
		/*-------the 1/8.8 part--------*/
		/*----part A----*/
		LDR R2, [R0, #42]
		LDR R3, [R0, #45]
		LDR R4, [R0, #90]
		LDR R5, [R0, #93]
		// add round key
		EOR R2, R4
		EOR R3, R5
		/*----multiply by A----*/
		AND R4, R11, R2
		AND R5, R11, R2, LSR #8
		AND R6, R11, R2, LSR #16
		AND R7, R11, R3
		AND R8, R11, R3, LSR #8
		AND R9, R11, R3, LSR #16
		/*----refresh with R2----*/
		LDR R2, [R12, #168]
		EOR R4, R2
		LDR R2, [R12, #172]
		EOR R5, R2
		LDR R2, [R12, #176]
		EOR R6, R2
		LDR R2, [R12, #180]
		EOR R7, R2
		LDR R2, [R12, #184]
		EOR R8, R2
		LDR R2, [R12, #188]
		EOR R9, R2
		/*----eor and store----*/
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #42]
		EOR R7, R8
		EOR R7, R9
		STRH R7, [R1, #45]
		LSR R7, R7, #16
		STRB R7, [R1, #47]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R11,R12,LR}
		/* return */
		BX LR
.end
	