/***************************************************************************************/
/* @file   gadget_CodeLs_MixColumns.s                                                  */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1,0,0;1,1,0;1,0,1] and the f function is fixed as */
/*		   mixColumns of AES-128.                                                      */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_MixColumns
	
Gadget_CodeLs_MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */
		
		/* fix some Regs */
		LDR R12, =random2
		MOV R11, R0
		MOV R10, R1
		MOV R9, #0xFF
		
		/*-------the 1/4.1 part--------*/
		// for [0, 1, 2, 3]
		/*----Part A----*/
		LDRH R2, [R11]
		LDRH R3, [R11, #3]
		LDRH R4, [R11, #6]
		LDRH R5, [R11, #9]
		AND R0, R9, R2
		AND R1, R9, R3
		ORR R0, R0, R1, LSL #8
		AND R1, R9, R4
		ORR R0, R0, R1, LSL #16
		AND R1, R9, R5
		ORR R0, R0, R1, LSL #24
		BL mul_mat_vec_4
		MOV R6, R1
		// r6 stores the vec1
		AND R2, R9, R2, LSR #8
		AND R3, R9, R3, LSR #8
		AND R4, R9, R4, LSR #8
		AND R5, R9, R5, LSR #8
		MOV R0, R2
		ORR R0, R0, R3, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec2
		/*----multiply by A----*/
		AND R0, R9, R6
		AND R2, R9, R6, LSR #8
		AND R3, R9, R6, LSR #16
		AND R4, R9, R6, LSR #24
		AND R5, R9, R1
		AND R6, R9, R1, LSR #8
		AND R7, R9, R1, LSR #16
		AND R8, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12]
		EOR R0, R1
		LDR R1, [R12, #4]
		EOR R5, R1
		LDR R1, [R12, #12]
		EOR R2, R1
		LDR R1, [R12, #16]
		EOR R6, R1
		LDR R1, [R12, #24]
		EOR R3, R1
		LDR R1, [R12, #28]
		EOR R7, R1
		LDR R1, [R12, #36]
		EOR R4, R1
		LDR R1, [R12, #40]
		EOR R8, R1
		/*----eor----*/
		EOR R5, R0
		EOR R6, R2
		EOR R7, R3
		EOR R8, R4
		/*----part A again----*/
		LDRB R0, [R11, #2]
		LDRB R1, [R11, #5]
		LDRB R2, [R11, #8]
		LDRB R3, [R11, #11]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		ORR R0, R0, R3, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec3
		/*----multiply by A again----*/
		AND R0, R9, R1
		AND R2, R9, R1, LSR #8
		AND R3, R9, R1, LSR #16
		AND R4, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #8]
		EOR R0, R1
		LDR R1, [R12, #20]
		EOR R2, R1
		LDR R1, [R12, #32]
		EOR R3, R1
		LDR R1, [R12, #44]
		EOR R4, R1
		/*----eor and store----*/
		EOR R0, R5
		EOR R2, R6
		EOR R3, R7
		EOR R4, R8
		STR R0, [R10]
		STR R2, [R10, #3]
		STR R3, [R10, #6]
		STR R4, [R10, #9]
		
		/*-------the 1/4.2 part--------*/
		// for [4, 5, 6, 7]
		/*----Part A----*/
		LDRH R2, [R11, #12]
		LDRH R3, [R11, #15]
		LDRH R4, [R11, #18]
		LDRH R5, [R11, #21]
		AND R0, R9, R2
		AND R1, R9, R3
		ORR R0, R0, R1, LSL #8
		AND R1, R9, R4
		ORR R0, R0, R1, LSL #16
		AND R1, R9, R5
		ORR R0, R0, R1, LSL #24
		BL mul_mat_vec_4
		MOV R6, R1
		// r6 stores the vec1
		AND R2, R9, R2, LSR #8
		AND R3, R9, R3, LSR #8
		AND R4, R9, R4, LSR #8
		AND R5, R9, R5, LSR #8
		MOV R0, R2
		ORR R0, R0, R3, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec2
		/*----multiply by A----*/
		AND R0, R9, R6
		AND R2, R9, R6, LSR #8
		AND R3, R9, R6, LSR #16
		AND R4, R9, R6, LSR #24
		AND R5, R9, R1
		AND R6, R9, R1, LSR #8
		AND R7, R9, R1, LSR #16
		AND R8, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #48]
		EOR R0, R1
		LDR R1, [R12, #52]
		EOR R5, R1
		LDR R1, [R12, #60]
		EOR R2, R1
		LDR R1, [R12, #64]
		EOR R6, R1
		LDR R1, [R12, #72]
		EOR R3, R1
		LDR R1, [R12, #76]
		EOR R7, R1
		LDR R1, [R12, #84]
		EOR R4, R1
		LDR R1, [R12, #88]
		EOR R8, R1
		/*----eor----*/
		EOR R5, R0
		EOR R6, R2
		EOR R7, R3
		EOR R8, R4
		/*----part A again----*/
		LDRB R0, [R11, #14]
		LDRB R1, [R11, #17]
		LDRB R2, [R11, #20]
		LDRB R3, [R11, #23]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		ORR R0, R0, R3, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec3
		/*----multiply by A again----*/
		AND R0, R9, R1
		AND R2, R9, R1, LSR #8
		AND R3, R9, R1, LSR #16
		AND R4, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #56]
		EOR R0, R1
		LDR R1, [R12, #68]
		EOR R2, R1
		LDR R1, [R12, #80]
		EOR R3, R1
		LDR R1, [R12, #92]
		EOR R4, R1
		/*----eor and store----*/
		EOR R0, R5
		EOR R2, R6
		EOR R3, R7
		EOR R4, R8
		STR R0, [R10, #12]
		STR R2, [R10, #15]
		STR R3, [R10, #18]
		STR R4, [R10, #21]
		
		/*-------the 1/4.3 part--------*/
		// for [8, 9, 10, 11]
		/*----Part A----*/
		LDRH R2, [R11, #24]
		LDRH R3, [R11, #27]
		LDRH R4, [R11, #30]
		LDRH R5, [R11, #33]
		AND R0, R9, R2
		AND R1, R9, R3
		ORR R0, R0, R1, LSL #8
		AND R1, R9, R4
		ORR R0, R0, R1, LSL #16
		AND R1, R9, R5
		ORR R0, R0, R1, LSL #24
		BL mul_mat_vec_4
		MOV R6, R1
		// r6 stores the vec1
		AND R2, R9, R2, LSR #8
		AND R3, R9, R3, LSR #8
		AND R4, R9, R4, LSR #8
		AND R5, R9, R5, LSR #8
		MOV R0, R2
		ORR R0, R0, R3, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec2
		/*----multiply by A----*/
		AND R0, R9, R6
		AND R2, R9, R6, LSR #8
		AND R3, R9, R6, LSR #16
		AND R4, R9, R6, LSR #24
		AND R5, R9, R1
		AND R6, R9, R1, LSR #8
		AND R7, R9, R1, LSR #16
		AND R8, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #96]
		EOR R0, R1
		LDR R1, [R12, #100]
		EOR R5, R1
		LDR R1, [R12, #108]
		EOR R2, R1
		LDR R1, [R12, #112]
		EOR R6, R1
		LDR R1, [R12, #120]
		EOR R3, R1
		LDR R1, [R12, #124]
		EOR R7, R1
		LDR R1, [R12, #132]
		EOR R4, R1
		LDR R1, [R12, #136]
		EOR R8, R1
		/*----eor----*/
		EOR R5, R0
		EOR R6, R2
		EOR R7, R3
		EOR R8, R4
		/*----part A again----*/
		LDRB R0, [R11, #26]
		LDRB R1, [R11, #29]
		LDRB R2, [R11, #32]
		LDRB R3, [R11, #35]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		ORR R0, R0, R3, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec3
		/*----multiply by A again----*/
		AND R0, R9, R1
		AND R2, R9, R1, LSR #8
		AND R3, R9, R1, LSR #16
		AND R4, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #104]
		EOR R0, R1
		LDR R1, [R12, #116]
		EOR R2, R1
		LDR R1, [R12, #128]
		EOR R3, R1
		LDR R1, [R12, #140]
		EOR R4, R1
		/*----eor and store----*/
		EOR R0, R5
		EOR R2, R6
		EOR R3, R7
		EOR R4, R8
		STR R0, [R10, #24]
		STR R2, [R10, #27]
		STR R3, [R10, #30]
		STR R4, [R10, #33]
		
		/*-------the 1/4.4 part--------*/
		// for [12, 13, 14, 15]
		/*----Part A----*/
		LDRH R2, [R11, #36]
		LDRH R3, [R11, #39]
		LDRH R4, [R11, #42]
		LDRH R5, [R11, #45]
		AND R0, R9, R2
		AND R1, R9, R3
		ORR R0, R0, R1, LSL #8
		AND R1, R9, R4
		ORR R0, R0, R1, LSL #16
		AND R1, R9, R5
		ORR R0, R0, R1, LSL #24
		BL mul_mat_vec_4
		MOV R6, R1
		// r6 stores the vec1
		AND R2, R9, R2, LSR #8
		AND R3, R9, R3, LSR #8
		AND R4, R9, R4, LSR #8
		AND R5, R9, R5, LSR #8
		MOV R0, R2
		ORR R0, R0, R3, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec2
		/*----multiply by A----*/
		AND R0, R9, R6
		AND R2, R9, R6, LSR #8
		AND R3, R9, R6, LSR #16
		AND R4, R9, R6, LSR #24
		AND R5, R9, R1
		AND R6, R9, R1, LSR #8
		AND R7, R9, R1, LSR #16
		AND R8, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #144]
		EOR R0, R1
		LDR R1, [R12, #148]
		EOR R5, R1
		LDR R1, [R12, #156]
		EOR R2, R1
		LDR R1, [R12, #160]
		EOR R6, R1
		LDR R1, [R12, #168]
		EOR R3, R1
		LDR R1, [R12, #172]
		EOR R7, R1
		LDR R1, [R12, #180]
		EOR R4, R1
		LDR R1, [R12, #184]
		EOR R8, R1
		/*----eor----*/
		EOR R5, R0
		EOR R6, R2
		EOR R7, R3
		EOR R8, R4
		/*----part A again----*/
		LDRB R0, [R11, #38]
		LDRB R1, [R11, #41]
		LDRB R2, [R11, #44]
		LDRB R3, [R11, #47]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		ORR R0, R0, R3, LSL #24
		BL mul_mat_vec_4
		// r1 stores the vec3
		/*----multiply by A again----*/
		AND R0, R9, R1
		AND R2, R9, R1, LSR #8
		AND R3, R9, R1, LSR #16
		AND R4, R9, R1, LSR #24
		/*----refresh with R2----*/
		LDR R1, [R12, #152]
		EOR R0, R1
		LDR R1, [R12, #164]
		EOR R2, R1
		LDR R1, [R12, #176]
		EOR R3, R1
		LDR R1, [R12, #188]
		EOR R4, R1
		/*----eor and store----*/
		EOR R0, R5
		EOR R2, R6
		EOR R3, R7
		EOR R4, R8
		STR R0, [R10, #36]
		STR R2, [R10, #39]
		STR R3, [R10, #42]
		STRH R4, [R10, #45]
		LSR R4, R4, #16
		STRB R4, [R10, #47]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	