/****************************************************************************/
/* @file   gadget_CodeMuL.s                                                 */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of gadget CodeMuL proposed in XXXXXXXXXXXX.       */
/*         This function calculates the byte_wise multiplication of two     */
/*		   vectors over F_{2^8}^{3} in the masked domain. And the generator */
/*         matrix is fixed as [1,0,0;1,1,0;1,0,1].                          */
/****************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeMuL
	
Gadget_CodeMuL:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src1 */
		/* r1 stores the address of src2 */
		/* r2 stores the address of result */
		
		/* fix some Regs */
		LDR R3, =bhtable // for field multiplication
		LDR R4, =bltable
		LDR R11, =random1 // store R1
		LDR R12, =random2 // store r2
		LDR R1, [R1] // r1 stores the vec2
		LDR R0, [R0] // r0 stores the vec1
		MOV R10, 0xFF
		
		/*-------The 0th row of matrix--------*/
		/*--outer product--*/
		// prepare the byte
		AND R5, R0, R10 
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R1, R10
		ORR R7, R5, R6, LSR #4
		LDRB R7, [R3, R7]
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R7, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		ORR R7, R7, R8, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		ORR R7, R7, R8, LSL #16
		// r7 stores the vec : (1, 3)
		
		/*--refresh with R1--*/
		LDR R5, [R11]
		EOR R7, R5
		// r7 stores the vec : (1, 3)
		
		/*--multiply by M0--*/
		EOR R7, R7, R7, LSR #8
		EOR R7, R7, R7, LSR #16
		// r7 stores the vec : (1, 1)
		
		/*--multiply by A--*/
		AND R7, #0xFF
		// r7 stores the vec : (1, 3)
		
		/*--refresh with R2--*/
		LDR R5, [R12]
		EOR R7, R5
		// r7 stores the vec : (1, 3)
		
		
		/*-------The 1th row of matrix--------*/
		/*--outer product--*/
		// prepare the byte
		AND R5, R10, R0, LSR #8 
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R1, R10
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		ORR R8, R8, R9, LSL #8
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R8, R6, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		ORR R8, R8, R9, LSR #16
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R8, R6, LSL #16
		// r8 stores the vec : (1, 3)
		
		/*--refresh with R1--*/
		LDR R5, [R11, #4]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		/*--multiply by M1--*/
		EOR R8, R8, R8, LSR #8
		EOR R8, R8, R8, LSR #16
		// r8 stores the vec : (1, 1)
		
		/*--multiply by A--*/
		AND R8, #0xFF
		// r8 stores the vec : (1, 3)
		
		/*--refresh with R2--*/
		LDR R5, [R12, #4]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		/*---Eor and store---*/
		EOR R7, R8
		
		/*-------The 2th row of matrix--------*/
		/*--outer product--*/
		// prepare the byte
		AND R5, R10, R0, LSR #16
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R1, R10
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		ORR R8, R8, R9, LSL #8
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R8, R6, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		ORR R8, R8, R9, LSR #16
		AND R6, 0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R8, R6, LSL #16
		// r8 stores the vec : (1, 3)
		
		/*--refresh with R1--*/
		LDR R5, [R11, #8]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		/*--multiply by M2--*/
		EOR R8, R8, R8, LSR #8
		EOR R8, R8, R8, LSR #16
		// r8 stores the vec : (1, 1)
		
		/*--multiply by A--*/
		AND R8, #0xFF
		// r8 stores the vec : (1, 3)
		
		/*--refresh with R2--*/
		LDR R5, [R12, #8]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		/*-------Eor and store--------*/
		EOR R7, R8
		STRH R7, [R2]
		LSR R7, R7, #16
		STRB R7, [R2, #2]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	