/**************************************************************************/
/* @file   aes.s                                                          */
/* @author yangx                                                          */
/* @date   2022-4-11                                                      */
/* @brief  eight functions                                                */
/*         -- AES_Encrypt_GCB [public]                                    */
/*            Implementation of AES-128 masked by code-based masking,     */ 
/*            please refer to xxxxxxx to see the detailed masking         */
/*            scheme. The generator matrix is fixed as                    */
/*            [1,0,0;91,1,0;239,0,1]. This function has no output and the */
/*            input is the address of cyphertext.                         */ 
/*         -- AddRoundKey [private]                                       */
/*            Implementation of AddRoundKey masked by code-based masking. */
/*         -- Sbox [private]                                              */
/*            Implementation of SubBytes masked by code-based masking.    */
/*            Note that S-box can be implemented by an inverse and an     */
/*            affine transformation.                                      */
/*         -- ShiftRows [private]                                         */
/*            Implementation of ShiftRows masked by code-based masking.   */
/*         -- MixColumns [private]                                        */
/*            Implementation of MixColumns masked by code-based masking.  */
/*         -- three functions for constructing R2                         */
/*            --- GetR21 for addRoundKey                                  */
/*            --- GetR22 for shiftRows                                    */
/*            --- GetR23 for mixColumns                                   */
/**************************************************************************/

.section .data
        temp: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	 code_pt: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	pt_state: .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
		      .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
			  .byte 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	  temp_2: .byte 0x00, 0x00, 0x00
	  temp_4: .byte 0x00, 0x00, 0x00
	 temp_16: .byte 0x00, 0x00, 0x00
	  temp_n: .byte 0x00, 0x00, 0x00
			
.section .text
.syntax unified
.thumb
.global AES_Encrypt_GCB, Sbox
	
		/*====================================*/
		/*-----function: AES_Encrypt_GCB------*/
		/*====================================*/
AES_Encrypt_GCB:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of result */
		
		/* fix some Regs */
		MOV R12, R0
		// randomSeed stores all the random numbers utlized in AES
		LDR R11, =randomSeed
		
		/* encode the plaintexts */
		LDR R3, =tabGByte1 // for field multiplication with 91
		LDR R4, =tabGByte2 // for field multiplication with 239
		LDR R5, =temp // container for the encoded plaintext and key
		LDR R6, =pt_aes // plaintext
		// 0
		LDRB R0, [R6]
		LDRB R1, [R11]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #1]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5]
		// 1
		LDRB R0, [R6, #1]
		LDRB R1, [R11, #2]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #3]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #3]
		// 2
		LDRB R0, [R6, #2]
		LDRB R1, [R11, #4]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #5]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #6]
		// 3
		LDRB R0, [R6, #3]
		LDRB R1, [R11, #6]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #7]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #9]
		// 4
		LDRB R0, [R6, #4]
		LDRB R1, [R11, #8]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #9]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #12]
		// 5
		LDRB R0, [R6, #5]
		LDRB R1, [R11, #10]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #11]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #15]
		// 6
		LDRB R0, [R6, #6]
		LDRB R1, [R11, #12]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #13]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #18]
		// 7
		LDRB R0, [R6, #7]
		LDRB R1, [R11, #14]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #15]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #21]
		// 8
		LDRB R0, [R6, #8]
		LDRB R1, [R11, #16]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #17]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #24]
		// 9
		LDRB R0, [R6, #9]
		LDRB R1, [R11, #18]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #19]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #27]
		// 10
		LDRB R0, [R6, #10]
		LDRB R1, [R11, #20]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #21]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #30]
		// 11
		LDRB R0, [R6, #11]
		LDRB R1, [R11, #22]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #23]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #33]
		// 12
		LDRB R0, [R6, #12]
		LDRB R1, [R11, #24]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #25]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #36]
		// 13
		LDRB R0, [R6, #13]
		LDRB R1, [R11, #26]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #27]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #39]
		// 14
		LDRB R0, [R6, #14]
		LDRB R1, [R11, #28]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #29]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #42]
		// 15
		LDRB R0, [R6, #15]
		LDRB R1, [R11, #30]
		LDRB R7, [R3, R1]
		EOR R0, R7
		LDRB R2, [R11, #31]
		LDRB R7, [R4, R2]
		EOR R0, R7
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R2, LSL #16
		STR R0, [R5, #45]
		ADD R11, #32 // move the pointer of randomSeed
		
		/* add encoded key to temp */
		LDR R8, =maskedKeys // note that the key has been extended and encoded
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		ADD R8, #48 // move the pointer of maskedKeys
		
		LDR R10, =random1 // container for R1
		LDR R9, =random2 // container for R2 * 16
		LDR R7, =code_pt
		LDR R6, =pt_state
		
		/*--add Round Key--*/
		BL AddRoundKey
		ADD R11, #96 // move the pointer of randomSeed
		
		/*----AES Round----*/
		MOV R4, #0
loopRounds:
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3 // move the address of pt_state
		ADD R7, #3 // move the address of code_pt
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		//ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		//ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72
		// move the pointer of randomSeed
		LDR R6, =pt_state // reset the address of pt_state
		LDR R7, =code_pt // reset the address of code_pt
		
		/*--ShiftRows--*/
		BL ShiftRows
		ADD R11, #96 // move the pointer of randomSeed
		
		/*--MixColumns--*/
		BL MixColumns
		ADD R11, #96 // move the pointer of randomSeed
		
		/* add encoded key of next round to temp */
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		ADD R8, #48 // move the pointer of maskedKeys
		
		/*--AddRoundKey--*/
		BL AddRoundKey
		ADD R11, #96 // move the pointer of randomSeed
		
		ADD R4, #1
		CMP R4, #9
		IT NE 
		BNE loopRounds

		/*----the last round----*/
		/*--SubBytes--*/
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		ADD R6, #3
		ADD R7, #3
		BL Sbox
		ADD R11, #72 // move the pointer of randomSeed
		LDR R6, =pt_state // reset pt_state
		LDR R7, =code_pt // reset code_pt
		
		/*--ShiftRows--*/
		MOV R6, R5
		BL ShiftRows
		ADD R11, #96 // move the pointer of randomSeed
		
		/* add the encoded key of the last round to temp */
		LDR R0, [R8]
		STR R0, [R5, #48]
		LDR R0, [R8, #4]
		STR R0, [R5, #52]
		LDR R0, [R8, #8]
		STR R0, [R5, #56]
		LDR R0, [R8, #12]
		STR R0, [R5, #60]
		LDR R0, [R8, #16]
		STR R0, [R5, #64]
		LDR R0, [R8, #20]
		STR R0, [R5, #68]
		LDR R0, [R8, #24]
		STR R0, [R5, #72]
		LDR R0, [R8, #28]
		STR R0, [R5, #76]
		LDR R0, [R8, #32]
		STR R0, [R5, #80]
		LDR R0, [R8, #36]
		STR R0, [R5, #84]
		LDR R0, [R8, #40]
		STR R0, [R5, #88]
		LDR R0, [R8, #44]
		STR R0, [R5, #92]
		
		/*--AddRoundKey--*/
		LDR R6, =pt_state
		BL AddRoundKey
		
		/* decode */
		LDR R3, =tabGByte1 // for field multiplication with 91
		LDR R4, =tabGByte2 // for field multiplication with 239
		// 0
		LDRB R0, [R6]
		LDRB R1, [R6, #1]
		LDRB R2, [R6, #2]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12]
		// 1
		LDRB R0, [R6, #3]
		LDRB R1, [R6, #4]
		LDRB R2, [R6, #5]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #1]
		// 2
		LDRB R0, [R6, #6]
		LDRB R1, [R6, #7]
		LDRB R2, [R6, #8]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #2]
		// 3
		LDRB R0, [R6, #9]
		LDRB R1, [R6, #10]
		LDRB R2, [R6, #11]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #3]
		// 4
		LDRB R0, [R6, #12]
		LDRB R1, [R6, #13]
		LDRB R2, [R6, #14]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #4]
		// 5
		LDRB R0, [R6, #15]
		LDRB R1, [R6, #16]
		LDRB R2, [R6, #17]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #5]
		// 6
		LDRB R0, [R6, #18]
		LDRB R1, [R6, #19]
		LDRB R2, [R6, #20]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #6]
		// 7
		LDRB R0, [R6, #21]
		LDRB R1, [R6, #22]
		LDRB R2, [R6, #23]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #7]
		// 8
		LDRB R0, [R6, #24]
		LDRB R1, [R6, #25]
		LDRB R2, [R6, #26]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #8]
		// 9
		LDRB R0, [R6, #27]
		LDRB R1, [R6, #28]
		LDRB R2, [R6, #29]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #9]
		// 10
		LDRB R0, [R6, #30]
		LDRB R1, [R6, #31]
		LDRB R2, [R6, #32]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #10]
		// 11
		LDRB R0, [R6, #33]
		LDRB R1, [R6, #34]
		LDRB R2, [R6, #35]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #11]
		// 12
		LDRB R0, [R6, #36]
		LDRB R1, [R6, #37]
		LDRB R2, [R6, #38]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #12]
		// 13
		LDRB R0, [R6, #39]
		LDRB R1, [R6, #40]
		LDRB R2, [R6, #41]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #13]
		// 14
		LDRB R0, [R6, #42]
		LDRB R1, [R6, #43]
		LDRB R2, [R6, #44]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #14]
		// 15
		LDRB R0, [R6, #45]
		LDRB R1, [R6, #46]
		LDRB R2, [R6, #47]
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R0, R1
		EOR R0, R2
		STRB R0, [R12, #15]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-------function: AddRoundKey--------*/
		/*====================================*/
AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R4,R12,LR}
		/* r4 stores the pointer of R9 */
		/* r5 stores the address of src */
		/* r6 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* fix some Regs */
		LDR R0, =tabGByte1 // for field multiplication with 91
		LDR R1, =tabGByte2 // for field multiplication with 239
		// initialize the pointers
		MOV R4, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		BL GetR21
		
		// add round key
		MOV R0, R5
		MOV R1, R6
		BL Gadget_CodeLs_AddRoundKey
	
		/* copy registers value from the stack */
		POP {R0,R1,R4,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*-----------function: Sbox-----------*/
		/*====================================*/
Sbox:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* r6 stores the address of src */
		/* r7 stores the address of dst */
		/* r9 stores the address of random2 */
		/* r10 stores the address of random1 */
		/* r11 stores the address of randomSeed */
		
		/* fix some Regs */
		LDR R3, =tabGByte1 // for field multiplication with 91
		LDR R4, =tabGByte2 // for field multiplication with 239
		LDR R8, =temp_n
		
		// R2 construction
		LDRH R0, [R11]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #2]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #4]
		LDRH R0, [R11, #4]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// x^2
		MOV R0, R6 // src
		LDR R1, =temp_2
		MOV R2, #0 // parameter indicating .^2
		BL Gadget_CodeL
		
		// R1 construction
		LDRB R0, [R11, #6]
		LDRB R1, [R11, #7]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10]
		STRB R0, [R10, #4]
		STRB R1, [R10, #8]
		LDRB R0, [R11, #8]
		LDRB R1, [R11, #9]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #1]
		STRB R0, [R10, #5]
		STRB R1, [R10, #9]
		LDRB R0, [R11, #10]
		LDRB R1, [R11, #11]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #2]
		STRB R0, [R10, #6]
		STRB R1, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #12]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #14]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #4]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// tempn = x * x ^ 2
		MOV R0, R6 // src
		LDR R1, =temp_2
		MOV R2, R8 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #18]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #20]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #22]
		LDRH R0, [R11, #4]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// tempn^4
		MOV R0, R8 // temp_n
		LDR R1, =temp_4
		MOV R2, #1 // parameter indicating .^4
		BL Gadget_CodeL
		
		// R1 construction
		LDRB R0, [R11, #24]
		LDRB R1, [R11, #25]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10]
		STRB R0, [R10, #4]
		STRB R1, [R10, #8]
		LDRB R0, [R11, #26]
		LDRB R1, [R11, #27]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #1]
		STRB R0, [R10, #5]
		STRB R1, [R10, #9]
		LDRB R0, [R11, #28]
		LDRB R1, [R11, #29]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #2]
		STRB R0, [R10, #6]
		STRB R1, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #30]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #32]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #34]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// dst = (temp4) * (tempn)
		LDR R0, =temp_4
		MOV R1, R8 // temp_n
		MOV R2, R7 // dst
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #36]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #38]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #40]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// dst^16
		MOV R0, R7 // dst
		LDR R1, =temp_16
		MOV R2, #2 // parameter indicating .^16
		BL Gadget_CodeL
		
		// R1 construction
		LDRB R0, [R11, #42]
		LDRB R1, [R11, #43]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10]
		STRB R0, [R10, #4]
		STRB R1, [R10, #8]
		LDRB R0, [R11, #44]
		LDRB R1, [R11, #45]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #1]
		STRB R0, [R10, #5]
		STRB R1, [R10, #9]
		LDRB R0, [R11, #46]
		LDRB R1, [R11, #47]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #2]
		STRB R0, [R10, #6]
		STRB R1, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #48]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #50]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #52]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// dst = (temp4) * (temp16)
		LDR R0, =temp_16
		LDR R1, =temp_4
		MOV R2, R6 // src
		BL Gadget_CodeMuL
		
		// R1 construction
		LDRB R0, [R11, #54]
		LDRB R1, [R11, #55]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10]
		STRB R0, [R10, #4]
		STRB R1, [R10, #8]
		LDRB R0, [R11, #56]
		LDRB R1, [R11, #57]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #1]
		STRB R0, [R10, #5]
		STRB R1, [R10, #9]
		LDRB R0, [R11, #58]
		LDRB R1, [R11, #59]
		LDRB R2, [R3, R0]
		LDRB R5, [R4, R1]
		EOR R2, R5
		STRB R2, [R10, #2]
		STRB R0, [R10, #6]
		STRB R1, [R10, #10]
		// R2 construction
		LDRH R0, [R11, #60]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #62]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #64]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// tempn = src * (temp2)
		MOV R0, R6 // src
		LDR R1, =temp_2 
		MOV R2, R8 // temp_n
		BL Gadget_CodeMuL
		
		// R2 construction
		LDRH R0, [R11, #66]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9]
		LDRH R0, [R11, #68]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #16]
		LDRH R0, [R11, #70]
		AND R1, R0, #0xFF
		LSR R2, R0, #8
		LDRB R1, [R3, R1]
		LDRB R2, [R4, R2]
		EOR R1, R2
		ORR R0, R1, R0, LSL #8
		STR R0, [R9, #8]
		// dst = aff(tempn)
		MOV R0, R8 // temp_n
		MOV R1, R7 // dst
		MOV R2, #3 // parameter indicating affine transformation
		BL Gadget_CodeL
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R8,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: ShiftRows---------*/
		/*====================================*/
ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R4,R12,LR}		
		/* r4 stores the pointer of R9 */
		/* r6 stores the address of dst */
		/* r7 stores the address of src */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* fix some Regs */
		LDR R0, =tabGByte1 // for field multiplication with 91
		LDR R1, =tabGByte2 // for field multiplication with 239
		
		/* initialize the pointer */
		MOV R4, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		BL GetR22
		
		// shift rows
		MOV R0, R7
		MOV R1, R6
		BL Gadget_CodeLs_ShiftRows
	
		/* copy registers value from the stack */
		POP {R0,R1,R4,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: MixColumns--------*/
		/*====================================*/
MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R4,R12,LR}
		/* r4 stores the pointer of R9 */
		/* r5 stores the address of dst */
		/* r6 stores the address of src */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of R11 */
		
		/* fix some Regs */
		LDR R0, =tabGByte1 // for field multiplication with 91
		LDR R1, =tabGByte2 // for field multiplication with 239
		
		/* initialize the pointers */
		MOV R4, #0 // pointer of random2
		MOV R12, #0 // pointer of randomSeed
		
		// R2 construction
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		BL GetR23
		
		// mix columns
		MOV R0, R6
		MOV R1, R5
		BL Gadget_CodeLs_MixColumns
	
		/* copy registers value from the stack */
		POP {R0,R1,R4,R12,LR}
		/* return */
		BX LR
		/*====================================*/
		
		
		/*====================================*/
		/*--------function: R2 for add--------*/
		/*====================================*/
GetR21:
		/* copy registers value to the stack */
		PUSH {R5,R6,R7,LR}
		/* r0 stores the tabGByte1 */
		/* r1 stores the tabGByte2 */
		/* r4 stores the pointer of R9 */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of randomSeed */
		
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		
		/* copy registers value from the stack */
		POP {R5,R6,R7,LR}
		/* return */
		BX LR
		/*====================================*/
		
		/*====================================*/
		/*-----function: R2 for shiftRows-----*/
		/*====================================*/
GetR22:
		/* copy registers value to the stack */
		PUSH {R5,R6,R7,LR}
		/* r0 stores the tabGByte1 */
		/* r1 stores the tabGByte2 */
		/* r4 stores the pointer of R9 */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of randomSeed */
		
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		
		/* copy registers value from the stack */
		POP {R5,R6,R7,LR}
		/* return */
		BX LR
		/*====================================*/
		
		/*====================================*/
		/*-----function: R2 for mixColumn-----*/
		/*====================================*/
GetR23:
		/* copy registers value to the stack */
		PUSH {R5,R6,R7,LR}
		/* r0 stores the tabGByte1 */
		/* r1 stores the tabGByte2 */
		/* r4 stores the pointer of R9 */
		/* r9 stores the address of random2 */
		/* r11 stores the address of randomSeed */
		/* r12 stores the pointer of randomSeed */
		
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		LDRH R5, [R11, R12]
		ADD R12, #2
		AND R6, R5, #0xFF
		LSR R7, R5, #8
		LDRB R6, [R0, R6]
		LDRB R7, [R1, R7]
		EOR R6, R7
		ORR R5, R6, R5, LSL #8
		STR R5, [R9, R4]
		ADD R4, #4
		
		/* copy registers value from the stack */
		POP {R5,R6,R7,LR}
		/* return */
		BX LR
		/*====================================*/
		
.end
	