/***************************************************************************************/
/* @file   gadget_CodeLs_AddRoundKey.s                                                 */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1,0,0;91,1,0;239,0,1] and the f function is fixed */
/*		   as addRoundKey, or saying an Eor operation between two vectors defined in   */
/*		   F_{2^8}^{16}.                                                               */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_AddRoundKey
	
Gadget_CodeLs_AddRoundKey:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */
		
		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		MOV R11, #0xFF
		LDR R2, =tabGByte1 // for field multiplication with 91
		LDR R3, =tabGByte2 // for field multiplication with 239
		
		/*-------1/16.1 part--------*/
		// Part A
		LDR R4, [R0]
		LDR R5, [R0, #48]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12]
		EOR R8, R4
		LDR R4, [R12, #4]
		EOR R9, R4
		LDR R4, [R12, #8]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1]
		
		/*-------1/16.2 part--------*/
		// Part A
		LDR R4, [R0, #3]
		LDR R5, [R0, #51]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #12]
		EOR R8, R4
		LDR R4, [R12, #16]
		EOR R9, R4
		LDR R4, [R12, #20]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #3]
		
		/*-------1/16.3 part--------*/
		// Part A
		LDR R4, [R0, #6]
		LDR R5, [R0, #54]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #24]
		EOR R8, R4
		LDR R4, [R12, #28]
		EOR R9, R4
		LDR R4, [R12, #32]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #6]
		
		/*-------1/16.4 part--------*/
		// Part A
		LDR R4, [R0, #9]
		LDR R5, [R0, #57]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #36]
		EOR R8, R4
		LDR R4, [R12, #40]
		EOR R9, R4
		LDR R4, [R12, #44]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #9]
		
		/*-------1/16.5 part--------*/
		// Part A
		LDR R4, [R0, #12]
		LDR R5, [R0, #60]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #48]
		EOR R8, R4
		LDR R4, [R12, #52]
		EOR R9, R4
		LDR R4, [R12, #56]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #12]
		
		/*-------1/16.6 part--------*/
		// Part A
		LDR R4, [R0, #15]
		LDR R5, [R0, #63]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #60]
		EOR R8, R4
		LDR R4, [R12, #64]
		EOR R9, R4
		LDR R4, [R12, #68]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #15]
		
		/*-------1/16.7 part--------*/
		// Part A
		LDR R4, [R0, #18]
		LDR R5, [R0, #66]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #72]
		EOR R8, R4
		LDR R4, [R12, #76]
		EOR R9, R4
		LDR R4, [R12, #80]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #18]
		
		/*-------1/16.8 part--------*/
		// Part A
		LDR R4, [R0, #21]
		LDR R5, [R0, #69]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #84]
		EOR R8, R4
		LDR R4, [R12, #88]
		EOR R9, R4
		LDR R4, [R12, #92]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #21]
		
		/*-------1/16.9 part--------*/
		// Part A
		LDR R4, [R0, #24]
		LDR R5, [R0, #72]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #96]
		EOR R8, R4
		LDR R4, [R12, #100]
		EOR R9, R4
		LDR R4, [R12, #104]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #24]
		
		/*-------1/16.10 part--------*/
		// Part A
		LDR R4, [R0, #27]
		LDR R5, [R0, #75]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #108]
		EOR R8, R4
		LDR R4, [R12, #112]
		EOR R9, R4
		LDR R4, [R12, #116]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #27]
		
		/*-------1/16.11 part--------*/
		// Part A
		LDR R4, [R0, #30]
		LDR R5, [R0, #78]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #120]
		EOR R8, R4
		LDR R4, [R12, #124]
		EOR R9, R4
		LDR R4, [R12, #128]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #30]
		
		/*-------1/16.12 part--------*/
		// Part A
		LDR R4, [R0, #33]
		LDR R5, [R0, #81]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #132]
		EOR R8, R4
		LDR R4, [R12, #136]
		EOR R9, R4
		LDR R4, [R12, #140]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #33]
		
		/*-------1/16.13 part--------*/
		// Part A
		LDR R4, [R0, #36]
		LDR R5, [R0, #84]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #144]
		EOR R8, R4
		LDR R4, [R12, #148]
		EOR R9, R4
		LDR R4, [R12, #152]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #36]
		
		/*-------1/16.14 part--------*/
		// Part A
		LDR R4, [R0, #39]
		LDR R5, [R0, #87]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #156]
		EOR R8, R4
		LDR R4, [R12, #160]
		EOR R9, R4
		LDR R4, [R12, #164]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #39]
		
		/*-------1/16.15 part--------*/
		// Part A
		LDR R4, [R0, #42]
		LDR R5, [R0, #90]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #168]
		EOR R8, R4
		LDR R4, [R12, #172]
		EOR R9, R4
		LDR R4, [R12, #176]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STR R8, [R1, #42]
	
		/*-------1/16.16 part--------*/
		// Part A
		LDR R4, [R0, #45]
		LDR R5, [R0, #93]
		// Multiply by Ms
		EOR R8, R4, R5
		AND R8, R11
		AND R6, R11, R4, LSR #8
		AND R7, R11, R5, LSR #8
		LDRB R6, [R2, R6]
		LDRB R7, [R2, R7]
		EOR R9, R6, R7
		AND R6, R11, R4, LSR #16
		AND R7, R11, R5, LSR #16
		LDRB R6, [R3, R6]
		LDRB R7, [R3, R7]
		EOR R10, R6, R7
		// Refresh with R2
		LDR R4, [R12, #180]
		EOR R8, R4
		LDR R4, [R12, #184]
		EOR R9, R4
		LDR R4, [R12, #188]
		EOR R10, R4
		// Eor and Store
		EOR R8, R9
		EOR R8, R10
		STRH R8, [R1, #45]
		LSR R8, R8, #16
		STRB R8, [R1, #47]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	