/***************************************************************************************/
/* @file   gadget_CodeLs_MixColumns.s                                                  */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1,0,0;91,1,0;239,0,1] and the f function is fixed */
/*		   as mixColumns of AES-128.                                                   */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_MixColumns
	
Gadget_CodeLs_MixColumns:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of dst */
		
		/* fix some Regs */
		MOV R11, R0
		MOV R10, R1
		LDR R2, =tabGByte1 // for field multiplication with 91
		LDR R12, =random2 // store R2 * 16
		
		/*-------the 1/4.1 part--------*/
		// for [0, 1, 2, 3]
		// Part A
		LDRB R0, [R11]
		LDRB R1, [R11, #3]
		LDRB R4, [R11, #6]
		LDRB R5, [R11, #9]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		MOV R7, R1
		// r6 stores the res
		LDRB R0, [R11, #1]
		LDRB R1, [R11, #4]
		LDRB R4, [R11, #7]
		LDRB R5, [R11, #10]
		LDRB R0, [R2, R0]
		LDRB R1, [R2, R1]
		LDRB R4, [R2, R4]
		LDRB R5, [R2, R5]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A 
		MOV R9, #0xFF
		AND R0, R9, R7
		AND R4, R9, R1
		AND R5, R9, R7, LSR #8
		AND R6, R9, R1, LSR #8
		AND R8, R9, R7, LSR #16
		LSR R9, R1, #16
		AND R9, #0xFF
		LSR R7, R7, #24
		LSR R1, R1, #24
		// Refresh with R2
		LDR R3, [R12]
		EOR R0, R3
		LDR R3, [R12, #4]
		EOR R4, R3
		LDR R3, [R12, #12]
		EOR R5, R3
		LDR R3, [R12, #16]
		EOR R6, R3
		LDR R3, [R12, #24]
		EOR R8, R3
		LDR R3, [R12, #28]
		EOR R9, R3
		LDR R3, [R12, #36]
		EOR R7, R3
		LDR R3, [R12, #40]
		EOR R1, R3
		// Eor
		EOR R4, R0
		EOR R5, R6
		EOR R6, R8, R9
		EOR R7, R1
		// Part A again
		LDR R3, =tabGByte2 // for field multiplication with 239
		LDRB R0, [R11, #2]
		LDRB R1, [R11, #5]
		LDRB R8, [R11, #8]
		LDRB R9, [R11, #11]
		LDRB R0, [R3, R0]
		LDRB R1, [R3, R1]
		LDRB R8, [R3, R8]
		LDRB R9, [R3, R9]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R8, LSL #16
		ORR R0, R0, R9, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A again
		MOV R3, #0xFF
		AND R0, R3, R1
		AND R8, R3, R1, LSR #8
		AND R9, R3, R1, LSR #16
		LSR R1, R1, #24
		// Refresh with R2 again
		LDR R3, [R12, #8]
		EOR R0, R3
		LDR R3, [R12, #20]
		EOR R8, R3
		LDR R3, [R12, #32]
		EOR R9, R3
		LDR R3, [R12, #44]
		EOR R1, R3
		// Eor and Store
		EOR R4, R0
		EOR R5, R8
		EOR R6, R9
		EOR R7, R1
		STR R4, [R10]
		STR R5, [R10, #3]
		STR R6, [R10, #6]
		STR R7, [R10, #9]
		
		/*-------the 1/4.2 part--------*/
		// for [4, 5, 6, 7]
		// Part A
		LDRB R0, [R11, #12]
		LDRB R1, [R11, #15]
		LDRB R4, [R11, #18]
		LDRB R5, [R11, #21]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		MOV R7, R1
		// r6 stores the res
		LDRB R0, [R11, #13]
		LDRB R1, [R11, #16]
		LDRB R4, [R11, #19]
		LDRB R5, [R11, #22]
		LDRB R0, [R2, R0]
		LDRB R1, [R2, R1]
		LDRB R4, [R2, R4]
		LDRB R5, [R2, R5]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A 
		MOV R9, #0xFF
		AND R0, R9, R7
		AND R4, R9, R1
		AND R5, R9, R7, LSR #8
		AND R6, R9, R1, LSR #8
		AND R8, R9, R7, LSR #16
		LSR R9, R1, #16
		AND R9, #0xFF
		LSR R7, R7, #24
		LSR R1, R1, #24
		// Refresh with R2
		LDR R3, [R12, #48]
		EOR R0, R3
		LDR R3, [R12, #52]
		EOR R4, R3
		LDR R3, [R12, #60]
		EOR R5, R3
		LDR R3, [R12, #64]
		EOR R6, R3
		LDR R3, [R12, #72]
		EOR R8, R3
		LDR R3, [R12, #76]
		EOR R9, R3
		LDR R3, [R12, #84]
		EOR R7, R3
		LDR R3, [R12, #88]
		EOR R1, R3
		// Eor
		EOR R4, R0
		EOR R5, R6
		EOR R6, R8, R9
		EOR R7, R1
		// part A again
		LDR R3, =tabGByte2
		LDRB R0, [R11, #14]
		LDRB R1, [R11, #17]
		LDRB R8, [R11, #20]
		LDRB R9, [R11, #23]
		LDRB R0, [R3, R0]
		LDRB R1, [R3, R1]
		LDRB R8, [R3, R8]
		LDRB R9, [R3, R9]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R8, LSL #16
		ORR R0, R0, R9, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A again
		MOV R3, #0xFF
		AND R0, R3, R1
		AND R8, R3, R1, LSR #8
		AND R9, R3, R1, LSR #16
		LSR R1, R1, #24
		// Refresh with R2 again
		LDR R3, [R12, #56]
		EOR R0, R3
		LDR R3, [R12, #68]
		EOR R8, R3
		LDR R3, [R12, #80]
		EOR R9, R3
		LDR R3, [R12, #92]
		EOR R1, R3
		// Eor and Store
		EOR R4, R0
		EOR R5, R8
		EOR R6, R9
		EOR R7, R1
		STR R4, [R10, #12]
		STR R5, [R10, #15]
		STR R6, [R10, #18]
		STR R7, [R10, #21]
		
		/*-------the 1/4.3 part--------*/
		// for [8, 9, 10, 11]
		// Part A
		LDRB R0, [R11, #24]
		LDRB R1, [R11, #27]
		LDRB R4, [R11, #30]
		LDRB R5, [R11, #33]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		MOV R7, R1
		// r6 stores the res
		LDRB R0, [R11, #25]
		LDRB R1, [R11, #28]
		LDRB R4, [R11, #31]
		LDRB R5, [R11, #34]
		LDRB R0, [R2, R0]
		LDRB R1, [R2, R1]
		LDRB R4, [R2, R4]
		LDRB R5, [R2, R5]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A 
		MOV R9, #0xFF
		AND R0, R9, R7
		AND R4, R9, R1
		AND R5, R9, R7, LSR #8
		AND R6, R9, R1, LSR #8
		AND R8, R9, R7, LSR #16
		LSR R9, R1, #16
		AND R9, #0xFF
		LSR R7, R7, #24
		LSR R1, R1, #24
		// Refresh with R2
		LDR R3, [R12, #96]
		EOR R0, R3
		LDR R3, [R12, #100]
		EOR R4, R3
		LDR R3, [R12, #108]
		EOR R5, R3
		LDR R3, [R12, #112]
		EOR R6, R3
		LDR R3, [R12, #120]
		EOR R8, R3
		LDR R3, [R12, #124]
		EOR R9, R3
		LDR R3, [R12, #132]
		EOR R7, R3
		LDR R3, [R12, #136]
		EOR R1, R3
		// Eor
		EOR R4, R0
		EOR R5, R6
		EOR R6, R8, R9
		EOR R7, R1
		// Part A again
		LDR R3, =tabGByte2
		LDRB R0, [R11, #26]
		LDRB R1, [R11, #29]
		LDRB R8, [R11, #32]
		LDRB R9, [R11, #35]
		LDRB R0, [R3, R0]
		LDRB R1, [R3, R1]
		LDRB R8, [R3, R8]
		LDRB R9, [R3, R9]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R8, LSL #16
		ORR R0, R0, R9, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A again
		MOV R3, #0xFF
		AND R0, R3, R1
		AND R8, R3, R1, LSR #8
		AND R9, R3, R1, LSR #16
		LSR R1, R1, #24
		// Refresh with R2 again
		LDR R3, [R12, #104]
		EOR R0, R3
		LDR R3, [R12, #116]
		EOR R8, R3
		LDR R3, [R12, #128]
		EOR R9, R3
		LDR R3, [R12, #140]
		EOR R1, R3
		// Eor and Store
		EOR R4, R0
		EOR R5, R8
		EOR R6, R9
		EOR R7, R1
		STR R4, [R10, #24]
		STR R5, [R10, #27]
		STR R6, [R10, #30]
		STR R7, [R10, #33]
		
		/*-------the 1/4.4 part--------*/
		// for [12, 13, 14, 15]
		// Part A
		LDRB R0, [R11, #36]
		LDRB R1, [R11, #39]
		LDRB R4, [R11, #42]
		LDRB R5, [R11, #45]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		MOV R7, R1
		// r6 stores the res
		LDRB R0, [R11, #37]
		LDRB R1, [R11, #40]
		LDRB R4, [R11, #43]
		LDRB R5, [R11, #46]
		LDRB R0, [R2, R0]
		LDRB R1, [R2, R1]
		LDRB R4, [R2, R4]
		LDRB R5, [R2, R5]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R4, LSL #16
		ORR R0, R0, R5, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A 
		MOV R9, #0xFF
		AND R0, R9, R7
		AND R4, R9, R1
		AND R5, R9, R7, LSR #8
		AND R6, R9, R1, LSR #8
		AND R8, R9, R7, LSR #16
		LSR R9, R1, #16
		AND R9, #0xFF
		LSR R7, R7, #24
		LSR R1, R1, #24
		// Refresh with R2
		LDR R3, [R12, #144]
		EOR R0, R3
		LDR R3, [R12, #148]
		EOR R4, R3
		LDR R3, [R12, #156]
		EOR R5, R3
		LDR R3, [R12, #160]
		EOR R6, R3
		LDR R3, [R12, #168]
		EOR R8, R3
		LDR R3, [R12, #172]
		EOR R9, R3
		LDR R3, [R12, #180]
		EOR R7, R3
		LDR R3, [R12, #184]
		EOR R1, R3
		// Eor
		EOR R4, R0
		EOR R5, R6
		EOR R6, R8, R9
		EOR R7, R1
		// Part A again
		LDR R3, =tabGByte2
		LDRB R0, [R11, #38]
		LDRB R1, [R11, #41]
		LDRB R8, [R11, #44]
		LDRB R9, [R11, #47]
		LDRB R0, [R3, R0]
		LDRB R1, [R3, R1]
		LDRB R8, [R3, R8]
		LDRB R9, [R3, R9]
		ORR R0, R0, R1, LSL #8
		ORR R0, R0, R8, LSL #16
		ORR R0, R0, R9, LSL #24
		BL mul_mat_vec_4
		// r1 stores the res
		// Multiply by A again
		MOV R3, #0xFF
		AND R0, R3, R1
		AND R8, R3, R1, LSR #8
		AND R9, R3, R1, LSR #16
		LSR R1, R1, #24
		// Refresh with R2 again
		LDR R3, [R12, #152]
		EOR R0, R3
		LDR R3, [R12, #164]
		EOR R8, R3
		LDR R3, [R12, #176]
		EOR R9, R3
		LDR R3, [R12, #188]
		EOR R1, R3
		// Eor and Store
		EOR R4, R0
		EOR R5, R8
		EOR R6, R9
		EOR R7, R1
		STR R4, [R10, #36]
		STR R5, [R10, #39]
		STR R6, [R10, #42]
		STRH R7, [R10, #45]
		LSR R7, R7, #16
		STRB R7, [R10, #47]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	