/***************************************************************************************/
/* @file   gadget_CodeLs_ShiftRows.s                                                   */
/* @author yangx                                                                       */
/* @date   2022.4.11                                                                   */
/* @brief  Implementation of Gadget CodeLs proposed in XXXXXX. This gadget is tailored */
/*         for the generator matrix [1,0,0;91,1,0;239,0,1] and the f function is fixed */
/*		   as shiftrows of AES-128.                                                    */
/***************************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs_ShiftRows
	
Gadget_CodeLs_ShiftRows:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R11,R12,LR}
		/* r0 stores the adress of src */
		/* r1 stores the adress of dst */
		
		/* fix some Regs */
		LDR R12, =random2 // store R2 * 16
		MOV R11, #0xFF
		LDR R2, =tabGByte1 // for field multiplication with 91
		LDR R3, =tabGByte2 // for field multiplication with 239
		
		/*-------1/16.1 part--------*/
		// for [0]
		// Part A
		LDR R4, [R0]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [0]
		LDR R7, [R12]
		EOR R4, R7
		LDR R7, [R12, #4]
		EOR R5, R7
		LDR R7, [R12, #8]
		EOR R6, R7
		// Eor and Store
		// store in [0]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1]
		
		/*-------1/16.2 part--------*/
		// for [1]
		// Part A
		LDR R4, [R0, #3]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [13]
		LDR R7, [R12, #156]
		EOR R4, R7
		LDR R7, [R12, #160]
		EOR R5, R7
		LDR R7, [R12, #164]
		EOR R6, R7
		// Eor and Store
		// store in [13]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #39]
		
		/*-------1/16.3 part--------*/
		// for [2]
		// Part A
		LDR R4, [R0, #6]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [10]
		LDR R7, [R12, #120]
		EOR R4, R7
		LDR R7, [R12, #124]
		EOR R5, R7
		LDR R7, [R12, #128]
		EOR R6, R7
		// Eor and Store
		// store in [10]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #30]
		
		/*-------1/16.4 part--------*/
		// for [3]
		// Part A
		LDR R4, [R0, #9]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [7]
		LDR R7, [R12, #84]
		EOR R4, R7
		LDR R7, [R12, #88]
		EOR R5, R7
		LDR R7, [R12, #92]
		EOR R6, R7
		// Eor and Store
		// store in [7]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #21]
		
		/*-------1/16.5 part--------*/
		// for [4]
		// Part A
		LDR R4, [R0, #12]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [4]
		LDR R7, [R12, #48]
		EOR R4, R7
		LDR R7, [R12, #52]
		EOR R5, R7
		LDR R7, [R12, #56]
		EOR R6, R7
		// Eor and Store
		// store in [4]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #12]
		
		/*-------1/16.6 part--------*/
		// for [5]
		// Part A
		LDR R4, [R0, #15]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [1]
		LDR R7, [R12, #12]
		EOR R4, R7
		LDR R7, [R12, #16]
		EOR R5, R7
		LDR R7, [R12, #20]
		EOR R6, R7
		// Eor and Store
		// store in [1]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #3]
		
		/*-------1/16.7 part--------*/
		// for [6]
		// Part A
		LDR R4, [R0, #18]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [14]
		LDR R7, [R12, #168]
		EOR R4, R7
		LDR R7, [R12, #172]
		EOR R5, R7
		LDR R7, [R12, #176]
		EOR R6, R7
		// Eor and Store
		// store in [14]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #42]
		
		/*-------1/16.8 part--------*/
		// for [7]
		// Part A
		LDR R4, [R0, #21]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [11]
		LDR R7, [R12, #132]
		EOR R4, R7
		LDR R7, [R12, #136]
		EOR R5, R7
		LDR R7, [R12, #140]
		EOR R6, R7
		// Eor and Store
		// store in [11]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #33]
		
		/*-------1/16.9 part--------*/
		// for [8]
		// Part A
		LDR R4, [R0, #24]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [8]
		LDR R7, [R12, #96]
		EOR R4, R7
		LDR R7, [R12, #100]
		EOR R5, R7
		LDR R7, [R12, #104]
		EOR R6, R7
		// Eor and Store
		// store in [8]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #24]
		
		/*-------1/16.10 part--------*/
		// for [9]
		// Part A
		LDR R4, [R0, #27]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [5]
		LDR R7, [R12, #60]
		EOR R4, R7
		LDR R7, [R12, #64]
		EOR R5, R7
		LDR R7, [R12, #68]
		EOR R6, R7
		// Eor and Store
		// store in [5]
		EOR R4, R5
		EOR R4, R6
		STR R4, [R1, #15]
		
		/*-------1/16.11 part--------*/
		// for [10]
		// Part A
		LDR R4, [R0, #30]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [2]
		LDR R7, [R12, #24]
		EOR R4, R7
		LDR R7, [R12, #28]
		EOR R5, R7
		LDR R7, [R12, #32]
		EOR R6, R7
		// Eor and Store
		// store in [2]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #6]
		LSR R4, R4, #16
		STRB R4, [R1, #8]
	
		/*-------1/16.12 part--------*/
		// for [11]
		// Part A
		LDR R4, [R0, #33]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [15]
		LDR R7, [R12, #180]
		EOR R4, R7
		LDR R7, [R12, #184]
		EOR R5, R7
		LDR R7, [R12, #188]
		EOR R6, R7
		// Eor and Store
		// store in [15]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #45]
		LSR R4, R4, #16
		STRB R4, [R1, #47]
		
		/*-------1/16.13 part--------*/
		// for [12]
		// Part A
		LDR R4, [R0, #36]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [12]
		LDR R7, [R12, #144]
		EOR R4, R7
		LDR R7, [R12, #148]
		EOR R5, R7
		LDR R7, [R12, #152]
		EOR R6, R7
		// Eor and Store
		// store in [12]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #36]
		LSR R4, R4, #16
		STRB R4, [R1, #38]
		
		/*-------1/16.14 part--------*/
		// for [13]
		// Part A
		LDR R4, [R0, #39]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [9]
		LDR R7, [R12, #108]
		EOR R4, R7
		LDR R7, [R12, #112]
		EOR R5, R7
		LDR R7, [R12, #116]
		EOR R6, R7
		// Eor and Store
		// store in [9]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #27]
		LSR R4, R4, #16
		STRB R4, [R1, #29]
		
		/*-------1/16.15 part--------*/
		// for [14]
		// Part A
		LDR R4, [R0, #42]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [6]
		LDR R7, [R12, #72]
		EOR R4, R7
		LDR R7, [R12, #76]
		EOR R5, R7
		LDR R7, [R12, #80]
		EOR R6, R7
		// Eor and Store
		// store in [6]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #18]
		LSR R4, R4, #16
		STRB R4, [R1, #20]
		
		/*-------1/16.16 part--------*/
		// for [15]
		// Part A
		LDR R4, [R0, #45]
		AND R5, R11, R4, LSR #8
		LDRB R5, [R2, R5]
		AND R6, R11, R4, LSR #16
		LDRB R6, [R3, R6]
		AND R4, R11
		// Refresh with R2
		// changed to [3]
		LDR R7, [R12, #36]
		EOR R4, R7
		LDR R7, [R12, #40]
		EOR R5, R7
		LDR R7, [R12, #44]
		EOR R6, R7
		// Eor and Store
		// store in [3]
		EOR R4, R5
		EOR R4, R6
		STRH R4, [R1, #9]
		LSR R4, R4, #16
		STRB R4, [R1, #11]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R11,R12,LR}
		/* return */
		BX LR
.end
	