/****************************************************************************/
/* @file   gadget_CodeMuL.s                                                 */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of gadget CodeMuL proposed in XXXXXXXXXXXX.       */
/*         This function calculates the byte_wise multiplication of two     */
/*		   vectors over F_{2^8}^{3} in the masked domain. And the generator */
/*         matrix is fixed as [1,0,0;91,1,0;239,0,1].                       */
/****************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeMuL
	
Gadget_CodeMuL:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src1 */
		/* r1 stores the address of src2 */
		/* r2 stores the address of result */
	
		/* fix some Regs */
		LDR R12, =random1 // store R1
		LDR R11, =random2 // store R2
		MOV R10, #0xFF
		LDR R3, =bhtable // for field multiplication
		LDR R4, =bltable
		LDR R0, [R0] // r0 stores the vec1 : (1, 3)
		LDR R1, [R1] // r1 stores the vec2 : (1, 3)
		
		/*-------The 0th row of matrix--------*/
		// Outer product
		AND R5, R10, R0 // prepare the byte 1
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R10, R1 // prepare the byte2
		ORR R7, R5, R6, LSR #4
		LDRB R7, [R3, R7]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R7, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8 // prepare the byte2
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		ORR R7, R7, R8, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16 // prepare the byte2
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		ORR R7, R7, R8, LSL #16
		// r7 stores the vec : (1, 3)
		
		// Refresh with R1
		LDR R5, [R12]
		EOR R7, R5
		// r7 stores the vec : (1, 3)
		
		// Multiply by M0
		LDR R5, =tabGByte1
		AND R8, R10, R7, LSR #8
		LDRB R8, [R5, R8]
		EOR R7, R8
		LDR R6, =tabGByte2
		AND R8, R10, R7, LSR #16
		LDRB R8, [R6, R8]
		EOR R7, R8
		// r7 stores the vec : (1, 1)
		
		// Multiply by A
		AND R7, R10
		// r7 stores the vec : (1, 3)
		
		/*--refresh with R2--*/
		LDR R8, [R11]
		EOR R7, R8
		// r7 stores the vec : (1, 3)
		
		/*-------The 1th row of matrix--------*/
		// Outer Product
		AND R5, R10, R0, LSR #8 // prepare the byte 1
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R10, R1 // prepare the byte2
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8 // prepare the byte2
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R9, R6
		ORR R8, R8, R9, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16 // prepare the byte2
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R9, R6
		ORR R8, R8, R9, LSL #16
		// r8 stores the vec : (1, 3)
		
		// Refresh with R1
		LDR R5, [R12, #4]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		// Multiply by M1
		LDR R5, =tabGByte1
		AND R6, R10, R8
		LDRB R6, [R5, R6]
		LDR R5, =tabMByte1
		AND R9, R10, R8, LSR #8
		LDRB R9, [R5, R9]
		EOR R6, R9
		LDR R5, =tabMByte2
		AND R9, R10, R8, LSR #16
		LDRB R9, [R5, R9]
		EOR R9, R6
		// r9 stores the vec : (1, 1)
		
		// Refresh with R2
		LDR R5, [R11, #4]
		EOR R9, R5
		// r9 stores the vec : (1, 3)
		
		// Eor
		EOR R7, R9
		
		/*-------The 2th row of matrix--------*/
		// Outer Product
		AND R5, R10, R0, LSR #16 // prepare the byte 1
		LSL R5, R5, #4
		// the first byte of vec (1, 3)
		AND R6, R10, R1 // prepare the byte2
		ORR R8, R5, R6, LSR #4
		LDRB R8, [R3, R8]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R8, R6
		// the second byte of vec (1, 3)
		AND R6, R10, R1, LSR #8 // prepare the byte2
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R9, R6
		ORR R8, R8, R9, LSL #8
		// the third byte of vec (1, 3)
		AND R6, R10, R1, LSR #16 // prepare the byte2
		ORR R9, R5, R6, LSR #4
		LDRB R9, [R3, R9]
		AND R6, #0xF
		ORR R6, R5, R6
		LDRB R6, [R4, R6]
		EOR R9, R6
		ORR R8, R8, R9, LSL #16
		// r8 stores the vec : (1, 3)
		
		// Refresh with R2
		LDR R5, [R12, #8]
		EOR R8, R5
		// r8 stores the vec : (1, 3)
		
		// Multiply by M2
		LDR R5, =tabGByte2
		AND R6, R10, R8
		LDRB R6, [R5, R6]
		LDR R5, =tabMByte2
		AND R9, R10, R8, LSR #8
		LDRB R9, [R5, R9]
		EOR R6, R9
		LDR R5, =tabMByte3
		AND R9, R10, R8, LSR #16
		LDRB R9, [R5, R9]
		EOR R9, R6
		// r9 stores the vec : (1, 1)
		
		// Refresh with R2
		LDR R5, [R11, #8]
		EOR R9, R5
		// r9 stores the vec : (1, 3)
		
		// Eor and Store
		EOR R9, R7
		STRH R9, [R2]
		LSR R9, R9, #16
		STRB R9, [R2, #2]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	