/*************************************************************************************************/
/* @file   aes.c                                                                                 */
/* @author yangx                                                                                 */
/* @date   2022-4-25                                                                             */
/* @brief  global tables and some functions for AES-128 algorithm masked by code-based masking.  */
/*         Note that this is a general version, which can be fed with any legal tuple of (k,m,n) */
/*         and generator matrix A. For normal operation, there are some parameters required to   */
/*         be set. And the involved parameters shall be marked by ## in the annotation. Note     */
/*         that the parameters are scattered in the code. For check, there are 33 parameters in  */
/*         total.                                                                                */
/*************************************************************************************************/

#include "common.h"
#include "DWT_clock.h"
#include "random.h"
#include "comm.h"
#include "stdlib.h"

// (k, m, n) = (1, 1, 2)
uint8 matA[] = { // ## generator matrix (k+m)*n A
   1,  91,
   0,   1,
};
uint8 matG[] = { // ## the upper k*n sub-matrix G of A
   1,
   0,
};
uint8 matH[] = { // ## the lower m*n sub-matrix H of A
  91,
   1,
};
uint8 matH1[] = { // ## the left m*(n-m) sub-matrix of H 
	91
};
uint8 matHT[] = { // ## the transpose of H
  91,
   1,
};
uint8 matAInvSub[] = { // ## the first k columns of the inverse of A
   1,  91,
};
uint8 matAInvSubSub[] = { // ## the lower left (n-k)*k sub-matrix of the inverse of A
	 91
};
uint8 matM[] = { // ## n matrices M_i over F^{n * k}
// M_0
   1,  91,
// M_1
  91, 245,
};

// (k, m, n) = (1, 2, 3)
/*uint8 matA[] = {
   1,  91, 239,
   0,   1,   0,
   0,   0,   1,
};
uint8 matH1[] = {
   91, 239
};
uint8 matG[] = {
   1,
   0,
   0,
};
uint8 matH[] = {
  91, 239,
   1,   0,
   0,   1,
};
uint8 matHT[] = {
  91, 239,
   1,   0,
   0,   1,
};
uint8 matAInvSub[] = {
   1,  91, 239,
};
uint8 matAInvSubSub[] = {
   91, 239,
};
uint8 matM[] = {
// M[0]
   1,  91, 239,
// M[1]
  91, 245, 178,
// M[2]
 239, 178,   8,
};*/

// (k , m , n) = (2, 1, 3)
/*uint8 matA[] = {
   1,   0,   1,
   0,   1,   1,
   0,   0,   1,
};
uint8 matG[] = {
   1,   0,
   0,   1,
   0,   0,
};
uint8 matH[] = {
   1,
   1,
   1,
};
uint8 matHT[] = {
   1,
   1,
   1,
};
uint8 matAInvSub[] = {
   1,   0,   1,
   0,   1,   1,
};
uint8 matM[] = {
// M[0]
   1,   0,   1,
   0,   0,   0,
// M[1]
   0,   0,   0,
   0,   1,   1,
// M[2]
   1,   0,   1,
   0,   1,   1,
};
uint8 matH1[] = {
    1, 1
};
uint8 matAInvSubSub[] = {
    1, 1
};*/

// (k , m , n) = (4, 1, 5)
/*uint8 matA[] = {
   1,   0,   0,   0,   1,
   0,   1,   0,   0,   1,
   0,   0,   1,   0,   1,
   0,   0,   0,   1,   1,
   0,   0,   0,   0,   1,
};
uint8 matG[] = {
   1,   0,   0,   0,
   0,   1,   0,   0,
   0,   0,   1,   0,
   0,   0,   0,   1,
   0,   0,   0,   0,
};
uint8 matH[] = {
   1,
   1,
   1,
   1,
   1,
};
uint8 matHT[] = {
   1,
   1,
   1,
   1,
   1,
};
uint8 matAInvSub[] = {
   1,   0,   0,   0,   1,
   0,   1,   0,   0,   1,
   0,   0,   1,   0,   1,
   0,   0,   0,   1,   1,
};
uint8 matM[] = {
// M[0]
   1,   0,   0,   0,   1,
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
// M[1]
   0,   0,   0,   0,   0,
   0,   1,   0,   0,   1,
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
// M[2]
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
   0,   0,   1,   0,   1,
   0,   0,   0,   0,   0,
// M[3]
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,
   0,   0,   0,   1,   1,
// M[4]
   1,   0,   0,   0,   1,
   0,   1,   0,   0,   1,
   0,   0,   1,   0,   1,
   0,   0,   0,   1,   1,
};
uint8 matH1[] = {
    1, 1, 1, 1
};
uint8 matAInvSubSub[] = {
    1, 1,
    1, 1
};*/

// (k , m , n) = (1, 1, 3)
/*uint8 matA[] = {
   1,  23,
   0,  29,
   0,   1,
};
uint8 matG[] = {
   1,
   0,
   0
};
uint8 matH[] = {
  23,
  29,
   1
};
uint8 matHT[] = {
  23,
  29,
   1
};
uint8 matAInvSub[] = {
   1, 240,  82
};
uint8 matAInvSubSub[] = {
   240, 82
};
uint8 matM[] = {
// M[0]
   1, 240,  82,
// M[1]
 240,  70,  48,
// M[2]
  82,  48, 180
};
uint8 matH1[] = {
  23,
  29
};*/

// (k , m , n) = (1, 1, 4)
/*uint8 matA[] = {
   1,  23,
   0,  29,
   0,  51,
   0,   1,
};
uint8 matG[] = {
   1,
   0,
   0,
   0,
};
uint8 matH[] = {
  23,
  29,
  51,
   1,
};
uint8 matHT[] = {
  23,
  29,
  51,
   1,
};
uint8 matAInvSub[] = {
   1,  11,  17, 246,
};
uint8 matAInvSubSub[] = {
   11,  17, 246
};
uint8 matM[] = {
// M[0]
   1,  11,  17, 246,
// M[1]
  11,  69, 187, 240,
// M[2]
  17, 187,  26,  15,
// M[3]
 246, 240,  15,  82,
};
uint8 matH1[] = {
  23,
  29,
  51
};*/

// (k , m , n) = (2, 4, 6)
/*uint8 matA[] = {
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
   0,   0,   1,   0,   0,   0,
   0,   0,   0,   1,   0,   0,
   0,   0,   0,   0,   1,   0,
   0,   0,   0,   0,   0,   1,
};
uint8 matG[] = {
   1,   0,
   0,   1,
   0,   0,
   0,   0,
   0,   0,
   0,   0,
};
uint8 matH[] = {
   1,   1,   1,   1,
   1,   1,   1,   1,
   1,   0,   0,   0,
   0,   1,   0,   0,
   0,   0,   1,   0,
   0,   0,   0,   1,
};
uint8 matHT[] = {
   1,   1,   1,   1,
   1,   1,   1,   1,
   1,   0,   0,   0,
   0,   1,   0,   0,
   0,   0,   1,   0,
   0,   0,   0,   1,
};
uint8 matAInvSub[] = {
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
};
uint8 matM[] = {
// M[0]
   1,   0,   1,   1,   1,   1,
   0,   0,   0,   0,   0,   0,
// M[1]
   0,   0,   0,   0,   0,   0,
   0,   1,   1,   1,   1,   1,
// M[2]
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
// M[3]
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
// M[4]
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
// M[5]
   1,   0,   1,   1,   1,   1,
   0,   1,   1,   1,   1,   1,
};
uint8 matH1[] = {
    1, 1, 1, 1,
    1, 1, 1, 1
};
uint8 matAInvSubSub[] = {
    1, 1,
    1, 1,
    1, 1,
    1, 1
};*/

// tables
uint8 bhtable[] = {
	0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 
  0,  16,  32,  48,  64,  80,  96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 
  0,  32,  64,  96, 128, 160, 192, 224,  27,  59,  91, 123, 155, 187, 219, 251, 
  0,  48,  96,  80, 192, 240, 160, 144, 155, 171, 251, 203,  91, 107,  59,  11, 
  0,  64, 128, 192,  27,  91, 155, 219,  54, 118, 182, 246,  45, 109, 173, 237, 
  0,  80, 160, 240,  91,  11, 251, 171, 182, 230,  22,  70, 237, 189,  77,  29, 
  0,  96, 192, 160, 155, 251,  91,  59,  45,  77, 237, 141, 182, 214, 118,  22, 
  0, 112, 224, 144, 219, 171,  59,  75, 173, 221,  77,  61, 118,   6, 150, 230, 
  0, 128,  27, 155,  54, 182,  45, 173, 108, 236, 119, 247,  90, 218,  65, 193, 
  0, 144,  59, 171, 118, 230,  77, 221, 236, 124, 215,  71, 154,  10, 161,  49, 
  0, 160,  91, 251, 182,  22, 237,  77, 119, 215,  44, 140, 193,  97, 154,  58, 
  0, 176, 123, 203, 246,  70, 141,  61, 247,  71, 140,  60,   1, 177, 122, 202, 
  0, 192, 155,  91,  45, 237, 182, 118,  90, 154, 193,   1, 119, 183, 236,  44, 
  0, 208, 187, 107, 109, 189, 214,   6, 218,  10,  97, 177, 183, 103,  12, 220, 
  0, 224, 219,  59, 173,  77, 118, 150,  65, 161, 154, 122, 236,  12,  55, 215, 
  0, 240, 251,  11, 237,  29,  22, 230, 193,  49,  58, 202,  44, 220, 215,  39, 
  0,  27,  54,  45, 108, 119,  90,  65, 216, 195, 238, 245, 180, 175, 130, 153, 
  0,  11,  22,  29,  44,  39,  58,  49,  88,  83,  78,  69, 116, 127,  98, 105, 
  0,  59, 118,  77, 236, 215, 154, 161, 195, 248, 181, 142,  47,  20,  89,  98, 
  0,  43,  86, 125, 172, 135, 250, 209,  67, 104,  21,  62, 239, 196, 185, 146, 
  0,  91, 182, 237, 119,  44, 193, 154, 238, 181,  88,   3, 153, 194,  47, 116, 
  0,  75, 150, 221,  55, 124, 161, 234, 110,  37, 248, 179,  89,  18, 207, 132, 
  0, 123, 246, 141, 247, 140,   1, 122, 245, 142,   3, 120,   2, 121, 244, 143, 
  0, 107, 214, 189, 183, 220,  97,  10, 117,  30, 163, 200, 194, 169,  20, 127, 
  0, 155,  45, 182,  90, 193, 119, 236, 180,  47, 153,   2, 238, 117, 195,  88, 
  0, 139,  13, 134,  26, 145,  23, 156,  52, 191,  57, 178,  46, 165,  35, 168, 
  0, 187, 109, 214, 218,  97, 183,  12, 175,  20, 194, 121, 117, 206,  24, 163, 
  0, 171,  77, 230, 154,  49, 215, 124,  47, 132,  98, 201, 181,  30, 248,  83, 
  0, 219, 173, 118,  65, 154, 236,  55, 130,  89,  47, 244, 195,  24, 110, 181, 
  0, 203, 141,  70,   1, 202, 140,  71,   2, 201, 143,  68,   3, 200, 142,  69, 
  0, 251, 237,  22, 193,  58,  44, 215, 153,  98, 116, 143,  88, 163, 181,  78, 
  0, 235, 205,  38, 129, 106,  76, 167,  25, 242, 212,  63, 152, 115,  85, 190, 
  0,  54, 108,  90, 216, 238, 180, 130, 171, 157, 199, 241, 115,  69,  31,  41, 
  0,  38,  76, 106, 152, 190, 212, 242,  43,  13, 103,  65, 179, 149, 255, 217, 
  0,  22,  44,  58,  88,  78, 116,  98, 176, 166, 156, 138, 232, 254, 196, 210, 
  0,   6,  12,  10,  24,  30,  20,  18,  48,  54,  60,  58,  40,  46,  36,  34, 
  0, 118, 236, 154, 195, 181,  47,  89, 157, 235, 113,   7,  94,  40, 178, 196, 
  0, 102, 204, 170, 131, 229,  79,  41,  29, 123, 209, 183, 158, 248,  82,  52, 
  0,  86, 172, 250,  67,  21, 239, 185, 134, 208,  42, 124, 197, 147, 105,  63, 
  0,  70, 140, 202,   3,  69, 143, 201,   6,  64, 138, 204,   5,  67, 137, 207, 
  0, 182, 119, 193, 238,  88, 153,  47, 199, 113, 176,   6,  41, 159,  94, 232, 
  0, 166,  87, 241, 174,   8, 249,  95,  71, 225,  16, 182, 233,  79, 190,  24, 
  0, 150,  55, 161, 110, 248,  89, 207, 220,  74, 235, 125, 178,  36, 133,  19, 
  0, 134,  23, 145,  46, 168,  57, 191,  92, 218,  75, 205, 114, 244, 101, 227, 
  0, 246, 247,   1, 245,   3,   2, 244, 241,   7,   6, 240,   4, 242, 243,   5, 
  0, 230, 215,  49, 181,  83,  98, 132, 113, 151, 166,  64, 196,  34,  19, 245, 
  0, 214, 183,  97, 117, 163, 194,  20, 234,  60,  93, 139, 159,  73,  40, 254, 
  0, 198, 151,  81,  53, 243, 162, 100, 106, 172, 253,  59,  95, 153, 200,  14, 
  0,  45,  90, 119, 180, 153, 238, 195, 115,  94,  41,   4, 199, 234, 157, 176, 
  0,  61, 122,  71, 244, 201, 142, 179, 243, 206, 137, 180,   7,  58, 125,  64, 
  0,  13,  26,  23,  52,  57,  46,  35, 104, 101, 114, 127,  92,  81,  70,  75, 
  0,  29,  58,  39, 116, 105,  78,  83, 232, 245, 210, 207, 156, 129, 166, 187, 
  0, 109, 218, 183, 175, 194, 117,  24,  69,  40, 159, 242, 234, 135,  48,  93, 
  0, 125, 250, 135, 239, 146,  21, 104, 197, 184,  63,  66,  42,  87, 208, 173, 
  0,  77, 154, 215,  47,  98, 181, 248,  94,  19, 196, 137, 113,  60, 235, 166, 
  0,  93, 186, 231, 111,  50, 213, 136, 222, 131, 100,  57, 177, 236,  11,  86, 
  0, 173,  65, 236, 130,  47, 195, 110,  31, 178,  94, 243, 157,  48, 220, 113, 
  0, 189,  97, 220, 194, 127, 163,  30, 159,  34, 254,  67,  93, 224,  60, 129, 
  0, 141,   1, 140,   2, 143,   3, 142,   4, 137,   5, 136,   6, 139,   7, 138, 
  0, 157,  33, 188,  66, 223,  99, 254, 132,  25, 165,  56, 198,  91, 231, 122, 
  0, 237, 193,  44, 153, 116,  88, 181,  41, 196, 232,   5, 176,  93, 113, 156, 
  0, 253, 225,  28, 217,  36,  56, 197, 169,  84,  72, 181, 112, 141, 145, 108, 
  0, 205, 129,  76,  25, 212, 152,  85,  50, 255, 179, 126,  43, 230, 170, 103, 
  0, 221, 161, 124,  89, 132, 248,  37, 178, 111,  19, 206, 235,  54,  74, 151, 
  0, 108, 216, 180, 171, 199, 115,  31,  77,  33, 149, 249, 230, 138,  62,  82, 
  0, 124, 248, 132, 235, 151,  19, 111, 205, 177,  53,  73,  38,  90, 222, 162, 
  0,  76, 152, 212,  43, 103, 179, 255,  86,  26, 206, 130, 125,  49, 229, 169, 
  0,  92, 184, 228, 107,  55, 211, 143, 214, 138, 110,  50, 189, 225,   5,  89, 
  0,  44,  88, 116, 176, 156, 232, 196, 123,  87,  35,  15, 203, 231, 147, 191, 
  0,  60, 120,  68, 240, 204, 136, 180, 251, 199, 131, 191,  11,  55, 115,  79, 
  0,  12,  24,  20,  48,  60,  40,  36,  96, 108, 120, 116,  80,  92,  72,  68, 
  0,  28,  56,  36, 112, 108,  72,  84, 224, 252, 216, 196, 144, 140, 168, 180, 
  0, 236, 195,  47, 157, 113,  94, 178,  33, 205, 226,  14, 188,  80, 127, 147, 
  0, 252, 227,  31, 221,  33,  62, 194, 161,  93,  66, 190, 124, 128, 159,  99, 
  0, 204, 131,  79,  29, 209, 158,  82,  58, 246, 185, 117,  39, 235, 164, 104, 
  0, 220, 163, 127,  93, 129, 254,  34, 186, 102,  25, 197, 231,  59,  68, 152, 
  0, 172,  67, 239, 134,  42, 197, 105,  23, 187,  84, 248, 145,  61, 210, 126, 
  0, 188,  99, 223, 198, 122, 165,  25, 151,  43, 244,  72,  81, 237,  50, 142, 
  0, 140,   3, 143,   6, 138,   5, 137,  12, 128,  15, 131,  10, 134,   9, 133, 
  0, 156,  35, 191,  70, 218, 101, 249, 140,  16, 175,  51, 202,  86, 233, 117, 
  0, 119, 238, 153, 199, 176,  41,  94, 149, 226, 123,  12,  82,  37, 188, 203, 
  0, 103, 206, 169, 135, 224,  73,  46,  21, 114, 219, 188, 146, 245,  92,  59, 
  0,  87, 174, 249,  71,  16, 233, 190, 142, 217,  32, 119, 201, 158, 103,  48, 
  0,  71, 142, 201,   7,  64, 137, 206,  14,  73, 128, 199,   9,  78, 135, 192, 
  0,  55, 110,  89, 220, 235, 178, 133, 163, 148, 205, 250, 127,  72,  17,  38, 
  0,  39,  78, 105, 156, 187, 210, 245,  35,   4, 109,  74, 191, 152, 241, 214, 
  0,  23,  46,  57,  92,  75, 114, 101, 184, 175, 150, 129, 228, 243, 202, 221, 
  0,   7,  14,   9,  28,  27,  18,  21,  56,  63,  54,  49,  36,  35,  42,  45, 
  0, 247, 245,   2, 241,   6,   4, 243, 249,  14,  12, 251,   8, 255, 253,  10, 
  0, 231, 213,  50, 177,  86, 100, 131, 121, 158, 172,  75, 200,  47,  29, 250, 
  0, 215, 181,  98, 113, 166, 196,  19, 226,  53,  87, 128, 147,  68,  38, 241, 
  0, 199, 149,  82,  49, 246, 164,  99,  98, 165, 247,  48,  83, 148, 198,   1, 
  0, 183, 117, 194, 234,  93, 159,  40, 207, 120, 186,  13,  37, 146,  80, 231, 
  0, 167,  85, 242, 170,  13, 255,  88,  79, 232,  26, 189, 229,  66, 176,  23, 
  0, 151,  53, 162, 106, 253,  95, 200, 212,  67, 225, 118, 190,  41, 139,  28, 
  0, 135,  21, 146,  42, 173,  63, 184,  84, 211,  65, 198, 126, 249, 107, 236, 
  0,  90, 180, 238, 115,  41, 199, 157, 230, 188,  82,   8, 149, 207,  33, 123, 
  0,  74, 148, 222,  51, 121, 167, 237, 102,  44, 242, 184,  85,  31, 193, 139, 
  0, 122, 244, 142, 243, 137,   7, 125, 253, 135,   9, 115,  14, 116, 250, 128, 
  0, 106, 212, 190, 179, 217, 103,  13, 125,  23, 169, 195, 206, 164,  26, 112, 
  0,  26,  52,  46, 104, 114,  92,  70, 208, 202, 228, 254, 184, 162, 140, 150, 
  0,  10,  20,  30,  40,  34,  60,  54,  80,  90,  68,  78, 120, 114, 108, 102, 
  0,  58, 116,  78, 232, 210, 156, 166, 203, 241, 191, 133,  35,  25,  87, 109, 
  0,  42,  84, 126, 168, 130, 252, 214,  75,  97,  31,  53, 227, 201, 183, 157, 
  0, 218, 175, 117,  69, 159, 234,  48, 138,  80,  37, 255, 207,  21,  96, 186, 
  0, 202, 143,  69,   5, 207, 138,  64,  10, 192, 133,  79,  15, 197, 128,  74, 
  0, 250, 239,  21, 197,  63,  42, 208, 145, 107, 126, 132,  84, 174, 187,  65, 
  0, 234, 207,  37, 133, 111,  74, 160,  17, 251, 222,  52, 148, 126,  91, 177, 
  0, 154,  47, 181,  94, 196, 113, 235, 188,  38, 147,   9, 226, 120, 205,  87, 
  0, 138,  15, 133,  30, 148,  17, 155,  60, 182,  51, 185,  34, 168,  45, 167, 
  0, 186, 111, 213, 222, 100, 177,  11, 167,  29, 200, 114, 121, 195,  22, 172, 
  0, 170,  79, 229, 158,  52, 209, 123,  39, 141, 104, 194, 185,  19, 246,  92, 
  0,  65, 130, 195,  31,  94, 157, 220,  62, 127, 188, 253,  33,  96, 163, 226, 
  0,  81, 162, 243,  95,  14, 253, 172, 190, 239,  28,  77, 225, 176,  67,  18, 
  0,  97, 194, 163, 159, 254,  93,  60,  37,  68, 231, 134, 186, 219, 120,  25, 
  0, 113, 226, 147, 223, 174,  61,  76, 165, 212,  71,  54, 122,  11, 152, 233, 
  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15, 
  0,  17,  34,  51,  68,  85, 102, 119, 136, 153, 170, 187, 204, 221, 238, 255, 
  0,  33,  66,  99, 132, 165, 198, 231,  19,  50,  81, 112, 151, 182, 213, 244, 
  0,  49,  98,  83, 196, 245, 166, 151, 147, 162, 241, 192,  87, 102,  53,   4, 
  0, 193, 153,  88,  41, 232, 176, 113,  82, 147, 203,  10, 123, 186, 226,  35, 
  0, 209, 185, 104, 105, 184, 208,   1, 210,   3, 107, 186, 187, 106,   2, 211, 
  0, 225, 217,  56, 169,  72, 112, 145,  73, 168, 144, 113, 224,   1,  57, 216, 
  0, 241, 249,   8, 233,  24,  16, 225, 201,  56,  48, 193,  32, 209, 217,  40, 
  0, 129,  25, 152,  50, 179,  43, 170, 100, 229, 125, 252,  86, 215,  79, 206, 
  0, 145,  57, 168, 114, 227,  75, 218, 228, 117, 221,  76, 150,   7, 175,  62, 
  0, 161,  89, 248, 178,  19, 235,  74, 127, 222,  38, 135, 205, 108, 148,  53, 
  0, 177, 121, 200, 242,  67, 139,  58, 255,  78, 134,  55,  13, 188, 116, 197, 
  0, 216, 171, 115,  77, 149, 230,  62, 154,  66,  49, 233, 215,  15, 124, 164, 
  0, 200, 139,  67,  13, 197, 134,  78,  26, 210, 145,  89,  23, 223, 156,  84, 
  0, 248, 235,  19, 205,  53,  38, 222, 129, 121, 106, 146,  76, 180, 167,  95, 
  0, 232, 203,  35, 141, 101,  70, 174,   1, 233, 202,  34, 140, 100,  71, 175, 
  0, 152,  43, 179,  86, 206, 125, 229, 172,  52, 135,  31, 250,  98, 209,  73, 
  0, 136,  11, 131,  22, 158,  29, 149,  44, 164,  39, 175,  58, 178,  49, 185, 
  0, 184, 107, 211, 214, 110, 189,   5, 183,  15, 220, 100,  97, 217,  10, 178, 
  0, 168,  75, 227, 150,  62, 221, 117,  55, 159, 124, 212, 161,   9, 234,  66, 
  0,  88, 176, 232, 123,  35, 203, 147, 246, 174,  70,  30, 141, 213,  61, 101, 
  0,  72, 144, 216,  59, 115, 171, 227, 118,  62, 230, 174,  77,   5, 221, 149, 
  0, 120, 240, 136, 251, 131,  11, 115, 237, 149,  29, 101,  22, 110, 230, 158, 
  0, 104, 208, 184, 187, 211, 107,   3, 109,   5, 189, 213, 214, 190,   6, 110, 
  0,  24,  48,  40,  96, 120,  80,  72, 192, 216, 240, 232, 160, 184, 144, 136, 
  0,   8,  16,  24,  32,  40,  48,  56,  64,  72,  80,  88,  96, 104, 112, 120, 
  0,  56, 112,  72, 224, 216, 144, 168, 219, 227, 171, 147,  59,   3,  75, 115, 
  0,  40,  80, 120, 160, 136, 240, 216,  91, 115,  11,  35, 251, 211, 171, 131, 
  0, 195, 157,  94,  33, 226, 188, 127,  66, 129, 223,  28,  99, 160, 254,  61, 
  0, 211, 189, 110,  97, 178, 220,  15, 194,  17, 127, 172, 163, 112,  30, 205, 
  0, 227, 221,  62, 161,  66, 124, 159,  89, 186, 132, 103, 248,  27,  37, 198, 
  0, 243, 253,  14, 225,  18,  28, 239, 217,  42,  36, 215,  56, 203, 197,  54, 
  0, 131,  29, 158,  58, 185,  39, 164, 116, 247, 105, 234,  78, 205,  83, 208, 
  0, 147,  61, 174, 122, 233,  71, 212, 244, 103, 201,  90, 142,  29, 179,  32, 
  0, 163,  93, 254, 186,  25, 231,  68, 111, 204,  50, 145, 213, 118, 136,  43, 
  0, 179, 125, 206, 250,  73, 135,  52, 239,  92, 146,  33,  21, 166, 104, 219, 
  0,  67, 134, 197,  23,  84, 145, 210,  46, 109, 168, 235,  57, 122, 191, 252, 
  0,  83, 166, 245,  87,   4, 241, 162, 174, 253,   8,  91, 249, 170,  95,  12, 
  0,  99, 198, 165, 151, 244,  81,  50,  53,  86, 243, 144, 162, 193, 100,   7, 
  0, 115, 230, 149, 215, 164,  49,  66, 181, 198,  83,  32,  98,  17, 132, 247, 
  0,   3,   6,   5,  12,  15,  10,   9,  24,  27,  30,  29,  20,  23,  18,  17, 
  0,  19,  38,  53,  76,  95, 106, 121, 152, 139, 190, 173, 212, 199, 242, 225, 
  0,  35,  70, 101, 140, 175, 202, 233,   3,  32,  69, 102, 143, 172, 201, 234, 
  0,  51, 102,  85, 204, 255, 170, 153, 131, 176, 229, 214,  79, 124,  41,  26, 
  0, 238, 199,  41, 149, 123,  82, 188,  49, 223, 246,  24, 164,  74,  99, 141, 
  0, 254, 231,  25, 213,  43,  50, 204, 177,  79,  86, 168, 100, 154, 131, 125, 
  0, 206, 135,  73,  21, 219, 146,  92,  42, 228, 173,  99,  63, 241, 184, 118, 
  0, 222, 167, 121,  85, 139, 242,  44, 170, 116,  13, 211, 255,  33,  88, 134, 
  0, 174,  71, 233, 142,  32, 201, 103,   7, 169,  64, 238, 137,  39, 206,  96, 
  0, 190, 103, 217, 206, 112, 169,  23, 135,  57, 224,  94,  73, 247,  46, 144, 
  0, 142,   7, 137,  14, 128,   9, 135,  28, 146,  27, 149,  18, 156,  21, 155, 
  0, 158,  39, 185,  78, 208, 105, 247, 156,   2, 187,  37, 210,  76, 245, 107, 
  0, 110, 220, 178, 163, 205, 127,  17,  93,  51, 129, 239, 254, 144,  34,  76, 
  0, 126, 252, 130, 227, 157,  31,  97, 221, 163,  33,  95,  62,  64, 194, 188, 
  0,  78, 156, 210,  35, 109, 191, 241,  70,   8, 218, 148, 101,  43, 249, 183, 
  0,  94, 188, 226,  99,  61, 223, 129, 198, 152, 122,  36, 165, 251,  25,  71, 
  0,  46,  92, 114, 184, 150, 228, 202, 107,  69,  55,  25, 211, 253, 143, 161, 
  0,  62, 124,  66, 248, 198, 132, 186, 235, 213, 151, 169,  19,  45, 111,  81, 
  0,  14,  28,  18,  56,  54,  36,  42, 112, 126, 108,  98,  72,  70,  84,  90, 
  0,  30,  60,  34, 120, 102,  68,  90, 240, 238, 204, 210, 136, 150, 180, 170, 
  0, 245, 241,   4, 249,  12,   8, 253, 233,  28,  24, 237,  16, 229, 225,  20, 
  0, 229, 209,  52, 185,  92, 104, 141, 105, 140, 184,  93, 208,  53,   1, 228, 
  0, 213, 177, 100, 121, 172, 200,  29, 242,  39,  67, 150, 139,  94,  58, 239, 
  0, 197, 145,  84,  57, 252, 168, 109, 114, 183, 227,  38,  75, 142, 218,  31, 
  0, 181, 113, 196, 226,  87, 147,  38, 223, 106, 174,  27,  61, 136,  76, 249, 
  0, 165,  81, 244, 162,   7, 243,  86,  95, 250,  14, 171, 253,  88, 172,   9, 
  0, 149,  49, 164,  98, 247,  83, 198, 196,  81, 245,  96, 166,  51, 151,   2, 
  0, 133,  17, 148,  34, 167,  51, 182,  68, 193,  85, 208, 102, 227, 119, 242, 
  0, 117, 234, 159, 207, 186,  37,  80, 133, 240, 111,  26,  74,  63, 160, 213, 
  0, 101, 202, 175, 143, 234,  69,  32,   5,  96, 207, 170, 138, 239,  64,  37, 
  0,  85, 170, 255,  79,  26, 229, 176, 158, 203,  52,  97, 209, 132, 123,  46, 
  0,  69, 138, 207,  15,  74, 133, 192,  30,  91, 148, 209,  17,  84, 155, 222, 
  0,  53, 106,  95, 212, 225, 190, 139, 179, 134, 217, 236, 103,  82,  13,  56, 
  0,  37,  74, 111, 148, 177, 222, 251,  51,  22, 121,  92, 167, 130, 237, 200, 
  0,  21,  42,  63,  84,  65, 126, 107, 168, 189, 130, 151, 252, 233, 214, 195, 
  0,   5,  10,  15,  20,  17,  30,  27,  40,  45,  34,  39,  60,  57,  54,  51, 
  0, 180, 115, 199, 230,  82, 149,  33, 215,  99, 164,  16,  49, 133,  66, 246, 
  0, 164,  83, 247, 166,   2, 245,  81,  87, 243,   4, 160, 241,  85, 162,   6, 
  0, 148,  51, 167, 102, 242,  85, 193, 204,  88, 255, 107, 170,  62, 153,  13, 
  0, 132,  19, 151,  38, 162,  53, 177,  76, 200,  95, 219, 106, 238, 121, 253, 
  0, 244, 243,   7, 253,   9,  14, 250, 225,  21,  18, 230,  28, 232, 239,  27, 
  0, 228, 211,  55, 189,  89, 110, 138,  97, 133, 178,  86, 220,  56,  15, 235, 
  0, 212, 179, 103, 125, 169, 206,  26, 250,  46,  73, 157, 135,  83,  52, 224, 
  0, 196, 147,  87,  61, 249, 174, 106, 122, 190, 233,  45,  71, 131, 212,  16, 
  0,  52, 104,  92, 208, 228, 184, 140, 187, 143, 211, 231, 107,  95,   3,  55, 
  0,  36,  72, 108, 144, 180, 216, 252,  59,  31, 115,  87, 171, 143, 227, 199, 
  0,  20,  40,  60,  80,  68, 120, 108, 160, 180, 136, 156, 240, 228, 216, 204, 
  0,   4,   8,  12,  16,  20,  24,  28,  32,  36,  40,  44,  48,  52,  56,  60, 
  0, 116, 232, 156, 203, 191,  35,  87, 141, 249, 101,  17,  70,  50, 174, 218, 
  0, 100, 200, 172, 139, 239,  67,  39,  13, 105, 197, 161, 134, 226,  78,  42, 
  0,  84, 168, 252,  75,  31, 227, 183, 150, 194,  62, 106, 221, 137, 117,  33, 
  0,  68, 136, 204,  11,  79, 131, 199,  22,  82, 158, 218,  29,  89, 149, 209, 
  0, 175,  69, 234, 138,  37, 207,  96,  15, 160,  74, 229, 133,  42, 192, 111, 
  0, 191, 101, 218, 202, 117, 175,  16, 143,  48, 234,  85,  69, 250,  32, 159, 
  0, 143,   5, 138,  10, 133,  15, 128,  20, 155,  17, 158,  30, 145,  27, 148, 
  0, 159,  37, 186,  74, 213, 111, 240, 148,  11, 177,  46, 222,  65, 251, 100, 
  0, 239, 197,  42, 145, 126,  84, 187,  57, 214, 252,  19, 168,  71, 109, 130, 
  0, 255, 229,  26, 209,  46,  52, 203, 185,  70,  92, 163, 104, 151, 141, 114, 
  0, 207, 133,  74,  17, 222, 148,  91,  34, 237, 167, 104,  51, 252, 182, 121, 
  0, 223, 165, 122,  81, 142, 244,  43, 162, 125,   7, 216, 243,  44,  86, 137, 
  0,  47,  94, 113, 188, 147, 226, 205,  99,  76,  61,  18, 223, 240, 129, 174, 
  0,  63, 126,  65, 252, 195, 130, 189, 227, 220, 157, 162,  31,  32,  97,  94, 
  0,  15,  30,  17,  60,  51,  34,  45, 120, 119, 102, 105,  68,  75,  90,  85, 
  0,  31,  62,  33, 124,  99,  66,  93, 248, 231, 198, 217, 132, 155, 186, 165, 
  0, 111, 222, 177, 167, 200, 121,  22,  85,  58, 139, 228, 242, 157,  44,  67, 
  0, 127, 254, 129, 231, 152,  25, 102, 213, 170,  43,  84,  50,  77, 204, 179, 
  0,  79, 158, 209,  39, 104, 185, 246,  78,   1, 208, 159, 105,  38, 247, 184, 
  0,  95, 190, 225, 103,  56, 217, 134, 206, 145, 112,  47, 169, 246,  23,  72, 
  0, 130,  31, 157,  62, 188,  33, 163, 124, 254,  99, 225,  66, 192,  93, 223, 
  0, 146,  63, 173, 126, 236,  65, 211, 252, 110, 195,  81, 130,  16, 189,  47, 
  0, 162,  95, 253, 190,  28, 225,  67, 103, 197,  56, 154, 217, 123, 134,  36, 
  0, 178, 127, 205, 254,  76, 129,  51, 231,  85, 152,  42,  25, 171, 102, 212, 
  0, 194, 159,  93,  37, 231, 186, 120,  74, 136, 213,  23, 111, 173, 240,  50, 
  0, 210, 191, 109, 101, 183, 218,   8, 202,  24, 117, 167, 175, 125,  16, 194, 
  0, 226, 223,  61, 165,  71, 122, 152,  81, 179, 142, 108, 244,  22,  43, 201, 
  0, 242, 255,  13, 229,  23,  26, 232, 209,  35,  46, 220,  52, 198, 203,  57, 
  0,   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,  22,  24,  26,  28,  30, 
  0,  18,  36,  54,  72,  90, 108, 126, 144, 130, 180, 166, 216, 202, 252, 238, 
  0,  34,  68, 102, 136, 170, 204, 238,  11,  41,  79, 109, 131, 161, 199, 229, 
  0,  50, 100,  86, 200, 250, 172, 158, 139, 185, 239, 221,  67, 113,  39,  21, 
  0,  66, 132, 198,  19,  81, 151, 213,  38, 100, 162, 224,  53, 119, 177, 243, 
  0,  82, 164, 246,  83,   1, 247, 165, 166, 244,   2,  80, 245, 167,  81,   3, 
  0,  98, 196, 166, 147, 241,  87,  53,  61,  95, 249, 155, 174, 204, 106,   8, 
  0, 114, 228, 150, 211, 161,  55,  69, 189, 207,  89,  43, 110,  28, 138, 248, 
  0, 153,  41, 176,  82, 203, 123, 226, 164,  61, 141,  20, 246, 111, 223,  70, 
  0, 137,   9, 128,  18, 155,  27, 146,  36, 173,  45, 164,  54, 191,  63, 182, 
  0, 185, 105, 208, 210, 107, 187,   2, 191,   6, 214, 111, 109, 212,   4, 189, 
  0, 169,  73, 224, 146,  59, 219, 114,  63, 150, 118, 223, 173,   4, 228,  77, 
  0, 217, 169, 112,  73, 144, 224,  57, 146,  75,  59, 226, 219,   2, 114, 171, 
  0, 201, 137,  64,   9, 192, 128,  73,  18, 219, 155,  82,  27, 210, 146,  91, 
  0, 249, 233,  16, 201,  48,  32, 217, 137, 112,  96, 153,  64, 185, 169,  80, 
  0, 233, 201,  32, 137,  96,  64, 169,   9, 224, 192,  41, 128, 105,  73, 160, 
  0,  25,  50,  43, 100, 125,  86,  79, 200, 209, 250, 227, 172, 181, 158, 135, 
  0,   9,  18,  27,  36,  45,  54,  63,  72,  65,  90,  83, 108, 101, 126, 119, 
  0,  57, 114,  75, 228, 221, 150, 175, 211, 234, 161, 152,  55,  14,  69, 124, 
  0,  41,  82, 123, 164, 141, 246, 223,  83, 122,   1,  40, 247, 222, 165, 140, 
  0,  89, 178, 235, 127,  38, 205, 148, 254, 167,  76,  21, 129, 216,  51, 106, 
  0,  73, 146, 219,  63, 118, 173, 228, 126,  55, 236, 165,  65,   8, 211, 154, 
  0, 121, 242, 139, 255, 134,  13, 116, 229, 156,  23, 110,  26,  99, 232, 145, 
  0, 105, 210, 187, 191, 214, 109,   4, 101,  12, 183, 222, 218, 179,   8,  97,
};
uint8 bltable[] = {
	0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 
  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15, 
  0,   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,  22,  24,  26,  28,  30, 
  0,   3,   6,   5,  12,  15,  10,   9,  24,  27,  30,  29,  20,  23,  18,  17, 
  0,   4,   8,  12,  16,  20,  24,  28,  32,  36,  40,  44,  48,  52,  56,  60, 
  0,   5,  10,  15,  20,  17,  30,  27,  40,  45,  34,  39,  60,  57,  54,  51, 
  0,   6,  12,  10,  24,  30,  20,  18,  48,  54,  60,  58,  40,  46,  36,  34, 
  0,   7,  14,   9,  28,  27,  18,  21,  56,  63,  54,  49,  36,  35,  42,  45, 
  0,   8,  16,  24,  32,  40,  48,  56,  64,  72,  80,  88,  96, 104, 112, 120, 
  0,   9,  18,  27,  36,  45,  54,  63,  72,  65,  90,  83, 108, 101, 126, 119, 
  0,  10,  20,  30,  40,  34,  60,  54,  80,  90,  68,  78, 120, 114, 108, 102, 
  0,  11,  22,  29,  44,  39,  58,  49,  88,  83,  78,  69, 116, 127,  98, 105, 
  0,  12,  24,  20,  48,  60,  40,  36,  96, 108, 120, 116,  80,  92,  72,  68, 
  0,  13,  26,  23,  52,  57,  46,  35, 104, 101, 114, 127,  92,  81,  70,  75, 
  0,  14,  28,  18,  56,  54,  36,  42, 112, 126, 108,  98,  72,  70,  84,  90, 
  0,  15,  30,  17,  60,  51,  34,  45, 120, 119, 102, 105,  68,  75,  90,  85, 
  0,  16,  32,  48,  64,  80,  96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 
  0,  17,  34,  51,  68,  85, 102, 119, 136, 153, 170, 187, 204, 221, 238, 255, 
  0,  18,  36,  54,  72,  90, 108, 126, 144, 130, 180, 166, 216, 202, 252, 238, 
  0,  19,  38,  53,  76,  95, 106, 121, 152, 139, 190, 173, 212, 199, 242, 225, 
  0,  20,  40,  60,  80,  68, 120, 108, 160, 180, 136, 156, 240, 228, 216, 204, 
  0,  21,  42,  63,  84,  65, 126, 107, 168, 189, 130, 151, 252, 233, 214, 195, 
  0,  22,  44,  58,  88,  78, 116,  98, 176, 166, 156, 138, 232, 254, 196, 210, 
  0,  23,  46,  57,  92,  75, 114, 101, 184, 175, 150, 129, 228, 243, 202, 221, 
  0,  24,  48,  40,  96, 120,  80,  72, 192, 216, 240, 232, 160, 184, 144, 136, 
  0,  25,  50,  43, 100, 125,  86,  79, 200, 209, 250, 227, 172, 181, 158, 135, 
  0,  26,  52,  46, 104, 114,  92,  70, 208, 202, 228, 254, 184, 162, 140, 150, 
  0,  27,  54,  45, 108, 119,  90,  65, 216, 195, 238, 245, 180, 175, 130, 153, 
  0,  28,  56,  36, 112, 108,  72,  84, 224, 252, 216, 196, 144, 140, 168, 180, 
  0,  29,  58,  39, 116, 105,  78,  83, 232, 245, 210, 207, 156, 129, 166, 187, 
  0,  30,  60,  34, 120, 102,  68,  90, 240, 238, 204, 210, 136, 150, 180, 170, 
  0,  31,  62,  33, 124,  99,  66,  93, 248, 231, 198, 217, 132, 155, 186, 165, 
  0,  32,  64,  96, 128, 160, 192, 224,  27,  59,  91, 123, 155, 187, 219, 251, 
  0,  33,  66,  99, 132, 165, 198, 231,  19,  50,  81, 112, 151, 182, 213, 244, 
  0,  34,  68, 102, 136, 170, 204, 238,  11,  41,  79, 109, 131, 161, 199, 229, 
  0,  35,  70, 101, 140, 175, 202, 233,   3,  32,  69, 102, 143, 172, 201, 234, 
  0,  36,  72, 108, 144, 180, 216, 252,  59,  31, 115,  87, 171, 143, 227, 199, 
  0,  37,  74, 111, 148, 177, 222, 251,  51,  22, 121,  92, 167, 130, 237, 200, 
  0,  38,  76, 106, 152, 190, 212, 242,  43,  13, 103,  65, 179, 149, 255, 217, 
  0,  39,  78, 105, 156, 187, 210, 245,  35,   4, 109,  74, 191, 152, 241, 214, 
  0,  40,  80, 120, 160, 136, 240, 216,  91, 115,  11,  35, 251, 211, 171, 131, 
  0,  41,  82, 123, 164, 141, 246, 223,  83, 122,   1,  40, 247, 222, 165, 140, 
  0,  42,  84, 126, 168, 130, 252, 214,  75,  97,  31,  53, 227, 201, 183, 157, 
  0,  43,  86, 125, 172, 135, 250, 209,  67, 104,  21,  62, 239, 196, 185, 146, 
  0,  44,  88, 116, 176, 156, 232, 196, 123,  87,  35,  15, 203, 231, 147, 191, 
  0,  45,  90, 119, 180, 153, 238, 195, 115,  94,  41,   4, 199, 234, 157, 176, 
  0,  46,  92, 114, 184, 150, 228, 202, 107,  69,  55,  25, 211, 253, 143, 161, 
  0,  47,  94, 113, 188, 147, 226, 205,  99,  76,  61,  18, 223, 240, 129, 174, 
  0,  48,  96,  80, 192, 240, 160, 144, 155, 171, 251, 203,  91, 107,  59,  11, 
  0,  49,  98,  83, 196, 245, 166, 151, 147, 162, 241, 192,  87, 102,  53,   4, 
  0,  50, 100,  86, 200, 250, 172, 158, 139, 185, 239, 221,  67, 113,  39,  21, 
  0,  51, 102,  85, 204, 255, 170, 153, 131, 176, 229, 214,  79, 124,  41,  26, 
  0,  52, 104,  92, 208, 228, 184, 140, 187, 143, 211, 231, 107,  95,   3,  55, 
  0,  53, 106,  95, 212, 225, 190, 139, 179, 134, 217, 236, 103,  82,  13,  56, 
  0,  54, 108,  90, 216, 238, 180, 130, 171, 157, 199, 241, 115,  69,  31,  41, 
  0,  55, 110,  89, 220, 235, 178, 133, 163, 148, 205, 250, 127,  72,  17,  38, 
  0,  56, 112,  72, 224, 216, 144, 168, 219, 227, 171, 147,  59,   3,  75, 115, 
  0,  57, 114,  75, 228, 221, 150, 175, 211, 234, 161, 152,  55,  14,  69, 124, 
  0,  58, 116,  78, 232, 210, 156, 166, 203, 241, 191, 133,  35,  25,  87, 109, 
  0,  59, 118,  77, 236, 215, 154, 161, 195, 248, 181, 142,  47,  20,  89,  98, 
  0,  60, 120,  68, 240, 204, 136, 180, 251, 199, 131, 191,  11,  55, 115,  79, 
  0,  61, 122,  71, 244, 201, 142, 179, 243, 206, 137, 180,   7,  58, 125,  64, 
  0,  62, 124,  66, 248, 198, 132, 186, 235, 213, 151, 169,  19,  45, 111,  81, 
  0,  63, 126,  65, 252, 195, 130, 189, 227, 220, 157, 162,  31,  32,  97,  94, 
  0,  64, 128, 192,  27,  91, 155, 219,  54, 118, 182, 246,  45, 109, 173, 237, 
  0,  65, 130, 195,  31,  94, 157, 220,  62, 127, 188, 253,  33,  96, 163, 226, 
  0,  66, 132, 198,  19,  81, 151, 213,  38, 100, 162, 224,  53, 119, 177, 243, 
  0,  67, 134, 197,  23,  84, 145, 210,  46, 109, 168, 235,  57, 122, 191, 252, 
  0,  68, 136, 204,  11,  79, 131, 199,  22,  82, 158, 218,  29,  89, 149, 209, 
  0,  69, 138, 207,  15,  74, 133, 192,  30,  91, 148, 209,  17,  84, 155, 222, 
  0,  70, 140, 202,   3,  69, 143, 201,   6,  64, 138, 204,   5,  67, 137, 207, 
  0,  71, 142, 201,   7,  64, 137, 206,  14,  73, 128, 199,   9,  78, 135, 192, 
  0,  72, 144, 216,  59, 115, 171, 227, 118,  62, 230, 174,  77,   5, 221, 149, 
  0,  73, 146, 219,  63, 118, 173, 228, 126,  55, 236, 165,  65,   8, 211, 154, 
  0,  74, 148, 222,  51, 121, 167, 237, 102,  44, 242, 184,  85,  31, 193, 139, 
  0,  75, 150, 221,  55, 124, 161, 234, 110,  37, 248, 179,  89,  18, 207, 132, 
  0,  76, 152, 212,  43, 103, 179, 255,  86,  26, 206, 130, 125,  49, 229, 169, 
  0,  77, 154, 215,  47,  98, 181, 248,  94,  19, 196, 137, 113,  60, 235, 166, 
  0,  78, 156, 210,  35, 109, 191, 241,  70,   8, 218, 148, 101,  43, 249, 183, 
  0,  79, 158, 209,  39, 104, 185, 246,  78,   1, 208, 159, 105,  38, 247, 184, 
  0,  80, 160, 240,  91,  11, 251, 171, 182, 230,  22,  70, 237, 189,  77,  29, 
  0,  81, 162, 243,  95,  14, 253, 172, 190, 239,  28,  77, 225, 176,  67,  18, 
  0,  82, 164, 246,  83,   1, 247, 165, 166, 244,   2,  80, 245, 167,  81,   3, 
  0,  83, 166, 245,  87,   4, 241, 162, 174, 253,   8,  91, 249, 170,  95,  12, 
  0,  84, 168, 252,  75,  31, 227, 183, 150, 194,  62, 106, 221, 137, 117,  33, 
  0,  85, 170, 255,  79,  26, 229, 176, 158, 203,  52,  97, 209, 132, 123,  46, 
  0,  86, 172, 250,  67,  21, 239, 185, 134, 208,  42, 124, 197, 147, 105,  63, 
  0,  87, 174, 249,  71,  16, 233, 190, 142, 217,  32, 119, 201, 158, 103,  48, 
  0,  88, 176, 232, 123,  35, 203, 147, 246, 174,  70,  30, 141, 213,  61, 101, 
  0,  89, 178, 235, 127,  38, 205, 148, 254, 167,  76,  21, 129, 216,  51, 106, 
  0,  90, 180, 238, 115,  41, 199, 157, 230, 188,  82,   8, 149, 207,  33, 123, 
  0,  91, 182, 237, 119,  44, 193, 154, 238, 181,  88,   3, 153, 194,  47, 116, 
  0,  92, 184, 228, 107,  55, 211, 143, 214, 138, 110,  50, 189, 225,   5,  89, 
  0,  93, 186, 231, 111,  50, 213, 136, 222, 131, 100,  57, 177, 236,  11,  86, 
  0,  94, 188, 226,  99,  61, 223, 129, 198, 152, 122,  36, 165, 251,  25,  71, 
  0,  95, 190, 225, 103,  56, 217, 134, 206, 145, 112,  47, 169, 246,  23,  72, 
  0,  96, 192, 160, 155, 251,  91,  59,  45,  77, 237, 141, 182, 214, 118,  22, 
  0,  97, 194, 163, 159, 254,  93,  60,  37,  68, 231, 134, 186, 219, 120,  25, 
  0,  98, 196, 166, 147, 241,  87,  53,  61,  95, 249, 155, 174, 204, 106,   8, 
  0,  99, 198, 165, 151, 244,  81,  50,  53,  86, 243, 144, 162, 193, 100,   7, 
  0, 100, 200, 172, 139, 239,  67,  39,  13, 105, 197, 161, 134, 226,  78,  42, 
  0, 101, 202, 175, 143, 234,  69,  32,   5,  96, 207, 170, 138, 239,  64,  37, 
  0, 102, 204, 170, 131, 229,  79,  41,  29, 123, 209, 183, 158, 248,  82,  52, 
  0, 103, 206, 169, 135, 224,  73,  46,  21, 114, 219, 188, 146, 245,  92,  59, 
  0, 104, 208, 184, 187, 211, 107,   3, 109,   5, 189, 213, 214, 190,   6, 110, 
  0, 105, 210, 187, 191, 214, 109,   4, 101,  12, 183, 222, 218, 179,   8,  97, 
  0, 106, 212, 190, 179, 217, 103,  13, 125,  23, 169, 195, 206, 164,  26, 112, 
  0, 107, 214, 189, 183, 220,  97,  10, 117,  30, 163, 200, 194, 169,  20, 127, 
  0, 108, 216, 180, 171, 199, 115,  31,  77,  33, 149, 249, 230, 138,  62,  82, 
  0, 109, 218, 183, 175, 194, 117,  24,  69,  40, 159, 242, 234, 135,  48,  93, 
  0, 110, 220, 178, 163, 205, 127,  17,  93,  51, 129, 239, 254, 144,  34,  76, 
  0, 111, 222, 177, 167, 200, 121,  22,  85,  58, 139, 228, 242, 157,  44,  67, 
  0, 112, 224, 144, 219, 171,  59,  75, 173, 221,  77,  61, 118,   6, 150, 230, 
  0, 113, 226, 147, 223, 174,  61,  76, 165, 212,  71,  54, 122,  11, 152, 233, 
  0, 114, 228, 150, 211, 161,  55,  69, 189, 207,  89,  43, 110,  28, 138, 248, 
  0, 115, 230, 149, 215, 164,  49,  66, 181, 198,  83,  32,  98,  17, 132, 247, 
  0, 116, 232, 156, 203, 191,  35,  87, 141, 249, 101,  17,  70,  50, 174, 218, 
  0, 117, 234, 159, 207, 186,  37,  80, 133, 240, 111,  26,  74,  63, 160, 213, 
  0, 118, 236, 154, 195, 181,  47,  89, 157, 235, 113,   7,  94,  40, 178, 196, 
  0, 119, 238, 153, 199, 176,  41,  94, 149, 226, 123,  12,  82,  37, 188, 203, 
  0, 120, 240, 136, 251, 131,  11, 115, 237, 149,  29, 101,  22, 110, 230, 158, 
  0, 121, 242, 139, 255, 134,  13, 116, 229, 156,  23, 110,  26,  99, 232, 145, 
  0, 122, 244, 142, 243, 137,   7, 125, 253, 135,   9, 115,  14, 116, 250, 128, 
  0, 123, 246, 141, 247, 140,   1, 122, 245, 142,   3, 120,   2, 121, 244, 143, 
  0, 124, 248, 132, 235, 151,  19, 111, 205, 177,  53,  73,  38,  90, 222, 162, 
  0, 125, 250, 135, 239, 146,  21, 104, 197, 184,  63,  66,  42,  87, 208, 173, 
  0, 126, 252, 130, 227, 157,  31,  97, 221, 163,  33,  95,  62,  64, 194, 188, 
  0, 127, 254, 129, 231, 152,  25, 102, 213, 170,  43,  84,  50,  77, 204, 179, 
  0, 128,  27, 155,  54, 182,  45, 173, 108, 236, 119, 247,  90, 218,  65, 193, 
  0, 129,  25, 152,  50, 179,  43, 170, 100, 229, 125, 252,  86, 215,  79, 206, 
  0, 130,  31, 157,  62, 188,  33, 163, 124, 254,  99, 225,  66, 192,  93, 223, 
  0, 131,  29, 158,  58, 185,  39, 164, 116, 247, 105, 234,  78, 205,  83, 208, 
  0, 132,  19, 151,  38, 162,  53, 177,  76, 200,  95, 219, 106, 238, 121, 253, 
  0, 133,  17, 148,  34, 167,  51, 182,  68, 193,  85, 208, 102, 227, 119, 242, 
  0, 134,  23, 145,  46, 168,  57, 191,  92, 218,  75, 205, 114, 244, 101, 227, 
  0, 135,  21, 146,  42, 173,  63, 184,  84, 211,  65, 198, 126, 249, 107, 236, 
  0, 136,  11, 131,  22, 158,  29, 149,  44, 164,  39, 175,  58, 178,  49, 185, 
  0, 137,   9, 128,  18, 155,  27, 146,  36, 173,  45, 164,  54, 191,  63, 182, 
  0, 138,  15, 133,  30, 148,  17, 155,  60, 182,  51, 185,  34, 168,  45, 167, 
  0, 139,  13, 134,  26, 145,  23, 156,  52, 191,  57, 178,  46, 165,  35, 168, 
  0, 140,   3, 143,   6, 138,   5, 137,  12, 128,  15, 131,  10, 134,   9, 133, 
  0, 141,   1, 140,   2, 143,   3, 142,   4, 137,   5, 136,   6, 139,   7, 138, 
  0, 142,   7, 137,  14, 128,   9, 135,  28, 146,  27, 149,  18, 156,  21, 155, 
  0, 143,   5, 138,  10, 133,  15, 128,  20, 155,  17, 158,  30, 145,  27, 148, 
  0, 144,  59, 171, 118, 230,  77, 221, 236, 124, 215,  71, 154,  10, 161,  49, 
  0, 145,  57, 168, 114, 227,  75, 218, 228, 117, 221,  76, 150,   7, 175,  62, 
  0, 146,  63, 173, 126, 236,  65, 211, 252, 110, 195,  81, 130,  16, 189,  47, 
  0, 147,  61, 174, 122, 233,  71, 212, 244, 103, 201,  90, 142,  29, 179,  32, 
  0, 148,  51, 167, 102, 242,  85, 193, 204,  88, 255, 107, 170,  62, 153,  13, 
  0, 149,  49, 164,  98, 247,  83, 198, 196,  81, 245,  96, 166,  51, 151,   2, 
  0, 150,  55, 161, 110, 248,  89, 207, 220,  74, 235, 125, 178,  36, 133,  19, 
  0, 151,  53, 162, 106, 253,  95, 200, 212,  67, 225, 118, 190,  41, 139,  28, 
  0, 152,  43, 179,  86, 206, 125, 229, 172,  52, 135,  31, 250,  98, 209,  73, 
  0, 153,  41, 176,  82, 203, 123, 226, 164,  61, 141,  20, 246, 111, 223,  70, 
  0, 154,  47, 181,  94, 196, 113, 235, 188,  38, 147,   9, 226, 120, 205,  87, 
  0, 155,  45, 182,  90, 193, 119, 236, 180,  47, 153,   2, 238, 117, 195,  88, 
  0, 156,  35, 191,  70, 218, 101, 249, 140,  16, 175,  51, 202,  86, 233, 117, 
  0, 157,  33, 188,  66, 223,  99, 254, 132,  25, 165,  56, 198,  91, 231, 122, 
  0, 158,  39, 185,  78, 208, 105, 247, 156,   2, 187,  37, 210,  76, 245, 107, 
  0, 159,  37, 186,  74, 213, 111, 240, 148,  11, 177,  46, 222,  65, 251, 100, 
  0, 160,  91, 251, 182,  22, 237,  77, 119, 215,  44, 140, 193,  97, 154,  58, 
  0, 161,  89, 248, 178,  19, 235,  74, 127, 222,  38, 135, 205, 108, 148,  53, 
  0, 162,  95, 253, 190,  28, 225,  67, 103, 197,  56, 154, 217, 123, 134,  36, 
  0, 163,  93, 254, 186,  25, 231,  68, 111, 204,  50, 145, 213, 118, 136,  43, 
  0, 164,  83, 247, 166,   2, 245,  81,  87, 243,   4, 160, 241,  85, 162,   6, 
  0, 165,  81, 244, 162,   7, 243,  86,  95, 250,  14, 171, 253,  88, 172,   9, 
  0, 166,  87, 241, 174,   8, 249,  95,  71, 225,  16, 182, 233,  79, 190,  24, 
  0, 167,  85, 242, 170,  13, 255,  88,  79, 232,  26, 189, 229,  66, 176,  23, 
  0, 168,  75, 227, 150,  62, 221, 117,  55, 159, 124, 212, 161,   9, 234,  66, 
  0, 169,  73, 224, 146,  59, 219, 114,  63, 150, 118, 223, 173,   4, 228,  77, 
  0, 170,  79, 229, 158,  52, 209, 123,  39, 141, 104, 194, 185,  19, 246,  92, 
  0, 171,  77, 230, 154,  49, 215, 124,  47, 132,  98, 201, 181,  30, 248,  83, 
  0, 172,  67, 239, 134,  42, 197, 105,  23, 187,  84, 248, 145,  61, 210, 126, 
  0, 173,  65, 236, 130,  47, 195, 110,  31, 178,  94, 243, 157,  48, 220, 113, 
  0, 174,  71, 233, 142,  32, 201, 103,   7, 169,  64, 238, 137,  39, 206,  96, 
  0, 175,  69, 234, 138,  37, 207,  96,  15, 160,  74, 229, 133,  42, 192, 111, 
  0, 176, 123, 203, 246,  70, 141,  61, 247,  71, 140,  60,   1, 177, 122, 202, 
  0, 177, 121, 200, 242,  67, 139,  58, 255,  78, 134,  55,  13, 188, 116, 197, 
  0, 178, 127, 205, 254,  76, 129,  51, 231,  85, 152,  42,  25, 171, 102, 212, 
  0, 179, 125, 206, 250,  73, 135,  52, 239,  92, 146,  33,  21, 166, 104, 219, 
  0, 180, 115, 199, 230,  82, 149,  33, 215,  99, 164,  16,  49, 133,  66, 246, 
  0, 181, 113, 196, 226,  87, 147,  38, 223, 106, 174,  27,  61, 136,  76, 249, 
  0, 182, 119, 193, 238,  88, 153,  47, 199, 113, 176,   6,  41, 159,  94, 232, 
  0, 183, 117, 194, 234,  93, 159,  40, 207, 120, 186,  13,  37, 146,  80, 231, 
  0, 184, 107, 211, 214, 110, 189,   5, 183,  15, 220, 100,  97, 217,  10, 178, 
  0, 185, 105, 208, 210, 107, 187,   2, 191,   6, 214, 111, 109, 212,   4, 189, 
  0, 186, 111, 213, 222, 100, 177,  11, 167,  29, 200, 114, 121, 195,  22, 172, 
  0, 187, 109, 214, 218,  97, 183,  12, 175,  20, 194, 121, 117, 206,  24, 163, 
  0, 188,  99, 223, 198, 122, 165,  25, 151,  43, 244,  72,  81, 237,  50, 142, 
  0, 189,  97, 220, 194, 127, 163,  30, 159,  34, 254,  67,  93, 224,  60, 129, 
  0, 190, 103, 217, 206, 112, 169,  23, 135,  57, 224,  94,  73, 247,  46, 144, 
  0, 191, 101, 218, 202, 117, 175,  16, 143,  48, 234,  85,  69, 250,  32, 159, 
  0, 192, 155,  91,  45, 237, 182, 118,  90, 154, 193,   1, 119, 183, 236,  44, 
  0, 193, 153,  88,  41, 232, 176, 113,  82, 147, 203,  10, 123, 186, 226,  35, 
  0, 194, 159,  93,  37, 231, 186, 120,  74, 136, 213,  23, 111, 173, 240,  50, 
  0, 195, 157,  94,  33, 226, 188, 127,  66, 129, 223,  28,  99, 160, 254,  61, 
  0, 196, 147,  87,  61, 249, 174, 106, 122, 190, 233,  45,  71, 131, 212,  16, 
  0, 197, 145,  84,  57, 252, 168, 109, 114, 183, 227,  38,  75, 142, 218,  31, 
  0, 198, 151,  81,  53, 243, 162, 100, 106, 172, 253,  59,  95, 153, 200,  14, 
  0, 199, 149,  82,  49, 246, 164,  99,  98, 165, 247,  48,  83, 148, 198,   1, 
  0, 200, 139,  67,  13, 197, 134,  78,  26, 210, 145,  89,  23, 223, 156,  84, 
  0, 201, 137,  64,   9, 192, 128,  73,  18, 219, 155,  82,  27, 210, 146,  91, 
  0, 202, 143,  69,   5, 207, 138,  64,  10, 192, 133,  79,  15, 197, 128,  74, 
  0, 203, 141,  70,   1, 202, 140,  71,   2, 201, 143,  68,   3, 200, 142,  69, 
  0, 204, 131,  79,  29, 209, 158,  82,  58, 246, 185, 117,  39, 235, 164, 104, 
  0, 205, 129,  76,  25, 212, 152,  85,  50, 255, 179, 126,  43, 230, 170, 103, 
  0, 206, 135,  73,  21, 219, 146,  92,  42, 228, 173,  99,  63, 241, 184, 118, 
  0, 207, 133,  74,  17, 222, 148,  91,  34, 237, 167, 104,  51, 252, 182, 121, 
  0, 208, 187, 107, 109, 189, 214,   6, 218,  10,  97, 177, 183, 103,  12, 220, 
  0, 209, 185, 104, 105, 184, 208,   1, 210,   3, 107, 186, 187, 106,   2, 211, 
  0, 210, 191, 109, 101, 183, 218,   8, 202,  24, 117, 167, 175, 125,  16, 194, 
  0, 211, 189, 110,  97, 178, 220,  15, 194,  17, 127, 172, 163, 112,  30, 205, 
  0, 212, 179, 103, 125, 169, 206,  26, 250,  46,  73, 157, 135,  83,  52, 224, 
  0, 213, 177, 100, 121, 172, 200,  29, 242,  39,  67, 150, 139,  94,  58, 239, 
  0, 214, 183,  97, 117, 163, 194,  20, 234,  60,  93, 139, 159,  73,  40, 254, 
  0, 215, 181,  98, 113, 166, 196,  19, 226,  53,  87, 128, 147,  68,  38, 241, 
  0, 216, 171, 115,  77, 149, 230,  62, 154,  66,  49, 233, 215,  15, 124, 164, 
  0, 217, 169, 112,  73, 144, 224,  57, 146,  75,  59, 226, 219,   2, 114, 171, 
  0, 218, 175, 117,  69, 159, 234,  48, 138,  80,  37, 255, 207,  21,  96, 186, 
  0, 219, 173, 118,  65, 154, 236,  55, 130,  89,  47, 244, 195,  24, 110, 181, 
  0, 220, 163, 127,  93, 129, 254,  34, 186, 102,  25, 197, 231,  59,  68, 152, 
  0, 221, 161, 124,  89, 132, 248,  37, 178, 111,  19, 206, 235,  54,  74, 151, 
  0, 222, 167, 121,  85, 139, 242,  44, 170, 116,  13, 211, 255,  33,  88, 134, 
  0, 223, 165, 122,  81, 142, 244,  43, 162, 125,   7, 216, 243,  44,  86, 137, 
  0, 224, 219,  59, 173,  77, 118, 150,  65, 161, 154, 122, 236,  12,  55, 215, 
  0, 225, 217,  56, 169,  72, 112, 145,  73, 168, 144, 113, 224,   1,  57, 216, 
  0, 226, 223,  61, 165,  71, 122, 152,  81, 179, 142, 108, 244,  22,  43, 201, 
  0, 227, 221,  62, 161,  66, 124, 159,  89, 186, 132, 103, 248,  27,  37, 198, 
  0, 228, 211,  55, 189,  89, 110, 138,  97, 133, 178,  86, 220,  56,  15, 235, 
  0, 229, 209,  52, 185,  92, 104, 141, 105, 140, 184,  93, 208,  53,   1, 228, 
  0, 230, 215,  49, 181,  83,  98, 132, 113, 151, 166,  64, 196,  34,  19, 245, 
  0, 231, 213,  50, 177,  86, 100, 131, 121, 158, 172,  75, 200,  47,  29, 250, 
  0, 232, 203,  35, 141, 101,  70, 174,   1, 233, 202,  34, 140, 100,  71, 175, 
  0, 233, 201,  32, 137,  96,  64, 169,   9, 224, 192,  41, 128, 105,  73, 160, 
  0, 234, 207,  37, 133, 111,  74, 160,  17, 251, 222,  52, 148, 126,  91, 177, 
  0, 235, 205,  38, 129, 106,  76, 167,  25, 242, 212,  63, 152, 115,  85, 190, 
  0, 236, 195,  47, 157, 113,  94, 178,  33, 205, 226,  14, 188,  80, 127, 147, 
  0, 237, 193,  44, 153, 116,  88, 181,  41, 196, 232,   5, 176,  93, 113, 156, 
  0, 238, 199,  41, 149, 123,  82, 188,  49, 223, 246,  24, 164,  74,  99, 141, 
  0, 239, 197,  42, 145, 126,  84, 187,  57, 214, 252,  19, 168,  71, 109, 130, 
  0, 240, 251,  11, 237,  29,  22, 230, 193,  49,  58, 202,  44, 220, 215,  39, 
  0, 241, 249,   8, 233,  24,  16, 225, 201,  56,  48, 193,  32, 209, 217,  40, 
  0, 242, 255,  13, 229,  23,  26, 232, 209,  35,  46, 220,  52, 198, 203,  57, 
  0, 243, 253,  14, 225,  18,  28, 239, 217,  42,  36, 215,  56, 203, 197,  54, 
  0, 244, 243,   7, 253,   9,  14, 250, 225,  21,  18, 230,  28, 232, 239,  27, 
  0, 245, 241,   4, 249,  12,   8, 253, 233,  28,  24, 237,  16, 229, 225,  20, 
  0, 246, 247,   1, 245,   3,   2, 244, 241,   7,   6, 240,   4, 242, 243,   5, 
  0, 247, 245,   2, 241,   6,   4, 243, 249,  14,  12, 251,   8, 255, 253,  10, 
  0, 248, 235,  19, 205,  53,  38, 222, 129, 121, 106, 146,  76, 180, 167,  95, 
  0, 249, 233,  16, 201,  48,  32, 217, 137, 112,  96, 153,  64, 185, 169,  80, 
  0, 250, 239,  21, 197,  63,  42, 208, 145, 107, 126, 132,  84, 174, 187,  65, 
  0, 251, 237,  22, 193,  58,  44, 215, 153,  98, 116, 143,  88, 163, 181,  78, 
  0, 252, 227,  31, 221,  33,  62, 194, 161,  93,  66, 190, 124, 128, 159,  99, 
  0, 253, 225,  28, 217,  36,  56, 197, 169,  84,  72, 181, 112, 141, 145, 108, 
  0, 254, 231,  25, 213,  43,  50, 204, 177,  79,  86, 168, 100, 154, 131, 125, 
  0, 255, 229,  26, 209,  46,  52, 203, 185,  70,  92, 163, 104, 151, 141, 114,
};
uint8 tab_sbox[256] = {
	0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
	0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0,	0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
	0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc,	0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
	0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a,	0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
	0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0,	0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
	0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b,	0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
	0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85,	0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
	0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5,	0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
	0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17,	0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
	0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88,	0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
	0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c,	0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
	0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9,	0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
	0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6,	0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
	0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e,	0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
	0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94,	0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
	0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68,	0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
};
uint8 Logtable[256] = {   
  0,   0,  25,   1,  50,   2,  26, 198,  75, 199,  27, 104,  51, 238, 223,   3,    
100,   4, 224,  14,  52, 141, 129, 239,  76, 113,   8, 200, 248, 105,  28, 193,    
125, 194,  29, 181, 249, 185,  39, 106,  77, 228, 166, 114, 154, 201,   9, 120,    
101,  47, 138,   5,  33,  15, 225,  36,  18, 240, 130,  69,  53, 147, 218, 142,    
150, 143, 219, 189,  54, 208, 206, 148,  19,  92, 210, 241,  64,  70, 131,  56,    
102, 221, 253,  48, 191,   6, 139,  98, 179,  37, 226, 152,  34, 136, 145,  16,    
126, 110,  72, 195, 163, 182,  30,  66,  58, 107,  40,  84, 250, 133,  61, 186,    
 43, 121,  10,  21, 155, 159,  94, 202,  78, 212, 172, 229, 243, 115, 167,  87,    
175,  88, 168,  80, 244, 234, 214, 116,  79, 174, 233, 213, 231, 230, 173, 232,    
 44, 215, 117, 122, 235,  22,  11, 245,  89, 203,  95, 176, 156, 169,  81, 160,    
127,  12, 246, 111,  23, 196,  73, 236, 216,  67,  31,  45, 164, 118, 123, 183,    
204, 187,  62,  90, 251,  96, 177, 134,  59,  82, 161, 108, 170,  85,  41, 157,    
151, 178, 135, 144,  97, 190, 220, 252, 188, 149, 207, 205,  55,  63,  91, 209,    
 83,  57, 132,  60,  65, 162, 109,  71,  20,  42, 158,  93,  86, 242, 211, 171,    
 68,  17, 146, 217,  35,  32,  46, 137, 180, 124, 184,  38, 119, 153, 227, 165,    
103,  74, 237, 222, 197,  49, 254,  24,  13,  99, 140, 128, 192, 247, 112,   7,    
};  
uint8 Alogtable[256] = {   
  1,   3,   5,  15,  17,  51,  85, 255,  26,  46, 114, 150, 161, 248,  19,  53,    
 95, 225,  56,  72, 216, 115, 149, 164, 247,   2,   6,  10,  30,  34, 102, 170,    
229,  52,  92, 228,  55,  89, 235,  38, 106, 190, 217, 112, 144, 171, 230,  49,    
 83, 245,   4,  12,  20,  60,  68, 204,  79, 209, 104, 184, 211, 110, 178, 205,    
 76, 212, 103, 169, 224,  59,  77, 215,  98, 166, 241,   8,  24,  40, 120, 136,    
131, 158, 185, 208, 107, 189, 220, 127, 129, 152, 179, 206,  73, 219, 118, 154,    
181, 196,  87, 249,  16,  48,  80, 240,  11,  29,  39, 105, 187, 214,  97, 163,    
254,  25,  43, 125, 135, 146, 173, 236,  47, 113, 147, 174, 233,  32,  96, 160,    
251,  22,  58,  78, 210, 109, 183, 194,  93, 231,  50,  86, 250,  21,  63,  65,    
195,  94, 226,  61,  71, 201,  64, 192,  91, 237,  44, 116, 156, 191, 218, 117,    
159, 186, 213, 100, 172, 239,  42, 126, 130, 157, 188, 223, 122, 142, 137, 128,    
155, 182, 193,  88, 232,  35, 101, 175, 234,  37, 111, 177, 200,  67, 197,  84,    
252,  31,  33,  99, 165, 244,   7,   9,  27,  45, 119, 153, 176, 203,  70, 202,    
 69, 207,  74, 222, 121, 139, 134, 145, 168, 227,  62,  66, 198,  81, 243,  14,    
 18,  54,  90, 238,  41, 123, 141, 140, 143, 138, 133, 148, 167, 242,  13,  23,    
 57,  75, 221, 124, 132, 151, 162, 253,  28,  36, 108, 180, 199,  82, 246,   1,    
}; 
// constants
uint8 Rcon[10][4] = {
	{ 0x01, 0x00, 0x00, 0x00 },
	{ 0x02, 0x00, 0x00, 0x00 },
	{ 0x04, 0x00, 0x00, 0x00 },
	{ 0x08, 0x00, 0x00, 0x00 },
	{ 0x10, 0x00, 0x00, 0x00 },
	{ 0x20, 0x00, 0x00, 0x00 },
	{ 0x40, 0x00, 0x00, 0x00 },
	{ 0x80, 0x00, 0x00, 0x00 },
	{ 0x1b, 0x00, 0x00, 0x00 },
	{ 0x36, 0x00, 0x00, 0x00 }
};
// inner variables
uint8 tempVecM[2] = {0}; // ## value : n
uint8 tempVec2M[2] = {0}; // ## value : n
uint8 tempVecL[2] = {0}; // ## value : n
uint8 tempVec2L[2] = {0}; // ## value : n
uint8 tempVecLs[64]= {0}; // ## value : 2 * n * L
uint8 tempVec2Ls[64] = {0}; // ## value : 2 * n * L
uint8 tempy[1] = {0}; // ## value : k
// for random
uint8 Rand[2] = {0}; // ## value : n * m store the true random number generated by RNG 
uint8 random1[4] = {0}; // ## value : n * n store the R1 
uint8 random2[64] = {0}; // ## value : n * n * L store the R2
// for CodeL
uint8 vecy[2] = {0}; // ## value : k + m
// Parameter
int k = 1; // ## 
int m = 1; // ##
int n = 2; // ##
int L = 16; // ## L = 16 / k
// -------------------

/* Public Functions */
void rawKeyExpansion(uint8* key, uint8* keys);
void AES_Encrypt_GCB(uint8* dst);
// -------------------

/* Private Functions */
// AES
void AddRoundKey(uint8* src1, uint8* src2, uint8* res);
void Sbox(uint8* src, uint8* dst);
void ShiftRows(uint8* src, uint8* dst);
void MixColumn(uint8* src, uint8* dst);
// Masking
extern void gene_Encode(uint8* src, uint8* res);
extern void gene_Decode(uint8* src, uint8* res);
extern void Gadget_CodeMuL(uint8* src1, uint8* src2, uint8* res);
extern void Gadget_CodeL(uint8* src, uint8* res, int para);
extern void Gadget_CodeLs(uint8* src, uint8* res, int L, int para);
// random
void getRandom1(); // set the random1
void getRandom2(uint8* randomAdr); // set the random2
// for keys generation
static void vecSetSub(uint8* src, uint8* dst, int start_index, int num);
static void vecExtraSub(uint8* src, uint8* dst, int start_index, int num);
static void Tbox(uint8* src, uint8* dst, int round);
// asm
extern void GetR1();
extern void GetR2(uint8* randomAdr);
// ------------------

// AES-128 masked by code-based masking
void AES_Encrypt_GCB(uint8* dst)
{

	int i = 0, j = 0;
	uint8 ptstate[32] = {0}; // ## value : n * L
	uint8 code_pt[32] = {0}; // ## value : n * L
	uint8 code_key[32] = {0}; // ## value : n * L
	uint8 input_pt[32] = {0}; // ## value : (k + m) * L
	uint8 input_key[32] = {0}; // ## value : (k + m) * L
   // uint8 mask1;
   // uint8 mask2;
	
	// prepare input
	for (i = 0;i < L;i++)
	{
		for (j = 0;j < k;j++)
		{
			input_pt[i * (k + m) + j] = pt_aes[(i * k) + j];
			input_key[i * (k + m) + j] = key_aes[(i * k) + j];
		}
		for (j = k;j < k + m;j++)
		{
			//input_pt[i * (k + m) + j] = 0x23;
			//input_key[i * (k + m) + j] = 0x45;
			genRandByte(&(input_pt[i * (k + m) + j]));
			genRandByte(&(input_key[i * (k + m) + j]));
		}
	}
	
	// encode
	for (i = 0;i < L;i++)
	{
		gene_Encode(&(input_pt[i * (k + m)]), &(code_pt[i * n]));
		gene_Encode(&(input_key[i * (k + m)]), &(code_key[i * n]));
	}

	ResetCYCCNT(); // start the counter

   //trigger_1();

	AddRoundKey(code_pt, code_key, ptstate);

   for(i = 0;i < 9;i++)
   {
      for (j = 0;j < L;j++)
      {
         //if (i==0 && j==0)
         //   trigger_1();
         Sbox(&(ptstate[j * n]), &(code_pt[j * n]));
         //if (i==0 && j==1)
         //   trigger_0();
      }
      //if (i==0) {mask1 = code_pt[4];} // for transfering the intermediate value
      //if (i == 2) trigger_0();
		ShiftRows(code_pt, ptstate);
		MixColumn(ptstate, code_pt);
		// prepare input key
		for (int ii = 0;ii < L;ii++)
		{
			for (int jj = 0;jj < k;jj++) input_key[ii * (k + m) + jj] = keys[(i + 1) * 16 + ii * k + jj];
			//for (int jj = k;jj < m;jj++) input_key[ii * (k + m) + jj] = 0x67;
			for (int jj = k;jj < k + m;jj++) genRandByte(&(input_key[ii * (k + m) + jj]));
		}
		for (int ii = 0;ii < L;ii++) gene_Encode(&(input_key[ii * (k + m)]), &(code_key[ii * n]));
		AddRoundKey(code_key, code_pt, ptstate);
	}	
	for (j = 0;j < L;j++) Sbox(&(ptstate[j * n]), &(code_pt[j * n]));
	ShiftRows(code_pt, ptstate);
	for (int ii = 0;ii < L;ii++)
		{
			for (int jj = 0;jj < k;jj++) input_key[ii * (k + m) + jj] = keys[10 * 16 + ii * k + jj];
			//for (int jj = k;jj < m;jj++) input_key[ii * (k + m) + jj] = 0x89;
			for (int jj = k;jj < k + m;jj++) genRandByte(&(input_key[ii * (k + m) + jj]));
		}
	for (int ii = 0;ii < L;ii++) gene_Encode(&(input_key[ii * (k + m)]), &(code_key[ii * n]));
	AddRoundKey(code_key, ptstate, code_pt);
	
	sys_val = ReadCYCCNT(); // read the counter
	
	// decode
	for (i = 0;i < L;i++) gene_Decode(&(code_pt[i * n]), &(dst[i * k]));

	//Usart_Send(&mask1, 1); // transfer the intermediate value
   //Usart_Send(&mask2, 1);
}

// key expansion (not in the masked domain)
void rawKeyExpansion(uint8* key, uint8* keys)
{
	int i, j;
	uint8 temp1[4] = { 0 };
	uint8 temp2[4] = { 0 };

	// the 0th key
	for (i = 0; i < 16; i++) keys[i] = key[i];

	// the ith key
	for (i = 0; i < 10; i++)
	{
		// the 0th sub 
		vecExtraSub(&(keys[i * 16]), temp1, 12, 4);
		Tbox(temp1, temp2, i);
		vecExtraSub(&(keys[i * 16]), temp1, 0, 4);
		for (j = 0; j < 4; j++) temp2[j] ^= temp1[j];
		vecSetSub(temp2, &(keys[(i + 1) * 16]), 0, 4);
		// the 1th sub
		vecExtraSub(&(keys[i * 16]), temp1, 4, 4);
		for (j = 0; j < 4; j++) temp2[j] ^= temp1[j];
		vecSetSub(temp2, &(keys[(i + 1) * 16]), 4, 4);
		// the 2th sub
		vecExtraSub(&(keys[i * 16]), temp1, 8, 4);
		for (j = 0; j < 4; j++) temp2[j] ^= temp1[j];
		vecSetSub(temp2, &(keys[(i + 1) * 16]), 8, 4);
		// the 3th sub
		vecExtraSub(&(keys[i * 16]), temp1, 12, 4);
		for (j = 0; j < 4; j++) temp2[j] ^= temp1[j];
		vecSetSub(temp2, &(keys[(i + 1) * 16]), 12, 4);
	}

}

// for AES

void AddRoundKey(uint8* src1, uint8* src2, uint8* res)
{
	for (int i = 0;i < L;i++) getRandom2(&(random2[i * n * n]));
	/* concatenation */
	uint8 temp[64] = {0}; // ## value : 2 * n * L
	for (int i = 0;i < n * L;i++)
	{
		temp[i] = src1[i];
		temp[n * L + i] = src2[i];
	}
	Gadget_CodeLs(temp, res, 2 * L, 0);
}

void Sbox(uint8* src, uint8* dst)
{
	uint8 temp_2[2] = {0}; // ## value : n
	uint8 temp_4[2] = {0}; // ## value : n
	uint8 temp_16[2] = {0}; // ## value : n
	uint8 temp_n[2] = {0}; // ## value : n

	getRandom2(random2);
	Gadget_CodeL(src, temp_2, 0);
	
	getRandom1();
	getRandom2(random2);
	Gadget_CodeMuL(src, temp_2, temp_n);
	
	getRandom2(random2);
	Gadget_CodeL(temp_n, temp_4, 1);
	
	getRandom1();
	getRandom2(random2);
	Gadget_CodeMuL(temp_4, temp_n, dst);
	
	getRandom2(random2);
	Gadget_CodeL(dst, temp_16, 2);
	
	getRandom1();
	getRandom2(random2);
	Gadget_CodeMuL(temp_16, temp_4, src);
	
	getRandom1();
	getRandom2(random2);
	Gadget_CodeMuL(src, temp_2, temp_n);
	
	getRandom2(random2);
	Gadget_CodeL(temp_n, dst, 3);
	
}

void ShiftRows(uint8* src, uint8* dst)
{
	for (int i = 0;i < L;i++) getRandom2(&(random2[i * n * n]));
	Gadget_CodeLs(src, dst, L, 1);
}

void MixColumn(uint8* src, uint8* dst)
{
	for (int i = 0;i < L;i++) getRandom2(&(random2[i * n * n]));
	Gadget_CodeLs(src, dst, L, 2);
}

// for raw key expansion
void vecSetSub(uint8* src, uint8* dst, int start_index, int num)
{
	for (int i = 0; i < num; i++) dst[start_index + i] = src[i];
}

void vecExtraSub(uint8* src, uint8* dst, int start_index, int num)
{
	for (int i = 0; i < num; i++) dst[i] = src[i + start_index];
}

void Tbox(uint8* src, uint8* dst, int round)
{
	int i;
	// rotate to left 1 byte
	dst[0] = src[1];
	dst[1] = src[2];
	dst[2] = src[3];
	dst[3] = src[0];
	// subByte
	for (i = 0; i < 4; i++) dst[i] = tab_sbox[dst[i]];
	// xor Rcon
	for (i = 0; i < 4; i++) dst[i] = dst[i] ^ Rcon[round][i];

}

// construct R1
void getRandom1()
{
	for (int i = 0; i < (n * m); i++)
	{
		genRandByte(&Rand[i]);
	   //Rand[i] = 0xab;
	}
	GetR1();
}

// construct R2
void getRandom2(uint8* randomAdr)
{
	for (int i = 0; i < (n * m); i++)
	{
	   genRandByte(&Rand[i]);
	   //Rand[i] = 0xcd;
	}
	GetR2(randomAdr);
}


