/***************************************************************************************/
/* @file   gadget_CodeLs.s                                                             */
/* @author yangx                                                                       */
/* @date   2022.4.25                                                                   */
/* @brief  Implementation of Gadget CodeLs masked by code-based masking. In this code, */
/*         the gadget is utilized to implement AddRoundKey, ShiftRows and MixColumns.  */
/*         The generator matrix can be fed with any legal input.                       */
/***************************************************************************************/

.section .data

		paraAdr: .word 0x00000000, 0x00000000, 0x00000000, 0x00000000
			
.section .text
.syntax unified
.thumb
.global Gadget_CodeLs
	
Gadget_CodeLs:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src */
		/* r1 stores the address of result */
		/* r2 stores the number of block l = 16 / k */
		/* r3 stores the parameter */
		/* 0 : addRoundKey */
		/* 1 : shiftRows */
		/* 2 : mixColumn */
		
		/* fix some Regs */
		MOV R4, R0
		MOV R5, R1
		MOV R6, R2
		MOV R7, R3
		LDR R3, =k
		LDRB R3, [R3]
		LDR R8, =n
		LDRB R8, [R8]
		LDR R9, =paraAdr
		STR R7, [R9] // store the para
		STR R4, [R9, #8] // store the address of src
		STR R5, [R9, #4] // store the address of dst
		
		/*-------The 0th row of matrix--------*/
		// Part A
		// call Part A L (16 / k) times
		LDR R9, =matM
		LDR R10, =tempVecLs
		MOV R12, #0 // counter
loopA1:
		// Outer Product
		LDRB R0, [R4] // prepare the byte
		// Multiply by M0
		MOV R11, #0 // counter
		MOV R1, R9 // refresh M0
loopk11:
		// r0 already stores the byte
		// r1 already stores the address of vec
		// r2 will store the res
		// r3 already stores the num of iterations : k
		BL mul_byte_vec_and_Eor
		// r2 stores the res 
		STRB R2, [R10, R11] // store
		// loop block
		ADD R11, #1
		ADD R1, R8
		CMP R11, R3 // loop <= k
		IT NE 
		BNE loopk11
		// next block
		ADD R12, #1
		ADD R4, R8
		ADD R10, R3
		CMP R12, R6
		IT NE
		BNE loopA1
		// tempVecLs(R10) stores the vec : (1, 16) or (1, 32)
		
		// Fs Function
		LDR R0, =tempVecLs // r0 stores the src
		LDR R1, =tempVec2Ls // r1 stores the res
		// according to the para
		CMP R7, #0
		ITTT EQ
		MOVEQ R2, #2
		UDIVEQ R6, R6, R2
		BEQ Fs_AddRoundKey1
		CMP R7, #1
		IT EQ
		BEQ Fs_ShiftRows1
		CMP R7, #2
		IT EQ
		BEQ Fs_MixColumns1
Fs_AddRoundKey1:
		LDR R2, [R0]
		LDR R10, [R0, #16]
		EOR R2, R10
		STR R2, [R1]
		LDR R2, [R0, #4]
		LDR R10, [R0, #20]
		EOR R2, R10
		STR R2, [R1, #4]
		LDR R2, [R0, #8]
		LDR R10, [R0, #24]
		EOR R2, R10
		STR R2, [R1, #8]
		LDR R2, [R0, #12]
		LDR R10, [R0, #28]
		EOR R2, R10
		STR R2, [R1, #12]
		B PartB1
Fs_ShiftRows1:
		LDR R2, [R0]
		STRB R2, [R1]
		LSR R2, R2, #8
		STRB R2, [R1, #13]
		LSR R2, R2, #8
		STRB R2, [R1, #10]
		LSR R2, R2, #8
		STRB R2, [R1, #7]
		LDR R2, [R0, #4]
		STRB R2, [R1, #4]
		LSR R2, R2, #8
		STRB R2, [R1, #1]
		LSR R2, R2, #8
		STRB R2, [R1, #14]
		LSR R2, R2, #8
		STRB R2, [R1, #11]
		LDR R2, [R0, #8]
		STRB R2, [R1, #8]
		LSR R2, R2, #8
		STRB R2, [R1, #5]
		LSR R2, R2, #8
		STRB R2, [R1, #2]
		LSR R2, R2, #8
		STRB R2, [R1, #15]
		LDR R2, [R0, #12]
		STRB R2, [R1, #12]
		LSR R2, R2, #8
		STRB R2, [R1, #9]
		LSR R2, R2, #8
		STRB R2, [R1, #6]
		LSR R2, R2, #8
		STRB R2, [R1, #3]
		B PartB1
Fs_MixColumns1:
		// r0 already stores the address of vec
		// r1 already stores the address of res
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		SUB R1, #12 // refresh the address of r1
		// tempVec2Ls(R1) stores the result : (1, 16)
		
PartB1:
		LDR R10, =random2
		MUL R9, R8, R8
		MOV R12, #0 // counter
loopB1:
		// Multiply by A & Refresh with R2 & Store
		MOV R11, #0 // counter
		// the first k byte
loopk12:
		LDRB R0, [R1, R11]
		LDRB R2, [R10, R11]
		EOR R0, R2
		STRB R0, [R5, R11]
		// loop block
		ADD R11, #1
		CMP R11, R3 // loop <= k
		IT NE
		BNE loopk12
		// the last (n - k) byte
loopnk1:
		LDRB R0, [R10, R11]
		STRB R0, [R5, R11]
		// loop block
		ADD R11, #1
		CMP R11, R8 // loop <= n
		IT NE
		BNE loopnk1
		// next block
		ADD R12, #1
		ADD R1, R3
		ADD R5, R8
		ADD R10, R9
		CMP R12, R6 // loop <= L
		IT NE
		BNE loopB1
		
		/*-------The ith row of matrix--------*/
		LDR R9, =matM
		MUL R10, R3, R8
		ADD R9, R10 // refresh Ms
		MOV R12, #1 // counter for row
loopRow:
		// Part A
		// call Part A L times
		LDR R10, =paraAdr
		LDR R4, [R10] // load the para
		CMP R4, #0
		ITT EQ
		MOVEQ R5, #2
		MULEQ R6, R6, R5
		LDR R4, [R10, #8] 
		ADD R4, R12 // refresh the address of src
		MOV R11, #0 // counter for L blocks
		LDR R5, =tempVecLs
loopA2:
		MOV R10, #0 // counter
		// Outer Product
		LDRB R0, [R4]
		MOV R1, R9 // refresh Mi
loopk21:
		// r0 already stores the byte
		// r1 already stores the address of vec
		// r2 will store the res
		// r3 already stores the num of iterations : k
		BL mul_byte_vec_and_Eor
		// r2 stores the res
		STRB R2, [R5, R10]
		// loop block
		ADD R10, #1
		ADD R1, R8 
		CMP R10, R3
		IT NE
		BNE loopk21
		// next block
		ADD R11, #1
		ADD R4, R8
		ADD R5, R3
		CMP R11, R6 // loop <= L
		IT NE
		BNE loopA2
		// tempVecLs(R5) stores the vec : (1, 16) or (1, 32)
		
		// Fs Function
		LDR R10, =paraAdr
		LDR R2, [R10] // r2 stores the para
		LDR R0, =tempVecLs // r10stores the address of src
		LDR R1, =tempVec2Ls // r1 stores the address of res
		// according to the para
		CMP R2, #0
		ITTT EQ
		MOVEQ R5, #2
		UDIVEQ R6, R6, R5 
		BEQ Fs_AddRoundKey2
		CMP R2, #1
		IT EQ
		BEQ Fs_ShiftRows2
		CMP R2, #2
		IT EQ
		BEQ Fs_MixColumns2
Fs_AddRoundKey2:
		LDR R5, [R0]
		LDR R2, [R0, #16]
		EOR R5, R2
		STR R5, [R1]
		LDR R5, [R0, #4]
		LDR R2, [R0, #20]
		EOR R5, R2
		STR R5, [R1, #4]
		LDR R5, [R0, #8]
		LDR R2, [R0, #24]
		EOR R5, R2
		STR R5, [R1, #8]
		LDR R5, [R0, #12]
		LDR R2, [R0, #28]
		EOR R5, R2
		STR R5, [R1, #12]
		B PartB2
Fs_ShiftRows2:
		LDR R2, [R0]
		STRB R2, [R1]
		LSR R2, R2, #8
		STRB R2, [R1, #13]
		LSR R2, R2, #8
		STRB R2, [R1, #10]
		LSR R2, R2, #8
		STRB R2, [R1, #7]
		LDR R2, [R0, #4]
		STRB R2, [R1, #4]
		LSR R2, R2, #8
		STRB R2, [R1, #1]
		LSR R2, R2, #8
		STRB R2, [R1, #14]
		LSR R2, R2, #8
		STRB R2, [R1, #11]
		LDR R2, [R0, #8]
		STRB R2, [R1, #8]
		LSR R2, R2, #8
		STRB R2, [R1, #5]
		LSR R2, R2, #8
		STRB R2, [R1, #2]
		LSR R2, R2, #8
		STRB R2, [R1, #15]
		LDR R2, [R0, #12]
		STRB R2, [R1, #12]
		LSR R2, R2, #8
		STRB R2, [R1, #9]
		LSR R2, R2, #8
		STRB R2, [R1, #6]
		LSR R2, R2, #8
		STRB R2, [R1, #3]
		B PartB2
Fs_MixColumns2:
		// r0 already stores the address of vec
		// r1 already stores the address of res
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		ADD R0, #4 // prepare the address of vec
		ADD R1, #4 // prepare the address of result
		BL mul_mat_vec_4
		SUB R1, #12 // refresh the address of r1
		// tempVec2Ls(R1) stores the result : (1, 16)
		
PartB2:
		LDR R10, =paraAdr
		LDR R5, [R10, #4] // r5 stores the address of result
		LDR R7, =random2
		MUL R4, R12, R8
		ADD R7, R4 // refresh R2
		MUL R4, R8, R8 // r4 = n * n
		MOV R11, #0 // counter for block
loopPartB2:
		// Multiply by A & Refresh with R2 & Eor and Store
		MOV R10, #0 // counter
		// the first k byte
loopk22:
		LDRB R0, [R7, R10]
		LDRB R2, [R1, R10]
		EOR R0, R2
		LDRB R2, [R5, R10]
		EOR R0, R2
		STRB R0, [R5, R10]
		// loop block
		ADD R10, #1
		CMP R10, R3 // loop <= k
		IT NE
		BNE loopk22
		// last (n - k) byte
loopnk2:
		LDRB R0, [R7, R10]
		LDRB R2, [R5, R10]
		EOR R0, R2
		STRB R0, [R5, R10]
		// loop block
		ADD R10, #1
		CMP R10, R8 // loop <= n
		IT NE
		BNE loopnk2
		// next block
		ADD R11, #1
		ADD R7, R4
		ADD R1, R3
		ADD R5, R8
		CMP R11, R6
		IT NE
		BNE loopPartB2
		
		// next row
		ADD R12, #1
		MUL R4, R3, R8
		ADD R9, R4 // refresh Ms
		CMP R12, R8 // loop <= n
		IT NE
		BNE loopRow
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	