/****************************************************************************/
/* @file   gadget_CodeMuL.s                                                 */
/* @author yangx                                                            */
/* @date   2022-4-25                                                        */
/* @brief  Implementation of gadget CodeMuL masked by code-based masking.   */
/*         This function calculates the byte_wise multiplication of two     */
/*		   vectors over F_{2^8}^{n} in the masked domain. And the generator */
/*         matrix can be fed with any legal input.                          */
/****************************************************************************/

.section .data
.section .text
.syntax unified
.thumb
.global Gadget_CodeMuL
	
Gadget_CodeMuL:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* r0 stores the address of src1 */
		/* r1 stores the address of src2 */
		/* r2 stores the address of result */
		
		/* fix some Regs */
		MOV R4, R0
		MOV R5, R1
		MOV R6, R2
		LDR R3, =n
		LDRB R3, [R3] 
		LDR R7, =matM // Ms
		LDR R8, =random1 // R1
		LDR R9, =random2 // R2
		
		/*-------The 0th row of matrix--------*/
		// Outer Product & Refresh with R1
		LDRB R0, [R4] // prepare the byte
		// r1 already stores the address of vec
		// r2 already stores the address of res
		// r3 already stores the num of iterations : n
		BL mul_byte_vec_and_Ref
		// result stores the vec : (1, n)
		
		// Multiply by M0 & Multiply by A & Refresh with R2 & Store
		MOV R12, #0 // counter 
		MOV R1, R7 // prepare the address of vec2
		MOV R2, R3 // prepare the num of iterations : n
		LDR R11, =k
		LDRB R11, [R11]
		// the forth k bytes
loopk1:
		MOV R0, R6 // prepare the address of vec1
		// r1 already stores the address of vec2
		// r2 already stores the num of iterations : n
		BL byte_mul_and_Eor
		// r0 stores the res
		LDRB R10, [R9, R12]
		EOR R0, R10
		STRB R0, [R6, R12]
		// loop block
		ADD R12, #1
		ADD R1, R3
		CMP R12, R11
		IT NE
		BNE loopk1
		
		// the last (n - k) byte
loopnk1:
		LDRB R10, [R9, R12]
		STRB R10, [R6, R12]
		// loop block
		ADD R12, #1
		CMP R12, R3
		IT NE 
		BNE loopnk1
		
		/*-------The ith row of matrix--------*/
		MOV R12, #1 // counter for (n - 1) rows
		MOV R10, R11 // r10 = k
loopn1:
		ADD R8, R3 // refresh R1
		ADD R9, R3 // refresh R2
		MUL R10, R3 // r10 = n * k
		ADD R7, R10 // move the pointer of Ms
		
		// Outer Product & Refresh with R1
		LDRB R0, [R4, R12] // prepare the byte
		MOV R1, R5 // prepare the address of vec
		LDR R2, =tempVecM // prepare the address of res
		// r3 already stores the num of iterations : n
		BL mul_byte_vec_and_Ref
		// tempVecM stores the vec : (1, n)
		
		// Multiply by Mi & Multiply by A & Refresh with R2 & Store
		MOV R11, #0 // counter 
		MOV R1, R7 // prepare the address of vec2
		MOV R2, R3 // prepare the num of iterations : n
		LDR R3, =k
		LDRB R3, [R3]
		// the forth k bytes
loopk2:
		LDR R0, =tempVecM // prepare the address of vec1
		// r1 already stores the address of vec2
		// r2 already stores the num of iterations : n
		BL byte_mul_and_Eor
		// r0 stores the res
		LDRB R10, [R9, R11]
		EOR R0, R10
		LDRB R10, [R6, R11]
		EOR R0, R10
		STRB R0, [R6, R11]
		// loop block
		ADD R11, #1
		ADD R1, R2
		CMP R11, R3
		IT NE
		BNE loopk2
		
		// the last (n - k) byte
		MOV R10, R3 // r10 = k
		MOV R3, R2 // r3 = n
loopnk2:
		LDRB R1, [R9, R11]
		LDRB R0, [R6, R11]
		EOR R0, R1
		STRB R0, [R6, R11]
		// loop block
		ADD R11, #1
		CMP R11, R3
		IT NE 
		BNE loopnk2
		
		// next row
		ADD R12, #1
		CMP R12, R3
		IT NE
		BNE loopn1
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	