/************************************************************************/
/* @file   main.c                                                       */
/* @author yangx                                                        */
/* @date   2022-4-25                                                    */
/* @brief  communicte with STM32F407 by USART                           */
/*         instruct the board to do                                     */
/*         -- set key                                                   */
/*         -- encryption by AES-128 masked code-based masking           */
/*         -- return the cyphertext                                     */
/*         Note that before encryption, the key sholud be set first.    */
/*         the generator matrix can be fed with legal input.            */
/************************************************************************/

#include <stm32f4xx.h>
#include <misc.h>			
#include <stm32f4xx_usart.h> 
#include <stm32f4xx_rcc.h>

#include "aes.h"
#include "comm.h"
#include "random.h"
#include "DWT_clock.h"
#include "trigger.h"

// Global variables
// ---------------------
uint8 key_aes[16] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f
};

uint8 pt_aes[16] = { 
	0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
	0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff
};

uint8 keys[16 * 11] = {0}; // keys for 11 rounds

unsigned int sys_val = 0; // for counter

int main()
{
	int i, j;

	uint8 cmd;
	uint8 APDUcmd[2], APDUdata[16];

	uint8 ct_aes[16] = {0};

	// Communication
	Usart_Init(9600);

	// random
	initRandSrc();

	// initialize the DWT for clock cycles counting
	InitCYCCNT();
	StartCYCCNT();

	// trigger
	trigger_init();
	trigger_0();

	// locked loop
	// ---------------------------
	while (1)
	{

		// get the instruction
		for (i = 0; i < 2; i++)
		{
			APDUcmd[i] = Usart_GetByte();
		}
		if (APDUcmd[0] == CLA) Usart_Send(&APDUcmd[1], 1);


		switch (APDUcmd[1])
		{
		// set the key
		case INS_MAKE_AES_KEY:
			for (i = 0; i < 16; i++)
			{
				APDUdata[i] = Usart_GetByte();
				key_aes[i] = APDUdata[i];
			}
			rawKeyExpansion(key_aes, keys);
			cmd = 0x91;
			Usart_Send(&cmd, 1); // reply
			break;
		case INS_ENCRYPT_AES:
			//ResetCYCCNT();
			for (i = 0; i < 16; i++)
			{
				APDUdata[i] = Usart_GetByte();
				pt_aes[i] = APDUdata[i];
			}
			//sys_val = ReadCYCCNT();
			cmd = 0x92;
			Usart_Send(&cmd, 1); // reply

			// AES encryption
			AES_Encrypt_GCB(ct_aes);
			cmd = 0x93;
			Usart_Send(&cmd, 1);

			// return the cycle count
			cmd = (uint8)sys_val;
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);

			break;
		// for test
		case INS_TEST:
			rawKeyExpansion(key_aes, keys);
			AES_Encrypt_GCB(ct_aes);
			cmd = 0x95;
			Usart_Send(&cmd, 1);

			// return the cycle count
			cmd = (uint8)sys_val;
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);
			sys_val = sys_val >> 8;
			cmd = (uint8)(sys_val);
			Usart_Send(&cmd, 1);

			break;
		// get the ciphertext
		case INS_GET_RESPONSE:
			for (i = 0;i < 16;i++)
			{
				Usart_Send(&(ct_aes[i]), 1);
			}
			cmd = 0x94;
			Usart_Send(&cmd, 1); // reply
			break;
		default:
			break;
		}

	}
	

}

