/********************************************************************************/
/* @file   mul_byte_vec.s                                                       */
/* @author yangx                                                                */
/* @date   2022-4-25                                                            */
/* @brief  This function calculates the multiplication of a vector by a scalar. */
/********************************************************************************/

.section .data	  
.section .text
.syntax unified
.thumb
.global mul_byte_vec

/* input */
/* r0 stores the scalar (byte) */
/* r1 stores the address of vector */
/* r2 stores the address of result */
/* r3 stores the number of iterations */

/* output */
mul_byte_vec:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R12}
		/* r0 stores the scalar (byte) */
		/* r1 stores the address of vector */
		/* r2 stores the address of result */
		/* r3 stores the number of iterations */
		
		/* prepare the table address */
		LDR R4, =bhtable 
		LDR R5, =bltable 
		
		// prepare ah and al
		LSL R0, R0, #4
		MOV R12, #0 // counter for iterations
loopmby:
		LDRB R6, [R1, R12] // load factor2
		// look up the table
		ORR R7, R0, R6, LSR #4
		LDRB R7, [R4, R7]
		AND R6, #0xF
		ORR R6, R0, R6
		LDRB R6, [R5, R6]
		EOR R6, R7
		// store
		STRB R6, [R2, R12]
		
		ADD R12, #1
		CMP R12, R3
		IT NE
		BNE loopmby
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R12}
		/* return */
		BX LR
.end
	