/*********************************************************************************/
/* @file   mul_byte_vec_and_Eor.s                                                */
/* @author yangx                                                                 */
/* @date   2022-4-25                                                             */
/* @brief  This function calculates the multiplication of a vector by a scalar   */
/*         and then sums up the elements in the vector to give the final result. */
/*********************************************************************************/

.section .data	  
.section .text
.syntax unified
.thumb
.global mul_byte_vec_and_Eor

/* input */
/* r0 stores the scalar (byte) */
/* r1 stores the adress of vectore */
/* r2 stores the result */
/* r3 stores the number of iterations */

/* output */
mul_byte_vec_and_Eor:
		/* copy registers value to the stack */
		PUSH {R0,R1,R3,R4,R5,R6,R7,R12}
		/* r0 stores the scalar (byte) */
		/* r1 stores the adress of vectore */
		/* r2 stores the result */
		/* r3 stores the number of iterations */
		
		/* prepare the table address */
		LDR R4, =bhtable 
		LDR R5, =bltable 
		
		// prepare ah and al
		LSL R0, R0, #4
		MOV R12, #0 // counter for iterations
		MOV R2, #0 // store the result
loopmby:
		LDRB R6, [R1, R12] // load factor2
		// look up the table
		ORR R7, R0, R6, LSR #4
		LDRB R7, [R4, R7]
		AND R6, #0xF
		ORR R6, R0, R6
		LDRB R6, [R5, R6]
		EOR R6, R7
		// eor
		EOR R2, R6
		
		ADD R12, #1
		CMP R12, R3
		IT NE
		BNE loopmby
		
		/* copy registers value from the stack */
		POP {R0,R1,R3,R4,R5,R6,R7,R12}
		/* return */
		BX LR
.end
	