/*******************************************************************************/
/* @file   mul_mat_mat_1.s                                                     */
/* @author yangx                                                               */
/* @date   2022-4-25                                                           */
/* @brief  This function calculates the multiplication of the transpose of m*n */
/*         matrix H and the transpose of R1. R1 is over F_{2^8}^{n*m}.         */
/*******************************************************************************/

.section .data		   
.section .text
.syntax unified
.thumb
.global GetR1

/* output */
GetR1:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R11,R12,LR}
		/* fix some Regs */
		LDR R5, =matH1
		LDR R6, =Rand
		LDR R7, =random1
		LDR R2, =m
		LDRB R2, [R2]
		LDR R3, =n
		LDRB R3, [R3]
		SUB R4, R3, R2 // r4 = n - m
		
		/* the first (n - m) rows */
		MOV R12, #0 // counter for rows
loopRow1:
		MOV R11, #0 // counter for columns
		MOV R1, R6 // reset the address of mat2
loopCol1:
		MOV R0, R5 // prepare the address of mat1 row
		// r1 already stores the address of mat2 column
		// r2 already stores the num of iterations : m
		BL byte_mul_and_Eor
		// r0 stores the res (byte)
		STRB R0, [R7, R11]
		// loop block
		ADD R11, #1
		ADD R1, R2
		CMP R11, R3 // loop <= n
		IT NE
		BNE loopCol1
		// next row
		ADD R12, #1
		ADD R5, R2
		ADD R7, R3
		CMP R12, R4
		IT NE
		BNE loopRow1

		/* the last m rows */
loopRow2:
		MOV R11, #0
		MOV R5, #0
loopCol2:
		LDRB R0, [R6, R11]
		STRB R0, [R7, R5]
		// loop block
		ADD R11, R2
		ADD R5, #1
		CMP R5, R3
		IT NE
		BNE loopCol2
		// next row
		ADD R12, #1
		ADD R6, #1
		ADD R7, R3
		CMP R12, R3
		IT NE
		BNE loopRow2
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R11,R12,LR}
		/* return */
		BX LR
.end
	
