/***********************************************************************************/
/* @file   mul_mat_mat_2.s                                                         */
/* @author yangx                                                                   */
/* @date   2022-4-25                                                               */
/* @brief  This function calculates the multiplication of R2 and the m*n matrix H. */
/*         R2 is over F_{2^8}^{n*m}.                                               */
/***********************************************************************************/

.section .data		   
.section .text
.syntax unified
.thumb
.global GetR2

/* output */
GetR2:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R10,R11,R12,LR}
		/* r0 stores the address of result */
		
		/* fix some Regs */
		LDR R5, =Rand
		LDR R6, =matH1
		MOV R7, R0
		LDR R2, =m
		LDRB R2, [R2]
		LDR R3, =n
		LDRB R3, [R3]
		SUB R4, R3, R2
		
		MOV R12, #0 // counter for rows
loopRow:
		MOV R11, #0 // counter for columns
		MOV R1, R6 // reset the address of mat2
		/* the first (n - m) columns */
loopCol1:
		MOV R0, R5 // prepare the address of mat1 row
		// r1 already stores the address of mat2 column
		// r2 already stores the num of iterations : m
		BL byte_mul_and_Eor
		// r0 stores the res
		STRB R0, [R7, R11]
		// loop block
		ADD R11, #1
		ADD R1, R2
		CMP R11, R4 // loop <= (n - m)
		IT NE
		BNE loopCol1
		/* the last m columns */
		MOV R10, #0 // counter
loopCol2:
		LDRB R0, [R5, R10]
		STRB R0, [R7, R11]
		// loop block
		ADD R10, #1
		ADD R11, #1
		CMP R10, R2
		IT NE
		BNE loopCol2
		// next row
		ADD R12, #1
		ADD R5, R2
		ADD R7, R3
		CMP R12, R3
		IT NE
		BNE loopRow
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R10,R11,R12,LR}
		/* return */
		BX LR
.end
	
