/****************************************************************************/
/* @file   mul_mat_vec_4.s                                                  */
/* @author yangx                                                            */
/* @date   2022-4-11                                                        */
/* @brief  Implementation of the product for a matrix defined as            */
/*         [2,3,1,1;1,2,3,1;1,1,2,3;3,1,1,2] and a vector over F_{2^8}^{4}. */
/*		   Actually this is the MixColumns operated on a column.            */
/****************************************************************************/

.section .data  
.section .text
.syntax unified
.thumb
.global mul_mat_vec_4

/* input */
/* r0 stores the adress of vec */
/* r1 stores the adress of result */

/* output */
mul_mat_vec_4:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R8,R9,R10,R11,R12}
		/* r0 stores the address of vector */
        /* r1 stores the address of result */
		
		// prepare the tables address 
		LDR R3, =bhtable
		LDR R4, =bltable
		// load the vector
		LDR R0, [R0]
		MOV R8, #0xFF // mask
		
		/* deal with the 0th byte of vec */
		AND R2, R0, R8
		// x1 = tab1[a0] + tab2[a2] + other
		// x2 = a + other 
		// x3 = a + other 
		// x4 = tab1[a0] + tab2[a3] + other
		EOR R10, R2, R0, LSR #24 // r10 stores x2, x2 = a + d
		EOR R11, R2, R0, LSR #8 // r11 stores x3, x3 = a + b
		LSL R2, R2, #4
		LDRB R5, [R3, R2] // tab1[a0]
		ORR R6, R2, #2
		LDRB R6, [R4, R6] // tab2[a2]
		EOR R9, R5, R6 // x3 = tab1[a0] + tab2[a2]
		ORR R2, R2, #3
		LDRB R2, [R4, R2] // tab2[a3]
		EOR R12, R5, R2 // x4 = tab1[a0] + tab2[a3]
		/* deal with the 1th byte of vec */
		AND R2, R8, R0, LSR #8
		// x1 = tab1[b0] + tab2[b3] + other
		// x2 = tab1[b0] + tab2[b2] + other 
		// x3 = b + other
		// x4 = b + other
		EOR R12, R2 // x4 = tab1[a0] + tab2[a3] + b
		LSL R2, R2, #4
		LDRB R5, [R3, R2] // tab1[b0]
		ORR R6, R2, #3
		LDRB R6, [R4, R6] // tab2[b3]
		EOR R9, R5
		EOR R9, R6 // x1 = x1 + tab1[b0] + tab2[b3]
		ORR R2, R2, #2
		LDRB R2, [R4, R2] // tab2[b2]
		EOR R10, R5
		EOR R10, R2 // x2 = x2 + tab1[b0] + tab2[b2]
		/* deal with the 2th byte of vec */
		AND R2, R8, R0, LSR #16
		// x1 = c + other
		// x2 = tab1[c0] + tab2[c3] + other 
		// x3 = tab1[c0] + tab2[c2] + other
		// x4 = c + other
		EOR R9, R2 // x1 = x1 + c
		EOR R12, R2 // x4 = x4 + c
		LSL R2, R2, #4
		LDRB R5, [R3, R2] // tab1[c0]
		ORR R6, R2, #3
		LDRB R6, [R4, R6] // tab2[c3]
		EOR R10, R5
		EOR R10, R6 // x2 = x2 + tab1[c0] + tab2[c3]
		ORR R2, R2, #2
		LDRB R2, [R4, R2] // tab2[c2]
		EOR R11, R5
		EOR R11, R2 // x3 = x3 + tab1[c0] + tab2[c2]
		/* deal with the 3th byte of vec */
		LSR R2, R0, #24
		// x1 = d + other
		// x2 = d + other 
		// x3 = tab1[d0] + tab2[d3] + other
		// x4 = tab1[d0] + tab2[d2] + other
		EOR R9, R2 // x1 = x1 + d
		LSL R2, R2, #4
		LDRB R5, [R3, R2] // tab1[d0]
		ORR R6, R2, #3
		LDRB R6, [R4, R6] // tab2[d3]
		EOR R11, R5
		EOR R11, R6 // x3 = x3 + tab1[d0] + tab2[d3]
		ORR R2, R2, #2
		LDRB R2, [R4, R2] // tab2[d2]
		EOR R12, R5
		EOR R12, R2 // x4 = x4 + tab1[d0] + tab2[d2]
		
		STRB R9, [R1]
		STRB R10, [R1, #1]
		STRB R11, [R1, #2]
		STRB R12, [R1, #3]
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R8,R9,R10,R11,R12}
		/* return */
		BX LR
.end
	