/***********************************************************/
/* @file   gene_Encode.s                                   */
/* @author yangx                                           */
/* @date   2022-4-25                                       */
/* @brief  Implementation of encoding in masked domain.    */
/*         The generator matrix can be fed with any input. */
/***********************************************************/

.section .data	
.section .text
.syntax unified
.thumb
.global gene_Encode
gene_Encode:
		/* copy registers value to the stack */
		PUSH {R0,R1,R2,R3,R4,R5,R6,R7,R12,LR}
		/* r0 stores the address of vector (1, k + m) to be endcoded */
		/* r1 stores the address of codeword */
		
		/* fix some Regs */
		MOV R5, R0
		MOV R6, R1
		LDR R2, =m
		LDRB R2, [R2]
		LDR R3, =k
		LDRB R3, [R3]
		LDR R4, =n
		LDRB R4, [R4]
		LDR R1, =matH1
		
		/* the forth k columns */
		MOV R12, #0 // counter
loopk:
		ADD R0, R5, R3 // prepare the address of vec1
		// r1 already stores the address of vec2
		// r2 already stores the num of iterations : m
		BL byte_mul_and_Eor
		// r0 stores the result
		LDRB R7, [R5, R12]
		EOR R0, R7
		STRB R0, [R6, R12]
		// loop block
		ADD R12, #1
		ADD R1, R2
		CMP R12, R3 // loop <= k
		IT NE 
		BNE loopk
		
		/* the middle (n - k - m) columns */
		SUB R7, R4, R2 // n - m
		CMP R7, R3
		IT EQ
		BEQ Colm
loopnmk:
		ADD R0, R5, R3 // prepare the address of vec1
		// r1 already stores the address of vec2
		// r2 already stores the num of iterations : m
		BL byte_mul_and_Eor
		// r0 stoare the result
		STRB R0, [R6, R12]
		// loop block
		ADD R12, #1
		ADD R1, R2
		CMP R12, R7 // loop <= n - m
		IT NE
		BNE loopnmk
		
		/* the last m columns */
Colm:
		ADD R0, R5, R3
		MOV R7, #0 // counter
loopm:
		LDRB R3, [R0, R7]
		STRB R3, [R6, R12]
		// loop block
		ADD R7, #1
		ADD R12, #1
		CMP R12, R4 // loop <= n
		IT NE 
		BNE loopm
		
		/* copy registers value from the stack */
		POP {R0,R1,R2,R3,R4,R5,R6,R7,R12,LR}
		/* return */
		BX LR
.end
	